package forge.cfg;

import java.util.AbstractSet;
import java.util.Iterator;
import java.util.Set;

import edu.mit.csail.sdg.annotations.Returns;
import edu.mit.csail.sdg.util.collections.Iterators;
import forge.program.ForgeExpression;
import forge.program.ForgeExpression.Modifiable;
import forge.util.ExpressionUtil;

/**
 * A branching node of a Forge control flow graph.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class BranchStmt extends ReferenceStmt {

	private final ForgeExpression condition;
	private Set<Modifiable> refs;
	private CFGStmt thenNode, elseNode;
	private final StmtSet thenElseSet = new ThenElseSet();
	
	BranchStmt(ForgeCFG cfg, ForgeExpression condition) {
		super(cfg);
		program().checkProgram(condition);
		this.condition = condition;
		this.thenNode = cfg.exit();
		this.elseNode = cfg.exit();
	}
	
	@Returns("this.condition")
	public ForgeExpression condition() {
		return condition;
	}
	
	public Set<Modifiable> referenced() {
		if (refs == null) {
			refs = ExpressionUtil.findModifiables(condition);
		}
		return refs;
	}
	
	@Override
	public StmtSet succs() {
		return thenElseSet;
	}
	
	@Returns("this.thenNode")
	public CFGStmt getThen() {
		return thenNode;
	}

	public void setThen(CFGStmt thenStmt) {
		super.addSucc(thenStmt, nullIfSame(this.thenNode));
		this.thenNode = thenStmt;
	}

	@Returns("this.elseNode")
	public CFGStmt getElse() {
		return elseNode;
	}

	public void setElse(CFGStmt elseStmt) {
		super.addSucc(elseStmt, nullIfSame(this.elseNode));
		this.elseNode = elseStmt;
	}
	
	@Override
	public void accept(CFGVisitor visitor) {
		visitor.visit(this);
	}
	
	void appendStmt(StringBuilder sb) {
		sb.append("if ").append(condition).append(" then ").
			append(thenNode.id()).append(" else ").append(elseNode.id());
	}
	
	private CFGStmt nullIfSame(CFGStmt succ) {
		return thenNode.equals(elseNode) ? null : succ;
	}
	
	private final class ThenElseSet extends StmtSet {
		private ThenElseSet() {
			super(BranchStmt.this.cfg(),
				new AbstractSet<CFGStmt>() {
					@Override
					public int size() {
						return thenNode.equals(elseNode) ? 1 : 2;
					}
	
					@Override
					public Iterator<CFGStmt> iterator() {
						return thenNode.equals(elseNode) ?
								Iterators.<CFGStmt>singleton(thenNode) :
								Iterators.<CFGStmt>doubleton(thenNode, elseNode);
					}
				});
		}
	}
}
