package forge.cfg;

import java.util.Collections;
import java.util.Set;

import edu.mit.csail.sdg.annotations.Effects;
import edu.mit.csail.sdg.annotations.Ensures;
import edu.mit.csail.sdg.annotations.Returns;
import edu.mit.csail.sdg.annotations.SpecField;
import edu.mit.csail.sdg.annotations.Throws;
import forge.program.ForgeProgram;
import forge.program.ForgeVariable;

/**
 * A stmt in a Forge control flow graph.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
@SpecField({"succs : set CFGStmt", "preds : set CFGStmt"})
public abstract class CFGStmt implements CFGElement {
	
	/** empty set of forge variables. */
	static Set<ForgeVariable> EMPTY_VAR_SET = Collections.emptySet();
	
	/** prefix to generated ids */
	private static final String ID_PREFIX = "Stmt";
	
	/** cfg to which this stmt belongs. */
	private final ForgeCFG cfg;
	
	/** unique id for printing/debugging purposes. */
	private final String id;
	
	/** weak reference set of predecessors and unmod view. */
	private final StmtSet preds, unmodPreds;
	
	CFGStmt(ForgeCFG cfg) {
		this.cfg = cfg;
		id = ID_PREFIX + cfg.nextId();
		preds = StmtSet.weak(cfg);
		unmodPreds = preds.unmodifiable();
	}
	
	@Returns("this.cfg")
	public final ForgeCFG cfg() {
		return cfg;
	}
	
	public final ForgeProgram program() {
		return cfg.program();
	}
	
	/**
	 * Returns a unique id for this stmt.
	 */
	@Returns("this.id")
	public final String id() {
		return id;
	}

	/**
	 * Returns an unmodifiable view of the predecessor stmts. The predecessors
	 * are held by weak reference, so if they are not otherwise held in the
	 * heap, they will disappear from the set of predecessors.
	 */
	@Effects("return.elems = preds")
	public final StmtSet preds() {
		return unmodPreds;
	}
	
	/**
	 * Returns an unmodifiable view of the successor stmts.
	 */
	@Ensures("return.elems = succs")
	public abstract StmtSet succs();
	
	/**
	 * Applies a CFG visitor to this stmt.
	 */
	public abstract void accept(CFGVisitor visitor);
	
	/**
	 * Adds the given node as a successor to this stmt, and removes the old
	 * successor stmt if one is given
	 */
	@Throws({"NullPointerException : succ == null",
		     "IllegalArgumentException : succ.cfg != this.cfg"})
	final void addSucc(CFGStmt succ, CFGStmt oldSucc) {
		cfg.checkCFG(succ);
		// must do the remove first, in case they are the same
		if (oldSucc != null) oldSucc.preds.remove(this);
		succ.preds.add(this);
	}
	
	public final String toString() {
		StringBuilder sb = new StringBuilder(id()).append(": ");
		appendStmt(sb);
		return sb.toString();
	}
	
	abstract void appendStmt(StringBuilder sb);
}
