package forge.cfg;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import edu.mit.csail.sdg.annotations.Returns;
import edu.mit.csail.sdg.util.collections.ArrayHashSetUniqueList;
import edu.mit.csail.sdg.util.collections.UniqueList;
import forge.program.ForgeExpression;
import forge.program.ForgeVariable;
import forge.program.LocalDecls;
import forge.program.LocalVariable;
import forge.program.ForgeExpression.Modifiable;
import forge.util.ExpressionUtil;

/**
 * A CFG node at which a procedure is called.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class CallStmt extends UpdateStmt {

	/** called procedure. */
	private ForgeCFG called;
	
	/** list of expression in-arguments. */
	private final List<ForgeExpression> inArgs;
	
	/** ordered set of out-arguments */
	private final UniqueList<ForgeVariable> outArgs;
	
	private Set<Modifiable> refs;
	
	/**
	 * CallStmt constructor, called only by ForgeProcedure.
	 */
	CallStmt(ForgeCFG cfg, ForgeCFG called,
			List<ForgeExpression> inArgs, List<ForgeVariable> outArgs) {
		super(cfg);
		program().checkProgram(called);
		LocalDecls inParams = called.procedure().inParams();
		List<LocalVariable> ins = inParams.locals();
		List<LocalVariable> outs = called.procedure().outParams().locals();
		
		// throws an exception if any duplicates
		this.outArgs = new ArrayHashSetUniqueList<ForgeVariable>(outArgs);
		
		// check number of in-arguments equals number of in-parameters
		if (inArgs.size() != ins.size()) {
			throw new IllegalArgumentException(
					"wrong number of in-arguments");
		}
		
		// check number of out-arguments equals number of out-parameters
		if (outArgs.size() != outs.size()) {
			System.out.println("outArgs: " + outArgs);
			System.out.println("outs: " + outs);
			throw new IllegalArgumentException(
					"wrong number of out-arguments");
		}
		
		// check that the in-arguments are valid
		// must be able to store the argument in the parameter
		for (int i = 0; i < ins.size(); i++) {
			ForgeExpression arg = inArgs.get(i);
			program().checkProgram(arg);
			StoreException.checkStore(arg, ins.get(i));
		}
		
		// check that the out-arguments are valid
		// must be able to store the parameter in the argument
		for (int i = 0; i < outs.size(); i++) {
			ForgeVariable arg = outArgs.get(i);
			program().checkProgram(arg);
			StoreException.checkStore(outs.get(i), arg);
		}
		
		this.called = called;
		this.inArgs = unmodCopy(inArgs);
	}
	
	@Returns("this.called")
	public ForgeCFG called() {
		return called;
	}
	
	@Returns("this.inArgs")
	public List<ForgeExpression> inArgs() {
		return inArgs;
	}
	
	@Returns("this.outArgs")
	public List<ForgeVariable> outArgs() {
		return outArgs;
	}
	
	/* inherits specification */
	public Set<Modifiable> referenced() {
		if (refs == null) {
			refs = new HashSet<Modifiable>();
			for (ForgeExpression arg : inArgs) {
				refs.addAll(ExpressionUtil.findModifiables(arg));
			}
			refs = Collections.unmodifiableSet(refs);
		}
		return refs;
	}
	
	/* inherits specification */
	public Set<ForgeVariable> modified() {
		return outArgs.asSet();
	}
	
	@Override
	public void accept(CFGVisitor visitor) {
		visitor.visit(this);
	}

	void appendUpdate(StringBuilder sb) {
		if (!outArgs.isEmpty()) {
			Iterator<ForgeVariable> outs = outArgs.iterator();
			sb.append(outs.next());
			while (outs.hasNext()) {
				sb.append(", ").append(outs.next());
			}
			sb.append(" := ");
		}
		
		sb.append(called.procedure().name()).append("(");
		
		if (!inArgs.isEmpty()) {
			Iterator<ForgeExpression> ins = inArgs.iterator();
			sb.append(ins.next());
			while (ins.hasNext()) {
				sb.append(", ").append(ins.next());
			}
		}
		
		sb.append(")");
	}
	
	private static <E> List<E> unmodCopy(List<E> list) {
		return Collections.unmodifiableList(new ArrayList<E>(list));
	}
}