package forge.cfg;

import java.util.List;
import java.util.Set;

import edu.mit.csail.sdg.annotations.Effects;
import edu.mit.csail.sdg.annotations.Invariant;
import edu.mit.csail.sdg.annotations.Modifies;
import edu.mit.csail.sdg.annotations.Returns;
import edu.mit.csail.sdg.annotations.SpecField;
import edu.mit.csail.sdg.annotations.Throws;
import edu.mit.csail.sdg.util.collections.UnionSet;
import forge.program.ForgeExpression;
import forge.program.ForgeProcedure;
import forge.program.ForgeProgram;
import forge.program.ForgeVariable;
import forge.program.GlobalVariable;
import forge.program.InstanceDomain;
import forge.program.LocalVariable;
import forge.program.ProgramElement;
import forge.util.BreadthFirstVisitor;

/**
 * The CFG of a procedure.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
@SpecField("entry: CFGStmt")
public abstract class ForgeCFG implements ProgramElement {
	
	/**
	 * Returns a new modifiable CFG that who entry is the exit stmt.
	 */
	public static ForgeCFG.Impl implementation(ForgeProcedure proc) {
		return new Impl(proc);
	}
	
	/**
	 * Returns a new unmodifiable CFG that who entry is a specification stmt
	 * that modifies the union of the given globals and output parameters and
	 * with the given condition.
	 */
	public static ForgeCFG.Spec specification(ForgeProcedure proc,
			Set<GlobalVariable> globals, ForgeExpression condition) {
		Set<ForgeVariable> mods = UnionSet.union(globals,
				proc.outParams().locals().asSet());
		return new Spec(proc, mods, condition, false);
	}
	
	/**
	 * Returns a new unmodifiable CFG that who entry is a specification stmt
	 * with the given set of modifiables. The modifiables must meet the
	 * following constraint: outs - ins in mods in outs + globals
	 */
	public static ForgeCFG.Spec generalSpecification(ForgeProcedure proc,
			Set<? extends ForgeVariable> mods, ForgeExpression condition) {
		List<LocalVariable> ins = proc.inParams().locals();
		List<LocalVariable> outs = proc.outParams().locals();
		
		// check that (out - ins in mods)
		for (LocalVariable out: proc.outParams().locals()) {
			if (!ins.contains(out) && !mods.contains(out)) {
				throw new IllegalArgumentException("specification CFG does " +
						"not assign output parameter " + out);
			}
		}
		
		// check that (mods in outs + globals) {
		for (ForgeVariable mod: mods) {
			if (!mod.isGlobal() && !outs.contains(mod)) {
				throw new IllegalArgumentException("specification CFG " +
						"assigns non-output local variable " + mod);
			}
		}
		
		return new Spec(proc, mods, condition, false);
	}
	
	/**
	 * Hack used only for JML Forge. 
	 */
	@Deprecated
	public static ForgeCFG.Spec jmlSpecification(ForgeProcedure proc,
			Set<? extends ForgeVariable> mods, ForgeExpression condition) {
		return new Spec(proc, mods, condition, true);
	}
	
	@Invariant({"this.proc != null", "this.exit.cfg = this",
		        "this.stmtCount = #(cfg.this & CFGStmt)"})
	
	private final ForgeProcedure proc;
	private final ExitStmt exit;
	private int stmtCount = 0;
	
	@Effects("this.(ForgeCFG@proc) = proc && this.exit.cfg = this && this.stmtCount = 0")
	@Modifies("this.proc, this.exit, this.stmtCount")
	private ForgeCFG(ForgeProcedure proc) {
		this.proc = proc;
		exit = new ExitStmt(this);
	}
	
	@Effects("this.stmtCount = @old(this.stmtCount) + 1")
	@Returns("@old(this.stmtCount)")
	@Modifies("this.stmtCount")
	int nextId() {
		return stmtCount++;
	}
	
	@Returns("this.proc")
	public final ForgeProcedure procedure() {
		return proc;
	}

	@Returns("this.proc.program")
	public final ForgeProgram program() {
		return proc.program();
	}
	
	/**
	 * Returns the entry stmt of the cfg.
	 */
	@Returns("this.entry")
	public abstract CFGStmt entry();
	
	/**
	 * Returns the exit stmt of the cfg.
	 */
	@Returns("this.exit")
	public final ExitStmt exit() {
		return exit;
	}
	
	/**
	 * @throws IllegalArgumentException - if elem does not belong to this cfg
	 * @throws NullPointerException - if the given elem is null
	 */
	@Throws({"NullPointerException : element = null",
		     "IllegalArgumentException : this != element.cfg"})
	public void checkCFG(CFGElement element) {
		if (!equals(element.cfg())) {
			throw new IllegalArgumentException(element +
					" does not belong to the cfg");
		}
	}
	
	/**
	 * An implementation CFG with several statements.
	 */
	public static final class Impl extends ForgeCFG {
		
		private CFGStmt entry;
		
		private Impl(ForgeProcedure proc) {
			super(proc);
			entry = super.exit;
		}

		public CFGStmt entry() {
			return entry;
		}
		
		/**
		 * Set the entry stmt of the implementation.
		 */
		public void setEntry(CFGStmt stmt) {
			checkCFG(stmt);
			this.entry = stmt;
		}

		/**
		 * Constructs a new assign stmt with the specified variable and expr,
		 * whose successor is the exit stmt of this cfg.
		 */
		public AssignStmt newAssign(ForgeVariable var, ForgeExpression expr) {
			return new AssignStmt(this, var, expr);
		}
		
		/**
		 * Constructs a new spec stmt with the specified condition and no modified
		 * variables, whose successor is the exit stmt of this cfg.
		 */
		public SpecStmt newAssume(ForgeExpression condition) {
			return new SpecStmt(this, condition);
		}
		
		/**
		 * Constructs a new assume stmt with the specified condition and the
		 * given set of modified variables, whose successor is the exit stmt of
		 * this cfg.
		 */
		public SpecStmt newSpec(Set<? extends ForgeVariable> modified,
				ForgeExpression condition) {
			return new SpecStmt(this, modified, condition);
		}
		
		/**
		 * Constructs a new call stmt with the specified values, whose
		 * successor is the exit stmt of this cfg.
		 */
		public CallStmt newCall(ForgeCFG called, List<ForgeExpression> inArgs,
				List<ForgeVariable> outArgs) {
			return new CallStmt(this, called, inArgs, outArgs);
		}
		
		/**
		 * Constructs a new create stmt with the specified values, whose
		 * successor is the exit stmt of this cfg.
		 */
		public CreateStmt newCreate(ForgeVariable var, InstanceDomain type) {
			return new CreateStmt(this, var, type);
		}
		
		/**
		 * Constructs a new branch node with the specified condition value,
		 * whose successors are both the exit stmt of this cfg.
		 */
		public BranchStmt newBranch(ForgeExpression condition) {
			return new BranchStmt(this, condition);
		}
	}
	
	/**
	 * A specification CFG containing only a single specification statement.
	 */
	public static final class Spec extends ForgeCFG {
		
		private final SpecStmt specStmt;
		private final boolean jmlHack;
		
		private Spec(ForgeProcedure proc, Set<? extends ForgeVariable> mods,
				ForgeExpression condition, boolean jmlHack) {
			super(proc);
			specStmt = new SpecStmt(this, mods, condition);
			this.jmlHack = jmlHack;
		}

		public SpecStmt entry() {
			return specStmt;
		}
		
		public boolean jmlHack() {
			return jmlHack;
		}
	}
	
	public String toString() {
		StringBuilder sb = new StringBuilder();
		new PrintVisitor(sb).traverse(this);
		return sb.toString();
	}
	
	private static final String NEW_LINE = System.getProperty("line.separator");
	private static final String NEW_LINE_INDENT = NEW_LINE + "  ";
	
	private static final class PrintVisitor extends BreadthFirstVisitor {
		
		private final StringBuilder sb;
		
		private PrintVisitor(StringBuilder sb) {
			this.sb = sb;
		}
		
		protected void visit(ForgeCFG cfg) {
			sb.append(cfg.procedure().toString()).append(" {");
		}
		
		protected void defaultVisit(CFGStmt stmt) {
			sb.append(NEW_LINE_INDENT).append(stmt);
		}
		
		public void visit(ExitStmt stmt) {
			defaultVisit(stmt);
			sb.append(NEW_LINE).append("}").append(NEW_LINE);
		}
	}
}