package forge.cfg;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import edu.mit.csail.sdg.annotations.Returns;
import edu.mit.csail.sdg.util.collections.Bag;
import edu.mit.csail.sdg.util.collections.HashBag;
import forge.program.BinaryExpression;
import forge.program.ConditionalExpression;
import forge.program.ExpressionVisitor;
import forge.program.ForgeDomain;
import forge.program.ForgeExpression;
import forge.program.ForgeLiteral;
import forge.program.ForgeType;
import forge.program.ForgeVariable;
import forge.program.InstanceDomain;
import forge.program.OldExpression;
import forge.program.ProjectionExpression;
import forge.program.QuantifyExpression;
import forge.program.UnaryExpression;
import forge.program.ForgeDomain.Kind;
import forge.program.ForgeExpression.Modifiable;
import forge.util.ExpressionUtil;

/**
 * A specification statement. Nondeterministically chooses new values for a
 * set of modified variables such that a given formula is true.
 * Searches for values to quantified variables.
 * <p>
 * SpecStmt ::= m1,...,mn := spec(expr)
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class SpecStmt extends UpdateStmt {
	
	private final ForgeExpression condition;
	private final Set<ForgeVariable> modified;
	
	// Lazily initialized
	private Set<Modifiable> refs;
	private Map<ForgeVariable, ForgeExpression> explicitExprs;
	
	SpecStmt(ForgeCFG cfg, ForgeExpression condition) {
		super(cfg);
		program().checkProgram(condition);
		this.condition = condition;
		this.modified = EMPTY_VAR_SET;
	}
	
	SpecStmt(ForgeCFG cfg, Set<? extends ForgeVariable> modified,
			ForgeExpression condition) {
		super(cfg);
		program().checkProgram(condition);
		for (ForgeVariable v : modified) {
			program().checkProgram(v);
		}
		
		this.condition = condition;
		this.modified = Collections.unmodifiableSet(
				new LinkedHashSet<ForgeVariable>(modified));
	}
	
	@Returns("this.condition")
	public ForgeExpression condition() {
		return condition;
	}
	
	public Set<ForgeVariable> modified() {
		return modified;
	}
	
	public Set<Modifiable> referenced() {
		findVars();
		return refs;
	}
	
	/**
	 * Returns a functional expression for the given modified variable, or null
	 * if the variable is not constrained in a functional way. Throws an
	 * IllegalArgumentException if the variable is not in the modified set.
	 */
	public ForgeExpression explicitExpr(ForgeVariable var) {
		if (!modified.contains(var)) {
			throw new IllegalArgumentException(var + " not modified");
		}
		
		findVars();
		return explicitExprs.get(var);
	}
	
	private void findVars () {
		if (refs == null) {
			assert explicitExprs == null;
			explicitExprs = new HashMap<ForgeVariable, ForgeExpression>();
			SpecVarFinder mf = new SpecVarFinder();
			condition.accept(mf);
			refs = Collections.unmodifiableSet(mf.refs);
			
			for (ForgeExpression clause : ExpressionUtil.flatten(condition)) {
				if (!(clause instanceof BinaryExpression)) continue;
				BinaryExpression be = (BinaryExpression)clause;
				if (!be.op().equals(BinaryExpression.Op.EQUALS)) continue;
				ForgeExpression left = be.left(), right = be.right();
				
				ForgeVariable lvar = isExplicit(left, right, mf);
				if (lvar != null) {
					explicitExprs.put(lvar, right);
				} else {
					ForgeVariable rvar = isExplicit(right, left, mf);
					if (rvar != null) explicitExprs.put(rvar, left);
				}
			}
		}
	}
	
	private ForgeVariable isExplicit(ForgeExpression left,
			ForgeExpression right, SpecVarFinder mf) {
		if (!(left instanceof ForgeVariable)) return null;
		ForgeVariable var = (ForgeVariable)left;
		return mf.modifies(right) ? null : var;
	}

	@Override
	public void accept(CFGVisitor visitor) {
		visitor.visit(this);
	}
	
	public void appendUpdate(StringBuilder sb) {
		if (!modified.isEmpty()) {
			Iterator<ForgeVariable> mods = modified.iterator();
			sb.append(mods.next());
			while (mods.hasNext()) {
				sb.append(", ").append(mods.next());
			}
			sb.append(" := ");
		}
		
		sb.append("spec(").append(condition).append(")");
	}
	
	private final class SpecVarFinder extends ExpressionVisitor<Boolean> {
		
		private final Set<Modifiable> refs;
		private final Bag<ForgeVariable> quants;
		
		private SpecVarFinder() {
			refs = new HashSet<Modifiable>();
			quants = new HashBag<ForgeVariable>();
		}
		
		/**
		 * Returns true if the given expression contains a modified variable.
		 */
		private boolean modifies(ForgeExpression expr) {
			return super.getCache(expr);
		}
		
		@Override
		protected Boolean visit(ForgeVariable expr) {
			if (!quants.contains(expr)) {
				if (modified.contains(expr)) {
					return cacheReturn(expr, Boolean.TRUE);
				} else {
					refs.add(expr);
				}
			}
			return cacheReturn(expr, Boolean.FALSE);
		}
		
		@Override
		protected Boolean visit(OldExpression expr) {
			refs.add(expr.variable());
			return cacheReturn(expr, Boolean.FALSE);
		}
		
		@Override
		protected Boolean visit(QuantifyExpression expr) {
			quants.addAll(expr.decls().locals());
			Boolean b = expr.sub().accept(this);
			quants.removeAll(expr.decls().locals());
			return cacheReturn(expr, b);
		}

		@Override
		protected Boolean visit(BinaryExpression expr) {
			return cacheReturn(expr,
					expr.left().accept(this) |
					expr.right().accept(this));
		}
		
		@Override
		protected Boolean visit(ConditionalExpression expr) {
			return cacheReturn(expr,
					expr.condition().accept(this) |
					expr.thenExpr().accept(this) |
					expr.elseExpr().accept(this));
		}
		@Override
		protected Boolean visit(ProjectionExpression expr) {
			return cacheReturn(expr, expr.sub().accept(this));
		}
		@Override
		protected Boolean visit(UnaryExpression expr) {
			return cacheReturn(expr, expr.sub().accept(this));
		}
		@Override
		protected Boolean visit(ForgeLiteral expr) {
			return cacheReturn(expr, Boolean.FALSE);
		}
		@Override
		protected Boolean visit(ForgeType expr) {
			if (expr.isDomain()) {
				ForgeDomain dom = (ForgeDomain)expr;
				if (dom.kind().equals(Kind.INSTANCE)) {
					refs.add((InstanceDomain)dom);
				}
			}
			return cacheReturn(expr, Boolean.FALSE);
		}
		
		private Boolean cacheReturn(ForgeExpression expr, Boolean b) {
			super.putCache(expr, b);
			return b;
		}
	}
}
