package forge.cfg;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import edu.mit.csail.sdg.annotations.Returns;
import edu.mit.csail.sdg.annotations.SpecField;
import edu.mit.csail.sdg.util.collections.WeakHashSet;
import forge.program.ForgeProgram;

/**
 * A set of statements belonging to a single CFG.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
@SpecField({"cfg: ForgeCFG", "elems: set CFGStmt"})
public abstract class StmtSet implements Set<CFGStmt>, CFGElement {
	
	private static final Set<CFGStmt> NO_STMTS = Collections.emptySet();
	
	public static StmtSet hash(ForgeCFG cfg) {
		return new StmtSet(cfg, new HashSet<CFGStmt>()) {};
	}
	
	public static StmtSet hash(StmtSet s) {
		return new StmtSet(s.cfg, new HashSet<CFGStmt>(s.stmts)) {};
	}
	
	public static StmtSet empty(ForgeCFG cfg) {
		return new StmtSet(cfg, NO_STMTS) {};
	}
	
	public static StmtSet singleton(CFGStmt stmt) {
		return new StmtSet(stmt.cfg(), Collections.singleton(stmt)) {};
	}
	
	public static StmtSet weak(ForgeCFG cfg) {
		return new StmtSet(cfg, new WeakHashSet<CFGStmt>()) {};
	}
	
	public StmtSet unmodifiable() {
		return new StmtSet(cfg, Collections.unmodifiableSet(stmts)) {};
	}
	
	private final ForgeCFG cfg;
	private final Set<CFGStmt> stmts;
	
	protected StmtSet(ForgeCFG cfg, Set<CFGStmt> stmts) {
		this.cfg = cfg;
		this.stmts = stmts;
	}
	
	public ForgeProgram program() {
		return cfg.program();
	}
	
	@Returns("this.cfg")
	public ForgeCFG cfg() {
		return cfg;
	}

	/* @see java.util.Set#add(java.lang.Object) */
	public boolean add(CFGStmt e) {
		cfg.checkCFG(e);
		return stmts.add(e);
	}

	/* @see java.util.Set#addAll(java.util.Collection) */
	public boolean addAll(Collection<? extends CFGStmt> c) {
		boolean changed = false;
		for (CFGStmt s : c) changed |= stmts.add(s);
		return changed;
	}
	
	//
	// the rest just delegate
	//

	public void clear() {
		stmts.clear();
	}
	public boolean contains(Object o) {
		return stmts.contains(o);
	}
	public boolean containsAll(Collection<?> c) {
		return stmts.containsAll(c);
	}
	public boolean equals(Object o) {
		return stmts.equals(o);
	}
	public int hashCode() {
		return stmts.hashCode();
	}
	public boolean isEmpty() {
		return stmts.isEmpty();
	}
	public Iterator<CFGStmt> iterator() {
		return stmts.iterator();
	}
	public boolean remove(Object o) {
		return stmts.remove(o);
	}
	public boolean removeAll(Collection<?> c) {
		return stmts.removeAll(c);
	}
	public boolean retainAll(Collection<?> c) {
		return stmts.retainAll(c);
	}
	public int size() {
		return stmts.size();
	}
	public Object[] toArray() {
		return stmts.toArray();
	}
	public <T> T[] toArray(T[] a) {
		return stmts.toArray(a);
	}
	public String toString() {
		return stmts.toString();
	}
}
