package forge.cfg;

import java.util.AbstractSet;
import java.util.Iterator;
import java.util.Set;

import edu.mit.csail.sdg.annotations.Effects;
import edu.mit.csail.sdg.annotations.Modifies;
import edu.mit.csail.sdg.annotations.Returns;
import edu.mit.csail.sdg.annotations.SpecField;
import edu.mit.csail.sdg.annotations.Throws;
import edu.mit.csail.sdg.util.collections.Iterators;
import forge.program.ForgeExpression.Modifiable;

/**
 * A stmt in the control flow graph that may change the state of the execution
 * and which has exactly one successor.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
@SpecField({"succs : set CFGStmt | succs = this.nextNode",
            "mods : set Modifiable"})
public abstract class UpdateStmt extends ReferenceStmt {
	
	private CFGStmt nextNode;
	private final StmtSet nextSet = new NextSet();
	
	/* inherits specification */
	UpdateStmt(ForgeCFG cfg) {
		super(cfg);
		this.nextNode = cfg.exit();
	}
	
	/**
	 * Returns the set of variables and instance domains that may be modified
	 * by the statement. The "defined" modifiables, in the def-use meaning.
	 */
	@Effects("return.elems = mods")
	public abstract Set<? extends Modifiable> modified();
	
	/* inherits specification */
	public final StmtSet succs() {
		return nextSet;
	}

	@Returns("this.nextNode")
	public final CFGStmt getNext() {
		return nextNode;
	}

	@Effects("this.nextNode = nextNode")
	@Modifies("this.next")
	@Throws({"NullPointerException : nextNode == null",
		     "IllegalArgumentException : nextNode.cfg != this.cfg"})
	public final void setNext(CFGStmt nextNode) {
		super.addSucc(nextNode, this.nextNode);
		this.nextNode = nextNode;
	}
	
	final void appendStmt(StringBuilder sb) {
		appendUpdate(sb);
		sb.append(" goto ").append(getNext().id());
	}
	
	abstract void appendUpdate(StringBuilder sb);
	
	private final class NextSet extends StmtSet {
		private NextSet() {
			super(UpdateStmt.this.cfg(),
				new AbstractSet<CFGStmt>() {
					@Override
					public int size() {
						return 1;
					}
	
					@Override
					public Iterator<CFGStmt> iterator() {
						return Iterators.<CFGStmt>singleton(nextNode);
					}
				});
		}
	}
}
