package forge.dataflow;

import java.util.Map;

import forge.cfg.BranchStmt;
import forge.cfg.CFGStmt;
import forge.cfg.UpdateStmt;

/**
 * The results of a dataflow analysis.
 * 
 * @specfield stmt2Before: CFGNode -> D
 * @specfield update2After: StmtNode -> D
 * @specfield branch2After: BranchNode -> BranchData<D>
 * @specfield exitData: D
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class AnalysisResults<D> {
	
	private final Map<CFGStmt, D> stmt2Before;
	private final Map<UpdateStmt, D> update2After;
	private final Map<BranchStmt, BranchData<D>> branch2After;
	private final D exitData;
	
	/**
	 * @effect this.stmt2Before' = stmt2Before && this.exitData' = exitData
	 */
	AnalysisResults(Map<CFGStmt, D> stmt2Before, Map<UpdateStmt, D> update2After,
			Map<BranchStmt, BranchData<D>> branch2After, D exitData) {
		this.stmt2Before  = stmt2Before;
		this.update2After = update2After;
		this.branch2After = branch2After;
		this.exitData = exitData;
	}

	/**
	 * Returns the data at the program point in the CFG immediately preceding
	 * the specified stmt.
	 * 
	 * @return this.stmt2Before[node]
	 */
	public D dataBefore(CFGStmt stmt) {
		return stmt2Before.get(stmt);
	}
	
	/**
	 * Returns the data at the program point in the CFG immediately following
	 * the specified update stmt.
	 * 
	 * @return this.update2After[node.next]
	 */
	public D dataAfter(UpdateStmt update) {
		return update2After.get(update);
	}
	
	/**
	 * Returns the data at the program points in the CFG that immediately
	 * follow the specified branch stmt.
	 * 
	 * @return [x : BranchData |
	 *           x.thenData = this.stmt2Before[branch.then] &&
	 *           x.elseData = this.stmt2Before[branch.else] ]
	 */
	public BranchData<D> dataAfter(BranchStmt branch) {
		BranchData<D> data = branch2After.get(branch);
		if (data == null) {
			data = new BranchData<D>(null);
			branch2After.put(branch, data);
		}
		return data;
	}
	
	/**
	 * Returns the data at the program point in the CFG immediately following
	 * the exit node.
	 * 
	 * @return this.exitData
	 */
	public D exitData() {
		return exitData;
	}
	
}
