package forge.dataflow;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;

import forge.cfg.AssignStmt;
import forge.cfg.BranchStmt;
import forge.cfg.CFGStmt;
import forge.cfg.CFGVisitor;
import forge.cfg.CallStmt;
import forge.cfg.CreateStmt;
import forge.cfg.ExitStmt;
import forge.cfg.ForgeCFG;
import forge.cfg.SpecStmt;
import forge.cfg.UpdateStmt;

/**
 * A template for a dataflow analysis.
 * <p>
 * Subclasses need only indicate the direction of the analysis as an argument
 * to the constructor and implement the initial, merge, and transfer functions,
 * though they must be implemented carefully.
 * <p>
 * An analysis may be <em>forward</em> or <em>backward</em>. A forward analysis
 * proceeds forward from the entry stmt in the CFG, and a backwards analysis
 * proceeds backwards through the CFG from the exit stmt. When implementing a
 * backwards analysis, the transfer function for a branch node must always
 * return branch data whose thenData and elseData are the exact same reference;
 * this can be ensured by using the BranchData constructor that takes a single
 * data argument.
 * <p>
 * <strong>The transfer functions must not modify the input data.</strong>.
 * That is, if the data has changed, a new data instance must be returned. The
 * transfer functions may return null (bottom) if the analysis deems the
 * subsequent stmt to be unreachable. For performance reasons, subclasses are
 * encouraged, but not required, to ensure their transfer functions return the
 * input data reference if the data has not changed. This avoids calls to the
 * merge method, which can be costly.
 * <p>
 * The data before or after a given stmt in the results may be mapped to null,
 * while is used as a bottom value to indicate that the analysis deems the stmt
 * to be unreachable.
 * 
 * @specfield backward : boolean
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public abstract class DataflowAnalysis<D> {
	
	private final boolean backward;
	
	/**
	 * Constructs a new dataflow analysis in the specified direction.
	 *
	 * @effect this.backward' = backward
	 */
	protected DataflowAnalysis(boolean backward) {
		this.backward = backward;
	}
	
	/**
	 * Returns true if this is a backwards analysis.
	 * 
	 * @return this.backward
	 */
	public final boolean isBackward() {
		return backward;
	}
	
	/**
	 * Performs the analysis and returns a mapping from each reachable node to
	 * the data at the program point immediately preceding the node.
	 */	
	public final AnalysisResults<D> analyze(ForgeCFG cfg) {
		Analyzer analyzer = backward ? new BackwardAnalyzer() :
				new ForwardAnalyzer();
		analyzer.run(cfg);
		return new AnalysisResults<D>(analyzer.stmt2Before,
				analyzer.update2After, analyzer.branch2After, analyzer.exitData);
	}
	
	/**
	 * Returns the initial data set. If this is a forward analysis, it is the
	 * data immediately preceding the CFG's entry node; in a backward analysis
	 * it is the exit data immediate following the terminal node.
	 */
	public abstract D initial();

	/**
	 * Transfer function at an assign stmt.
	 * This implementation returns the input data.
	 */
	public D transfer(AssignStmt stmt, D input) {
		return input;
	}
	
	/**
	 * Transfer function at a call stmt.
	 * This implementation returns the input data.
	 */
	public D transfer(CallStmt stmt, D input) {
		return input;
	}
	
	/**
	 * Transfer function at a create stmt.
	 * This implementation returns the input data.
	 */
	public D transfer(CreateStmt stmt, D input) {
		return input;
	}
	
	/**
	 * Transfer function at an spec stmt.
	 * This implementation returns the input data.
	 */
	public D transfer(SpecStmt stmt, D input) {
		return input;
	}
	
	/**
	 * Transfer function at a branch node.
	 * If this is a backward analysis, the then and else data must be the same;
	 * they must exhibit reference equality (==) with one another; otherwise,
	 * the analyze method may throw an IllegalStateException.
	 * This implementation returns new BranchData(input).
	 */
	public BranchData<D> transfer(BranchStmt stmt, D input) {
		return new BranchData<D>(input);
	}
	
	/**
	 * Transfer function at a terminal node.
	 * This implementation returns the input data.
	 */
	public D transfer(ExitStmt stmt, D input) {
		return input;
	}
	
	/**
	 * Merges "from" data into the "into" data and return true iff the into
	 * data has changed.
	 * 
	 * @pre from != null && into != null && from != into
	 */
	public abstract boolean merge(D from, D into);
	
	/**
	 * Makes a copy of the specified data.
	 */
	public abstract D copy(D data);
	
	/**
	 * A generic forward or backwards worklist algorithm.
	 */
	private abstract class Analyzer implements CFGVisitor {
		
		final Queue<CFGStmt> worklist = new LinkedList<CFGStmt>();
		final Map<CFGStmt, D> stmt2Before = new HashMap<CFGStmt, D>();
		final Map<UpdateStmt, D> update2After = new HashMap<UpdateStmt, D>();
		final Map<BranchStmt, BranchData<D>> branch2After =
			new HashMap<BranchStmt, BranchData<D>>();
		D exitData = null;
		
		private void run(ForgeCFG cfg) {
			initialize(cfg);
			while(!worklist.isEmpty()) {
				worklist.remove().accept(this);
			}
		}
		
		abstract void initialize(ForgeCFG cfg);
		
		abstract D stmtInput(UpdateStmt update);
		
		abstract D branchInput(BranchStmt branch);
		
		abstract D exitInput(ExitStmt exit);

		abstract void stmtOutput(UpdateStmt update, D output);
		
		abstract void branchOutput(BranchStmt branch, BranchData<D> output);
		
		abstract void exitOutput(ExitStmt exit, D output);
		
		public final void visit(AssignStmt stmt) {
			D output = transfer(stmt, stmtInput(stmt));
			stmtOutput(stmt, output);
		}
		
		public final void visit(CallStmt stmt) {
			D output = transfer(stmt, stmtInput(stmt));
			stmtOutput(stmt, output);
		}
		
		public final void visit(CreateStmt stmt) {
			D output = transfer(stmt, stmtInput(stmt));
			stmtOutput(stmt, output);
		}

		public final void visit(BranchStmt stmt) {
			BranchData<D> output = transfer(stmt, branchInput(stmt));
			branchOutput(stmt, output);
		}
		
		public final void visit(SpecStmt stmt) {
			D output = transfer(stmt, stmtInput(stmt));
			stmtOutput(stmt, output);
		}

		public final void visit(ExitStmt stmt) {
			D output = transfer(stmt, exitInput(stmt));
			exitOutput(stmt, output);
		}
		
		boolean mergeOutput(D output, D currData) {
			assert output != null && currData != null;
			return (currData == output) ? false : merge(output, currData);
		}
		
	}
	
	private final class ForwardAnalyzer extends Analyzer {
		
		@Override
		void initialize(ForgeCFG cfg) {
			stmt2Before.put(cfg.entry(), initial());
			worklist.offer(cfg.entry());
		}
		
		@Override
		D stmtInput(UpdateStmt update) {
			return stmt2Before.get(update);
		}

		@Override
		D branchInput(BranchStmt branch) {
			return stmt2Before.get(branch);
		}
		
		@Override
		D exitInput(ExitStmt exit) {
			return stmt2Before.get(exit);
		}
		
		@Override
		void stmtOutput(UpdateStmt update, D output) {
			update2After.put(update, output);
			ensureSucc(update.getNext(), output);
		}

		@Override
		void branchOutput(BranchStmt branch, BranchData<D> output) {
			branch2After.put(branch, output);
			ensureSucc(branch.getThen(), output.thenData());
			ensureSucc(branch.getElse(), output.elseData());
		}

		@Override
		void exitOutput(ExitStmt exit, D output) {
			exitData = output;
		}
		
		private void ensureSucc(CFGStmt succ, D output) {
			if (output == null) return;
			D currData = stmt2Before.get(succ);
			
			final boolean changed;			
			if (currData == null) {
				stmt2Before.put(succ, copy(output));
				changed = true;
			} else {
				changed = mergeOutput(output, currData);
			}
			
			if (changed) worklist.offer(succ);
		}
		
	}
	
	private final class BackwardAnalyzer extends Analyzer {
		
		void initialize(ForgeCFG cfg) {
			exitData = initial();
			worklist.offer(cfg.exit());
		}
		
		@Override
		D stmtInput(UpdateStmt update) {
			return update2After.get(update);
		}

		@Override
		D branchInput(BranchStmt branch) {
			return checkSame(branch2After.get(branch));
		}
		
		@Override
		D exitInput(ExitStmt exit) {
			return exitData;
		}
		
		@Override
		void stmtOutput(UpdateStmt update, D output) {
			ensurePreds(update, output);
		}

		@Override
		void branchOutput(BranchStmt branch, BranchData<D> output) {
			ensurePreds(branch, checkSame(output));
		}

		@Override
		void exitOutput(ExitStmt exit, D output) {
			ensurePreds(exit, output);
		}
		
		private D checkSame(BranchData<D> branchData) {
			if (branchData.thenData() != branchData.elseData()) {
				throw new IllegalStateException("thenData != elseData");
			}
			return branchData.thenData();
		}
		
		private void ensurePreds(CFGStmt stmt, D output) {
			if (output == null) return;
			stmt2Before.put(stmt, output);
			
			for (CFGStmt p : stmt.preds()) {
				final boolean changed;
				
				if (p instanceof UpdateStmt) {
					UpdateStmt update = (UpdateStmt)p;
					D stmtAfter = update2After.get(update);
					if (stmtAfter == null) {
						update2After.put(update, copy(output));
						changed = true;
					} else {
						changed = mergeOutput(output, stmtAfter);
					}					
				} else {
					BranchStmt branch = (BranchStmt)p;
					BranchData<D> branchAfter = branch2After.get(branch);
					if (branchAfter == null) {
						branch2After.put(branch, new BranchData<D>(copy(output)));
						changed = true;
					} else {
						changed = mergeOutput(output, checkSame(branchAfter));
					}					
				}
				
				if (changed) worklist.offer(p);
			}
		}
		
	}
	
}
