package forge.dataflow;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import forge.cfg.AssignStmt;
import forge.cfg.BranchStmt;
import forge.cfg.CFGStmt;
import forge.cfg.CallStmt;
import forge.cfg.CreateStmt;
import forge.cfg.ExitStmt;
import forge.cfg.SpecStmt;

/**
 * Computes the set of nodes reachable from each node in the CFG
 * (not including the node itself unless the node is part of a cycle).
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class ReachabilityAnalysis extends DataflowAnalysis<Set<CFGStmt>> {
	
	/** singleton. */
	public static final ReachabilityAnalysis ANALYSIS = new ReachabilityAnalysis();
	
	private static final Set<CFGStmt> EMPTY_NODE_SET = Collections.emptySet();
	
	private ReachabilityAnalysis() {
		super(true);
	}

	@Override
	public Set<CFGStmt> initial() {
		return EMPTY_NODE_SET;
	}

	@Override
	public Set<CFGStmt> transfer(AssignStmt stmt, Set<CFGStmt> input) {
		return addSet(stmt, input);
	}

	@Override
	public BranchData<Set<CFGStmt>> transfer(BranchStmt stmt, Set<CFGStmt> input) {
		return new BranchData<Set<CFGStmt>>(addSet(stmt, input));
	}

	@Override
	public Set<CFGStmt> transfer(CallStmt stmt, Set<CFGStmt> input) {
		return addSet(stmt, input);
	}

	@Override
	public Set<CFGStmt> transfer(CreateStmt stmt, Set<CFGStmt> input) {
		return addSet(stmt, input);
	}
	
	@Override
	public Set<CFGStmt> transfer(SpecStmt stmt, Set<CFGStmt> input) {
		return addSet(stmt, input);
	}
	
	@Override
	public Set<CFGStmt> transfer(ExitStmt stmt, Set<CFGStmt> input) {
		return addSet(stmt, input);
	}
	
	@Override
	public boolean merge(Set<CFGStmt> from, Set<CFGStmt> into) {
		return into.addAll(from); // union
	}
	
	@Override
	public Set<CFGStmt> copy(Set<CFGStmt> data) {
		return new HashSet<CFGStmt>(data);
	}

	private Set<CFGStmt> addSet(CFGStmt stmt, Set<CFGStmt> input) {
		Set<CFGStmt> reach = new HashSet<CFGStmt>(input);
		reach.add(stmt);
		return reach;
	}

}