package forge.program;

import forge.program.QuantifyExpression.Op;


/**
 * A Forge expression.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
abstract class AbstractExpression implements ForgeExpression {
	
	private final ForgeProgram program;
	
	/**
	 * Constructor for everything exception ForgeType
	 */
	AbstractExpression(ForgeProgram program) {
		this.program = program;
	}
	
	public final ForgeProgram program() {
		return program;
	}
	
	/**
	 * Returns the type of the expression
	 */
	public abstract ForgeType type();
	
	/**
	 * Returns type().arity()
	 */
	public int arity() {
		return type().arity();
	}
	
	public final boolean isUnary() {
		return arity() == 1;
	}
	
	//
	// projection expressions
	//
	public ForgeExpression domain() {
		return projection(0);
	}
	public ForgeExpression range() {
		return projection(arity() - 1);
	}
	public final ForgeExpression iden() {
		IllegalArityException.checkUnary(arity());
		return new ProjectionExpression(this, 0, 0);
	}
	public final ForgeExpression projection(int... columns) {
		return new ProjectionExpression(this, columns);
	}

	//
	// unary op expressions
	// if overridden, must call op.checkArity
	//
	public final ForgeExpression closure() {
		return apply(UnaryExpression.Op.CLOSURE);
	}
	public final ForgeExpression no() {
		return apply(UnaryExpression.Op.NO);
	}
	public final ForgeExpression lone() {
		return apply(UnaryExpression.Op.LONE);
	}
	public final ForgeExpression one() {
		return apply(UnaryExpression.Op.ONE);
	}
	public final ForgeExpression some() {
		return apply(UnaryExpression.Op.SOME);
	}
	public final ForgeExpression size() {
		return apply(UnaryExpression.Op.CARDINALITY);
	}
	public final ForgeExpression bool() {
		return apply(UnaryExpression.Op.BOOL);
	}
	public final ForgeExpression not() {
		return apply(UnaryExpression.Op.NOT);
	}
	public final ForgeExpression sum() {
		return apply(UnaryExpression.Op.SUM);
	}
	public final ForgeExpression neg() {
		return apply(UnaryExpression.Op.NEG);
	}
	public final ForgeExpression bitNot() {
		return apply(UnaryExpression.Op.BIT_NOT);
	}
	public ForgeExpression apply(UnaryExpression.Op op) {
		return new UnaryExpression(op, this);
	}
	
	//
	// binary op expressions
	// if overloaded, must call op.checkArities
	//
	
	public final ForgeExpression union(ForgeExpression expr) {
		if (equals(program.empty())) {
			BinaryExpression.check(this, BinaryExpression.Op.UNION, expr);
			return expr;
		}
		if (expr.equals(program.empty())) {
			BinaryExpression.check(this, BinaryExpression.Op.UNION, expr);
			return this;
		}
		if (this instanceof ForgeType && expr instanceof ForgeType) {
			return ((ForgeType)this).union((ForgeType)expr);
		}
		return compose(BinaryExpression.Op.UNION, expr);
	}
	
	public final ForgeExpression difference(ForgeExpression expr) {
		if (equals(program.empty())) {
			BinaryExpression.check(this, BinaryExpression.Op.DIFFERENCE, expr);
			return program.empty();
		}
		if (expr.equals(program.empty())) {
			BinaryExpression.check(this, BinaryExpression.Op.DIFFERENCE, expr);
			return this;
		}
		if (this instanceof ForgeType && expr instanceof ForgeType) {
			return ((ForgeType)this).difference((ForgeType)expr);
		}
		return compose(BinaryExpression.Op.DIFFERENCE, expr);
	}
	
	public final ForgeExpression intersection(ForgeExpression expr) {
		if (equals(program.empty()) || expr.equals(program.empty())) {
			BinaryExpression.check(this, BinaryExpression.Op.INTERSECTION, expr);
			return program.empty();
		}
		if (this instanceof ForgeType && expr instanceof ForgeType) {
			return ((ForgeType)this).intersection((ForgeType)expr);
		}
		return compose(BinaryExpression.Op.INTERSECTION, expr);
	}
	
	public final ForgeExpression join(ForgeExpression expr) {
		if (equals(program.empty()) || expr.equals(program.empty())) {
			BinaryExpression.check(this, BinaryExpression.Op.PRODUCT, expr);
			return program.empty(arity() + expr.arity() - 2);
		}
		if (this instanceof ForgeType && expr instanceof ForgeType) {
			return ((ForgeType)this).join((ForgeType)expr);
		}
		return compose(BinaryExpression.Op.JOIN, expr);
	}
	
	public final ForgeExpression product(ForgeExpression expr) {
		if (equals(program.empty()) || expr.equals(program.empty())) {
			BinaryExpression.check(this, BinaryExpression.Op.PRODUCT, expr);
			return program.empty(arity() + expr.arity());
		}
		if (this instanceof ForgeType && expr instanceof ForgeType) {
			return ((ForgeType)this).product((ForgeType)expr);
		}
		return compose(BinaryExpression.Op.PRODUCT, expr);
	}
	
	public final ForgeExpression override(ForgeExpression expr) {
		return compose(BinaryExpression.Op.OVERRIDE, expr);
	}
	
	public final ForgeExpression and(ForgeExpression expr) {
		if (equals(program.trueLiteral()) || expr.equals(program.falseLiteral())) {
			BinaryExpression.check(this, BinaryExpression.Op.AND, expr);
			return expr;
		}
		if (equals(program.falseLiteral()) || expr.equals(program.trueLiteral())) {
			BinaryExpression.check(this, BinaryExpression.Op.AND, expr);
			return this;
		}
		return compose(BinaryExpression.Op.AND, expr);
	}
	
	public final ForgeExpression or(ForgeExpression expr) {
		if (equals(program.trueLiteral()) || expr.equals(program.falseLiteral())) {
			BinaryExpression.check(this, BinaryExpression.Op.OR, expr);
			return this;
		}
		if (equals(program.falseLiteral()) || expr.equals(program.trueLiteral())) {
			BinaryExpression.check(this, BinaryExpression.Op.OR, expr);
			return expr;
		}
		return compose(BinaryExpression.Op.OR, expr);
	}
	
	public final ForgeExpression xor(ForgeExpression expr) {
		return compose(BinaryExpression.Op.XOR, expr);
	}
	
	public final ForgeExpression implies(ForgeExpression expr) {
		if (equals(program.falseLiteral()) || expr.equals(program.falseLiteral())) {
			BinaryExpression.check(this, BinaryExpression.Op.IMPLIES, expr);
			return not();
		}
		if (equals(program.trueLiteral()) || expr.equals(program.trueLiteral())) {
			BinaryExpression.check(this, BinaryExpression.Op.IMPLIES, expr);
			return expr;
		}
		return compose(BinaryExpression.Op.IMPLIES, expr);
	}
	
	public final ForgeExpression iff(ForgeExpression expr) {
		if (equals(program.trueLiteral())) {
			BinaryExpression.check(this, BinaryExpression.Op.IFF, expr);
			return expr;
		}
		if (equals(program.falseLiteral())) {
			BinaryExpression.check(this, BinaryExpression.Op.IFF, expr);
			return expr.not();
		}
		if (expr.equals(program.trueLiteral())) {
			BinaryExpression.check(this, BinaryExpression.Op.IFF, expr);
			return this;
		}
		if (expr.equals(program.falseLiteral())) {
			BinaryExpression.check(this, BinaryExpression.Op.IFF, expr);
			return not();
		}
		return compose(BinaryExpression.Op.IFF, expr);
	}
	
	public final ForgeExpression eq(ForgeExpression expr) {
		return compose(BinaryExpression.Op.EQUALS, expr);
	}
	public final ForgeExpression in(ForgeExpression expr) {
		return compose(BinaryExpression.Op.SUBSET, expr);
	}
	public final ForgeExpression lt(ForgeExpression expr) {
		return compose(BinaryExpression.Op.LT, expr);
	}
	public final ForgeExpression gt(ForgeExpression expr) {
		return compose(BinaryExpression.Op.GT, expr);
	}
	public final ForgeExpression lte(ForgeExpression expr) {
		return compose(BinaryExpression.Op.LTE, expr);
	}
	public final ForgeExpression gte(ForgeExpression expr) {
		return compose(BinaryExpression.Op.GTE, expr);
	}
	public final ForgeExpression plus(ForgeExpression expr) {
		return compose(BinaryExpression.Op.PLUS, expr);
	}
	public final ForgeExpression minus(ForgeExpression expr) {
		return compose(BinaryExpression.Op.MINUS, expr);
	}
	public final ForgeExpression times(ForgeExpression expr) {
		return compose(BinaryExpression.Op.TIMES, expr);
	}
	public final ForgeExpression divide(ForgeExpression expr) {
		return compose(BinaryExpression.Op.DIVIDE, expr);
	}
	public final ForgeExpression modulo(ForgeExpression expr) {
		return compose(BinaryExpression.Op.MODULO, expr);
	}
	public final ForgeExpression bitAnd(ForgeExpression expr) {
		return compose(BinaryExpression.Op.BIT_AND, expr);
	}
	public final ForgeExpression bitOr(ForgeExpression expr) {
		return compose(BinaryExpression.Op.BIT_OR, expr);
	}
	public final ForgeExpression bitXor(ForgeExpression expr) {
		return compose(BinaryExpression.Op.BIT_XOR, expr);
	}
	public final ForgeExpression shiftLeft(ForgeExpression expr) {
		return compose(BinaryExpression.Op.SHL, expr);
	}
	public final ForgeExpression shiftRight(ForgeExpression expr) {
		return compose(BinaryExpression.Op.SHR, expr);
	}
	public final ForgeExpression unsignedShiftRight(ForgeExpression expr) {
		return compose(BinaryExpression.Op.USHR, expr);
	}
	public final ForgeExpression compose(BinaryExpression.Op op, ForgeExpression expr) {
		return new BinaryExpression(this, op, expr);
	}
	
	// conditional expression
	public final ForgeExpression thenElse(ForgeExpression thenExpr, ForgeExpression elseExpr) {
		if (equals(program.trueLiteral())) {
			ConditionalExpression.check(this, thenExpr, elseExpr);
			return thenExpr;
		}
		if (equals(program.falseLiteral())) {
			ConditionalExpression.check(this, thenExpr, elseExpr);
			return elseExpr;
		}
		return new ConditionalExpression(this, thenExpr, elseExpr);
	}

	// quantify expressions
	public final ForgeExpression forAll(LocalDecls decls) {
		if (equals(program.trueLiteral())) {
			return this;
		}
		return quantify(QuantifyExpression.Op.ALL, decls);
	}
	public final ForgeExpression forSome(LocalDecls decls) {
		return quantify(QuantifyExpression.Op.SOME, decls);
	}
	public final ForgeExpression comprehension(LocalDecls decls) {
		return quantify(QuantifyExpression.Op.UNION, decls);
	}
	public final ForgeExpression summation(LocalDecls decls) {
		return quantify(QuantifyExpression.Op.SUM, decls);
	}
	public final ForgeExpression quantify(Op quant, LocalDecls decls) {
		return new QuantifyExpression(quant, this, decls);
	}
	
}
