package forge.program;

import edu.mit.csail.sdg.annotations.Returns;

/**
 * An expression composed of a binary operator and two subexpressions.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class BinaryExpression extends AbstractExpression {

	private final Op op;
	private final ForgeExpression left, right;
	private final ForgeType type;
	
	/**
	 * Constructs a new expression with a binary operator.
	 * 
	 * @throws IllegalArgumentException - left.program != right.program
	 */
	BinaryExpression(ForgeExpression left, Op op, ForgeExpression right) {
		super(left.program());
		check(left, op, right);
		this.op = op;
		this.left = left;
		this.right = right;
		type = op.inferType(left.type(), right.type());
	}
	
	static void check(ForgeExpression left, Op op, ForgeExpression right) {
		left.program().checkProgram(right);
		op.checkArities(left.arity(), right.arity());
	}
	
	@Returns("this.op")
	public Op op() {
		return op;
	}
	
	@Returns("this.left")
	public ForgeExpression left() {
		return left;
	}
	
	@Returns("this.right")
	public ForgeExpression right() {
		return right;
	}
	
	@Returns("this.type")
	public ForgeType type() {
		return type;
	}

	public <T> T accept(ExpressionVisitor<T> visitor) {
		return visitor.fullVisit(this);
	}
	
	public String toString() {
		return "(" + left + " " + op + " " + right + ")";
	}
	
	public static enum Op {
		
		/** relational union */
		UNION("+") {
			void checkArities(int left, int right) {
				IllegalArityException.checkMatch(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.union(right);
			}
		},
		
		/** relational difference */
		DIFFERENCE("-") {
			void checkArities(int left, int right) {
				IllegalArityException.checkMatch(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left;
			}
		},
		
		/** relational intersection */
		INTERSECTION("&") {
			void checkArities(int left, int right) {
				IllegalArityException.checkMatch(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.intersection(right);
			}
		},
		
		/** relational join */
		JOIN(".") {
			void checkArities(int left, int right) {
				IllegalArityException.checkJoin(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.join(right);
			}
		},
		
		/** relational product */
		PRODUCT("->") {
			void checkArities(int left, int right) {}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.product(right);
			}
		},
		
		/** relational override */
		OVERRIDE("++") {
			void checkArities(int left, int right) {
				IllegalArityException.checkMatch(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.union(right);
			}
		},
		
		
		/** relational equality */
		EQUALS("=") {
			void checkArities(int left, int right) {
				IllegalArityException.checkMatch(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().booleanDomain();
			}
		},
		
		/** relational subset */
		SUBSET("in") {
			void checkArities(int left, int right) {
				IllegalArityException.checkMatch(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().booleanDomain();
			}
		},
		
		/** boolean and */
		AND("&&") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().booleanDomain();
			}
		},
		
		/** boolean or */
		OR("||") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().booleanDomain();
			}
		},
		
		/** boolean or */
		XOR("^^") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().booleanDomain();
			}
		},
		
		/** boolean implies */
		IMPLIES("=>") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().booleanDomain();
			}
		},
		
		/** boolean if and only if */
		IFF("<=>") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().booleanDomain();
			}
		},
		
		/** integer less than */
		LT("<") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().booleanDomain();
			}
		},
		
		/** integer greater than */
		GT(">") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().booleanDomain();
			}
		},
		
		/** integer less than or equal */
		LTE("=<") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().booleanDomain();
			}
		},
		
		/** integer greater than or equal */
		GTE(">=") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().booleanDomain();
			}
		},
		
		/** integer addition */
		PLUS("plus") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().integerDomain();
			}
		},
		
		/** integer subtraction */
		MINUS("minus") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().integerDomain();
			}
		},
		
		/** integer multiplication */
		TIMES("times") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().integerDomain();
			}
		},
		
		/** integer division */
		DIVIDE("divide") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().integerDomain();
			}
		},
		
		/** integer modulo division */
		MODULO("modulo") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().integerDomain();
			}
		},
		
		/** bitwise and */
		BIT_AND("band") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().integerDomain();
			}
		},
		
		/** boolean xor */
		BIT_OR("bor") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().integerDomain();
			}
		},
		
		/** boolean xor */
		BIT_XOR("bxor") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().integerDomain();
			}
		},
		
		/** integer bit shift left */
		SHL("<<") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().integerDomain();
			}
		},
		
		/** integer bit shift right */
		SHR(">>") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().integerDomain();
			}
		},
		
		/** integer unsigned bit shift right */
		USHR(">>>") {
			void checkArities(int left, int right) {
				IllegalArityException.checkUnary(left, right);
			}
			ForgeType inferType(ForgeType left, ForgeType right) {
				return left.program().integerDomain();
			}
		};
		
		private final String s;
		Op(String s) { this.s = s; }
		public String toString() { return s; }
		
		abstract void checkArities(int left, int right);
		abstract ForgeType inferType(ForgeType left, ForgeType right);
	}
	
}
