package forge.program;

import edu.mit.csail.sdg.annotations.Returns;

/**
 * A conditional "if" expression
 * 
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class ConditionalExpression extends AbstractExpression {
	
	private final ForgeExpression cond, thenExpr, elseExpr;
	private final ForgeType type;
	
	/**
	 * Constructs a conditional expression.
	 * 
	 * @throws IllegalArgumentException -
	 *   !(condition.program = thenExpr.program = elseExpr.program)
	 */
	ConditionalExpression(ForgeExpression cond, ForgeExpression thenExpr,
			ForgeExpression elseExpr) {
		super(cond.program());
		this.cond = cond;
		this.thenExpr = thenExpr;
		this.elseExpr = elseExpr;
		type = thenExpr.type().union(elseExpr.type());
	}
	
	static void check(ForgeExpression cond, ForgeExpression thenExpr, ForgeExpression elseExpr) {
		IllegalArityException.checkUnary(cond.arity());
		IllegalArityException.checkMatch(thenExpr.arity(), elseExpr.arity());
		cond.program().checkProgram(thenExpr);
		cond.program().checkProgram(elseExpr);
	}
	
	@Returns("this.condition")
	public ForgeExpression condition() {
		return cond;
	}
	
	@Returns("this.thenExpr")
	public ForgeExpression thenExpr() {
		return thenExpr;
	}
	
	@Returns("this.elseExpr")
	public ForgeExpression elseExpr() {
		return elseExpr;
	}

	@Override
	/* inherits specification */
	public ForgeType type() {
		return type;
	}

	/* inherits specification */
	public <T> T accept(ExpressionVisitor<T> visitor) {
		return visitor.fullVisit(this);
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append(cond.toString());
		sb.append(" ? ");
		sb.append(thenExpr.toString());
		sb.append(" : ");
		sb.append(elseExpr.toString());
		return sb.toString();
	}

}
