package forge.program;

import java.util.HashMap;
import java.util.Map;

/**
 * A visitor to a Forge expression.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public abstract class ExpressionVisitor<T> {
	
	/** null if we're not repeating exprs. */
	private final Map<ForgeExpression, T> expr2Result;
	
	/**
	 * @effect no this.cache'
	 */
	protected ExpressionVisitor() {
		expr2Result = new HashMap<ForgeExpression, T>();
	}
	
	/**
	 * @effect this.cache' = copy(visitor.cache)
	 */
	protected ExpressionVisitor(ExpressionVisitor<T> visitor) {
		expr2Result = new HashMap<ForgeExpression, T>(visitor.expr2Result);
	}
	
	/**
	 * @return some this.cache[expr]
	 */
	protected final boolean isCached(ForgeExpression expr) {
		return expr2Result.containsKey(expr);
	}
	
	/**
	 * @effect this.cache' = this.cache ++ expr -> result
	 */
	protected final void putCache(ForgeExpression expr, T result) {
		expr2Result.put(expr, result);
	}
	
	/**
	 * @return this.cache[expr]
	 */
	protected final T getCache(ForgeExpression expr) {
		return expr2Result.get(expr);
	}
	
	/**
	 * @effect isCached(expr) => this.cache' = this.cache &&
	 *        !isCached(expr) => visit(expr);
	 * @return isCached(expr) ? getCache(expr) : visit(expr);
	 */
	final T fullVisit(ForgeType expr) {
		return isCached(expr) ? getCache(expr) : visit(expr);
	}
	
	/**
	 * @effect isCached(expr) => this.cache' = this.cache &&
	 *        !isCached(expr) => visit(expr);
	 * @return isCached(expr) ? getCache(expr) : visit(expr);
	 */
	final T fullVisit(ForgeLiteral expr) {
		return isCached(expr) ? getCache(expr) : visit(expr);
	}
	
	/**
	 * @effect isCached(expr) => this.cache' = this.cache &&
	 *        !isCached(expr) => visit(expr);
	 * @return isCached(expr) ? getCache(expr) : visit(expr);
	 */
	final T fullVisit(ForgeVariable expr) {
		return isCached(expr) ? getCache(expr) : visit(expr);
	}
	
	/**
	 * @effect isCached(expr) => this.cache' = this.cache &&
	 *        !isCached(expr) => visit(expr);
	 * @return isCached(expr) ? getCache(expr) : visit(expr);
	 */
	final T fullVisit(UnaryExpression expr) {
		return isCached(expr) ? getCache(expr) : visit(expr);
	}
	
	/**
	 * @effect isCached(expr) => this.cache' = this.cache &&
	 *        !isCached(expr) => visit(expr);
	 * @return isCached(expr) ? getCache(expr) : visit(expr);
	 */
	final T fullVisit(BinaryExpression expr) {
		return isCached(expr) ? getCache(expr) : visit(expr);
	}
	
	/**
	 * @effect isCached(expr) => this.cache' = this.cache &&
	 *        !isCached(expr) => visit(expr);
	 * @return isCached(expr) ? getCache(expr) : visit(expr);
	 */
	final T fullVisit(ConditionalExpression expr) {
		return isCached(expr) ? getCache(expr) : visit(expr);
	}
	
	/**
	 * @effect isCached(expr) => this.cache' = this.cache &&
	 *        !isCached(expr) => visit(expr);
	 * @return isCached(expr) ? getCache(expr) : visit(expr);
	 */
	final T fullVisit(ProjectionExpression expr) {
		return isCached(expr) ? getCache(expr) : visit(expr);
	}
	
	/**
	 * @effect isCached(expr) => this.cache' = this.cache &&
	 *        !isCached(expr) => visit(expr);
	 * @return isCached(expr) ? getCache(expr) : visit(expr);
	 */
	final T fullVisit(QuantifyExpression expr) {
		return isCached(expr) ? getCache(expr) : visit(expr);
	}
	
	/**
	 * @effect isCached(expr) => this.cache' = this.cache &&
	 *        !isCached(expr) => visit(expr);
	 * @return isCached(expr) ? getCache(expr) : visit(expr);
	 */
	final T fullVisit(OldExpression expr) {
		return isCached(expr) ? getCache(expr) : visit(expr);
	}
	
	/**
	 * Visits a ForgeType.
	 */
	protected abstract T visit(ForgeType expr);
	
	/**
	 * Visits a ForgeConstant.
	 */
	protected abstract T visit(ForgeLiteral expr);
	
	/**
	 * Visits a ForgeVariable.
	 */
	protected abstract T visit(ForgeVariable expr);
	
	/**
	 * Visits a UnaryExpression.
	 */
	protected abstract T visit(UnaryExpression expr);
	
	/**
	 * Visits a BinaryExpression.
	 */
	protected abstract T visit(BinaryExpression expr);
	
	/**
	 * Visits a ConditionalExpression.
	 */
	protected abstract T visit(ConditionalExpression expr);
	
	/**
	 * Visits a ProjectionExpression.
	 */
	protected abstract T visit(ProjectionExpression expr);
	
	/**
	 * Visits a QuantifyExpression.
	 */
	protected abstract T visit(QuantifyExpression expr);
	
	/**
	 * Visits an OldExpression.
	 */
	protected abstract T visit(OldExpression expr);
	
}
