package forge.program;

import edu.mit.csail.sdg.annotations.Returns;
import edu.mit.csail.sdg.annotations.SpecField;

/**
 * A Forge expression.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public interface ForgeExpression extends ProgramElement {
	
	/**
	 * Domains, variables, and literals -- expressions that have a name.
	 */
	@SpecField({"name: String"})
	public static interface Leaf extends ForgeExpression {
		/**
		 * Returns the expression's name.
		 */
		 public String name();
	}
	
	/**
	 * Variables and instance domains -- leafs that may be modified.
	 */
	@SpecField({"isVariable: boolean", "isGlobal : boolean"})
	public static interface Modifiable extends Leaf {
		/**
		 * True if instanceof ForgeVariable, false if InstanceDomain.
		 */
		@Returns("this.isVariable")
		public boolean isVariable();
		
		/**
		 * True if instanceof InstanceDomain or GlobalVariable, false if LocalVariable.
		 */
		@Returns("this.isGlobal")
		public boolean isGlobal();
	}

	/**
	 * Returns the arity of this expression.
	 */
	public int arity();
	
	/**
	 * Returns the type of this expression.
	 */
	public ForgeType type();
	
	/**
	 * Returns true if the arity is one.
	 */
	public boolean isUnary();

	/**
	 * Accept a visitor.
	 */
	public <T> T accept(ExpressionVisitor<T> visitor);
	
	// projection expressions
	public ForgeExpression domain();
	public ForgeExpression range();
	public ForgeExpression iden();
	public ForgeExpression projection(int... columns);
	
	// unary operator expressions
	public ForgeExpression closure();
	public ForgeExpression no();
	public ForgeExpression lone();
	public ForgeExpression one();
	public ForgeExpression some();
	public ForgeExpression size();
	public ForgeExpression sum();
	public ForgeExpression neg();
	public ForgeExpression bool();
	public ForgeExpression not();
	public ForgeExpression bitNot();
	public ForgeExpression apply(UnaryExpression.Op op);

	// binary operator expressions
	public ForgeExpression eq(ForgeExpression expr);
	public ForgeExpression in(ForgeExpression expr);
	public ForgeExpression union(ForgeExpression expr);
	public ForgeExpression difference(ForgeExpression expr);
	public ForgeExpression intersection(ForgeExpression expr);
	public ForgeExpression join(ForgeExpression expr);
	public ForgeExpression product(ForgeExpression expr);
	public ForgeExpression override(ForgeExpression expr);
	public ForgeExpression and(ForgeExpression expr);
	public ForgeExpression or(ForgeExpression expr);
	public ForgeExpression xor(ForgeExpression expr);
	public ForgeExpression implies(ForgeExpression expr);
	public ForgeExpression iff(ForgeExpression expr);
	public ForgeExpression lt(ForgeExpression expr);
	public ForgeExpression gt(ForgeExpression expr);
	public ForgeExpression lte(ForgeExpression expr);
	public ForgeExpression gte(ForgeExpression expr);
	public ForgeExpression plus(ForgeExpression expr);
	public ForgeExpression minus(ForgeExpression expr);
	public ForgeExpression times(ForgeExpression expr);
	public ForgeExpression divide(ForgeExpression expr);
	public ForgeExpression modulo(ForgeExpression expr);
	public ForgeExpression bitAnd(ForgeExpression expr);
	public ForgeExpression bitOr(ForgeExpression expr);
	public ForgeExpression bitXor(ForgeExpression expr);
	public ForgeExpression shiftLeft(ForgeExpression expr);
	public ForgeExpression shiftRight(ForgeExpression expr);
	public ForgeExpression unsignedShiftRight(ForgeExpression expr);
	public ForgeExpression compose(BinaryExpression.Op op, ForgeExpression expr);
	
	// conditional expression
	public ForgeExpression thenElse(ForgeExpression thenExpr, ForgeExpression elseExpr);
	
	// quantify expressions
	public ForgeExpression forAll(LocalDecls decls);
	public ForgeExpression forSome(LocalDecls decls);
	public ForgeExpression summation(LocalDecls decls);
	public ForgeExpression comprehension(LocalDecls decls);
	public ForgeExpression quantify(QuantifyExpression.Op quant, LocalDecls decls);

}