package forge.program;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import edu.mit.csail.sdg.annotations.Throws;
import forge.program.LocalDecls.EmptyDecls;

/**
 * Manages the Forge classes, fields, and procedures.
 *
 * @specfield domains: set DomainType
 * @specfield globals: set GlobalVariable
 * @specfield locals: set LocalVariable
 * @specfield procedures: set ForgeProcedure
 * @specfield booleanClass, integerClass: ForgeClass
 * @invariant (booleanClass + integerClass) in classes
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class ForgeProgram {
	
	private static final int MAX_DOMAINS = 100;
	
	private final TypeFactory typeFactory;
	private final LocalDecls emptyDecls;
	
	private final BooleanDomain boolDomain;
	private final IntegerDomain intDomain;
	private final Set<InstanceDomain> instDomains;
	
	private final BooleanLiteral trueLiteral;
	private final BooleanLiteral falseLiteral;
	private final Map<Integer, IntegerLiteral> int2Literal;
	private final Set<InstanceLiteral> instLiterals;
	private final Set<GlobalVariable> globals;

	/**
	 * Constructs an empty Forge program.
	 */
	public ForgeProgram() {
		// initialize the type factory
		typeFactory = new TypeFactory(this, MAX_DOMAINS);
		emptyDecls  = new EmptyDecls(this);
		
		// intialize the domains
		boolDomain  = new BooleanDomain(typeFactory);
		intDomain   = new IntegerDomain(typeFactory);
		instDomains = newLeafSet();
		
		// initialize the literals
		trueLiteral  = new BooleanLiteral(boolDomain, true);
		falseLiteral = new BooleanLiteral(boolDomain, false);
		int2Literal  = new HashMap<Integer, IntegerLiteral>();
		instLiterals = newLeafSet();
		
		// initialize the globals
		globals = newLeafSet();
	}
	
	@Throws({"NullPointerException : elem = null",
		     "IllegalArgumentException : elem.program != this"})
	public void checkProgram(ProgramElement elem) {
		if (this != elem.program()) {
			throw new IllegalArgumentException(elem +
					" does not belong to the program " + this);
		}
	}
	
	/**
	 * Returns the empty set constant/type.
	 */
	public ForgeType.Unary empty() {
		return typeFactory.empty();
	}
	
	/**
	 * Returns the empty constant/type of the specified arity.
	 */
	public ForgeType empty(int arity) {
		return typeFactory.empty(arity);
	}
	
	/**
	 * Returns the empty decls.
	 */
	public LocalDecls emptyDecls() {
		return emptyDecls;
	}
	
	/**
	 * Returns the boolean domain.
	 */
	public BooleanDomain booleanDomain() {
		return boolDomain;
	}
	
	/**
	 * Returns the integer domain.
	 */
	public IntegerDomain integerDomain() {
		return intDomain;
	}
	
	/**
	 * Constructs and returns a new instance domain.
	 */
	public InstanceDomain newInstanceDomain(String name) {
		InstanceDomain domain = new InstanceDomain(typeFactory, name);
		instDomains.add(domain);
		return domain;
	}
	
	/**
	 * Returns the instance domains.
	 */
	public Set<InstanceDomain> instanceDomains() {
		return Collections.unmodifiableSet(instDomains);
	}
	
	/**
	 * Returns the true literal.
	 */
	public BooleanLiteral trueLiteral() {
		return trueLiteral;
	}
	
	/**
	 * Returns the false literal.
	 */
	public BooleanLiteral falseLiteral() {
		return falseLiteral;
	}
	
	/**
	 * Returns the integer literal with the specified value.
	 */
	public IntegerLiteral integerLiteral(int value) {
		IntegerLiteral intLit = int2Literal.get(value);
		if (intLit == null) {
			intLit = new IntegerLiteral(intDomain, value);
			int2Literal.put(value, intLit);
		}
		return intLit;
	}
	
	/**
	 * Constructs and returns a new instance literal.
	 */
	public InstanceLiteral newInstanceLiteral(String name, InstanceDomain type) {
		InstanceLiteral lit = new InstanceLiteral(name, type);
		instLiterals.add(lit);
		return lit;
	}
	
	/**
	 * Returns the set of all instance literals.
	 */
	public Set<InstanceLiteral> instanceLiterals() {
		return Collections.unmodifiableSet(instLiterals);
	}
	
	/**
	 * Constructs and returns a new global variable.
	 */
	public GlobalVariable newGlobalVariable(String name, ForgeType type) {
		GlobalVariable global = new GlobalVariable(name, type);
		globals.add(global);
		return global;
	}
	
	/**
	 * Returns the set of all global variables in the program.
	 */
	public Set<GlobalVariable> globalVariables() {
		return Collections.unmodifiableSet(globals);
	}
	
	/**
	 * Constructs and returns a new local variable.
	 */
	public LocalVariable newLocalVariable(String name, ForgeType type) {
		return new LocalVariable(name, type);
	}
	
	/**
	 * Constructs and returns a new procedure.
	 */
	public ForgeProcedure newProcedure(String name, LocalDecls ins, LocalDecls outs) {
		return new ForgeProcedure(this, name, ins, outs);
	}
	
	/**
	 * Construct an empty set that orders program elements by their name.
	 */
	private static <T extends ForgeExpression.Leaf> Set<T> newLeafSet() {
		return new TreeSet<T>(new Comparator<T>() {
			public int compare(T e1, T e2) {
				int c = e1.name().compareTo(e2.name());
				if (c == 0) {
					c = System.identityHashCode(e1) - System.identityHashCode(e2);
				}
				return c;
			}
		});
	}
	
}
