package forge.program;

import java.util.Iterator;

/**
 * A quantified formula expression.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class QuantifyExpression extends AbstractExpression
		implements ForgeExpression {
	
	private final Op op;
	private final LocalDecls decls;
	private final ForgeExpression sub;
	private final ForgeType type;
	
	QuantifyExpression(Op op, ForgeExpression sub, LocalDecls decls) {
		super(sub.program());
		program().checkProgram(decls);
		if (decls.locals().isEmpty()) {
			throw new IllegalArgumentException("decls is empty");
		}
		for (ForgeVariable v : decls.locals()) {
			IllegalArityException.checkUnary(v.arity());
		}
		IllegalArityException.checkUnary(sub.arity());
		this.op = op;
		this.decls = decls;
		this.sub = sub;
		type = op.inferType(decls);
	}

	@Override
	public ForgeType type() {
		return type;
	}
	
	public Op op() {
		return op;
	}
	
	public LocalDecls decls() {
		return decls;
	}
	
	public ForgeExpression sub() {
		return sub;
	}

	public <T> T accept(ExpressionVisitor<T> visitor) {
		return visitor.fullVisit(this);
	}
	
	public String toString() {
		StringBuilder sb = new StringBuilder(op.toString()).append(" ");
		return sb.append(decls).append(" | ").append(sub).toString();
	}
	
	public static enum Op {
		
		ALL("all") {
			ForgeType inferType(LocalDecls decls) {
				return decls.program().booleanDomain();
			}
		},
		
		SOME("some") {
			ForgeType inferType(LocalDecls decls) {
				return decls.program().booleanDomain();
			}
		},
		
		SUM("SUM") {
			ForgeType inferType(LocalDecls decls) {
				return decls.program().integerDomain();
			}
		},
		
		UNION("UNION") {
			ForgeType inferType(LocalDecls decls) {
				Iterator<LocalVariable> i = decls.locals().iterator();
				ForgeType type = i.next().type();
				while (i.hasNext()) type = type.product(i.next().type());
				return type;
			}
		};
		
		private final String s;
		private Op(String s) { this.s = s; }
		public String toString() { return s; }
		
		abstract ForgeType inferType(LocalDecls decls);
	}

}