package forge.program;

import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import kodkod.engine.bool.BooleanMatrix;
import kodkod.instance.TupleSet;
import edu.mit.csail.sdg.util.collections.SDGCollections;
import edu.mit.csail.sdg.util.collections.UniqueList;
import forge.program.ForgeType.Tuple;
import forge.program.ForgeType.Unary;
import forge.util.RelationFactory;

/**
 * Factory for constructing types.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
final class TypeFactory extends
		RelationFactory<ForgeType, ForgeType.Unary, ForgeType.Tuple, ForgeDomain, TypeFactory.DomainHolder> {

	private int occupied = 0;
	
	TypeFactory(ForgeProgram program, int maxDomains) {
		super(program);
		Set<DomainHolder> holders = new HashSet<DomainHolder>();
		for (int i = 0; i < maxDomains; i++) {
			holders.add(new DomainHolder());
		}
		setUniverse(holders);
	}
	
	private DomainHolder nextHolder(ForgeDomain domain) {
		DomainHolder holder = super.atom(occupied);
		occupied++;
		holder.initDomain(domain);
		return holder;
	}
	
	//
	// implementation of abstract methods
	//
	
	@Override
	protected ForgeDomain toAtom(DomainHolder o) {
		return o.domain();
	}

	@Override
	protected ForgeType newBig(TupleSet tuples) {
		return new BigType(tuples);
	}

	@Override
	protected Tuple newTuple(TupleSet tuples) {
		return new TupleType(tuples);
	}

	@Override
	protected Unary newUnary(TupleSet tuples) {
		return new UnaryType(tuples);
	}
	
	/**
	 * Common supertype for all Forge types.
	 */
	private static abstract class AbstractType extends AbstractExpression
			implements ForgeType {
		
		final TypeFactory factory;
		private TupleSet tupleSet;
		private BooleanMatrix matrix = null;
		
		private AbstractType(TypeFactory factory, TupleSet tupleSet) {
			super(factory.program());
			this.factory = factory;
			this.tupleSet = tupleSet;
		}
		
		/**
		 * Super constructor for ForgeDomain only
		 */
		private AbstractType(TypeFactory factory) {
			super(factory.program());
			this.factory = factory;
		}
		
		final TupleSet tupleSet() {
			if (tupleSet == null) {
				DomainHolder holder = factory.nextHolder((ForgeDomain)this);
				tupleSet = factory.singletonTupleSet(holder);
			}
			return tupleSet;
		}
		
		final BooleanMatrix matrix() {
			if (matrix == null) {
				matrix = factory.toBooleanMatrix(tupleSet());
			}
			return matrix;
		}
		
		public final int arity() {
			return tupleSet().arity();
		}
		
		public final ForgeType type() {
			return this;
		}
		
		public final boolean isEmpty() {
			return tupleSet().isEmpty();
		}
		
		public final boolean subtypeOf(ForgeType type) {
			if (arity() != type.arity()) return false;
			return tupleSetOf(type).containsAll(tupleSet());
		}
		
		public final <T> T accept(ExpressionVisitor<T> visitor) {
			return visitor.fullVisit(this);
		}
		
		public final ForgeType join(ForgeType type) {
			return factory.make(factory.joinTupleSet(matrix(), matrixOf(type)));
		}

		public final ForgeType product(ForgeType type) {
			return factory.make(productTuples(type));
		}
		
		final TupleSet productTuples(ForgeType type) {
			return tupleSet().product(tupleSetOf(type));
		}
		
		final TupleSet projectTuples(int... columns) {
			return factory.projectTupleSet(matrix(), columns);
		}
		
		final TupleSet unionTuples(ForgeType type) {
			return factory.unionTupleSet(matrix(), matrixOf(type));
		}
		
		final TupleSet diffTuples(ForgeType type) {
			return factory.differenceTupleSet(matrix(), matrixOf(type));
		}
		
		final TupleSet intersectTuples(ForgeType type) {
			return factory.intersectTupleSet(matrix(), matrixOf(type));
		}
		
		// rest are overridden by subclasses
		
		public boolean isDomain() {
			return false;
		}
		
		public ForgeType.Unary domain() {
			return factory.makeUnary(tupleSet().project(0));
		}
		
		public ForgeType.Unary range() {
			return factory.makeUnary(tupleSet().project(arity() - 1));
		}
		
		public ForgeType projectType(int... columns) {
			return factory.make(projectTuples(columns));
		}
		
		public ForgeType union(ForgeType type) {
			return factory.make(unionTuples(type));
		}

		public ForgeType difference(ForgeType type) {
			return factory.make(diffTuples(type));
		}

		public ForgeType intersection(ForgeType type) {
			return factory.make(intersectTuples(type));
		}
	}
	
	/**
	 * Implementation of ForgeType (that are not unary or tuple).
	 */
	private final class BigType extends AbstractType {
		
		private final Set<ForgeType.Tuple> tuples;
		
		private BigType(TupleSet tupleSet) {
			super(TypeFactory.this, tupleSet);
			assert tupleSet.arity() > 1 && tupleSet.size() != 1;
			tuples = setOfTuples(tupleSet);
		}

		public Set<? extends Tuple> tupleTypes() {
			return tuples;
		}
		
		public String toString() {
			return tupleSet().toString();
		}
	}
	
	/**
	 * Implementation of ForgeType.Unary (that are not tuple)
	 */
	private final class UnaryType extends AbstractType
			implements ForgeType.Unary {
		
		private final Set<ForgeDomain> domains;
		
		private UnaryType(TupleSet tupleSet) {
			super(TypeFactory.this, tupleSet);
			assert tupleSet.arity() == 1 && tupleSet.size() != 1;
			domains = setOfAtoms(tupleSet);
		}
		
		public Set<ForgeDomain> tupleTypes() {
			return domains;
		}

		public boolean isDomain() {
			return false;
		}
		
		public ForgeType.Unary domain() {
			return this;
		}
		
		public ForgeType.Unary range() {
			return this;
		}
		
		public ForgeType.Unary union(ForgeType type) {
			return makeUnary(unionTuples(type));
		}

		public ForgeType.Unary difference(ForgeType type) {
			return makeUnary(diffTuples(type));
		}

		public ForgeType.Unary intersection(ForgeType type) {
			return makeUnary(intersectTuples(type));
		}
		
		public String toString() {
			return tupleSet().toString();
		}
	}
	
	/**
	 * Implementation of ForgeType.Tuple (that are not unary)
	 */
	private final class TupleType extends AbstractType
			implements ForgeType.Tuple {
		
		private final List<ForgeDomain> domains;
		private final Set<ForgeType.Tuple> tuples;
		
		private TupleType(TupleSet tupleSet) {
			super(TypeFactory.this, tupleSet);
			assert tupleSet.size() == 1 && tupleSet.arity() > 1;
			domains = listOfAtoms(tupleSet);
			tuples = Collections.<ForgeType.Tuple>singleton(this);
		}

		public Set<? extends Tuple> tupleTypes() {
			return tuples;
		}
		
		public List<ForgeDomain> domains() {
			return domains;
		}

		public ForgeDomain domain() {
			return domains.get(0);
		}
		
		public ForgeDomain range() {
			return domains.get(arity() - 1);
		}

		public Tuple projectType(int... columns) {
			return makeTuple(projectTuples(columns));
		}
		
		public Tuple product(ForgeType.Tuple tupleType) {
			return makeTuple(productTuples(tupleType));
		}
		
		public String toString() {
			return tupleSet().toString();
		}
	}
	
	/**
	 * Implementation support for ForgeDomain.
	 */
	static abstract class DomainType extends AbstractType
			implements ForgeType.Unary, ForgeType.Tuple {
		
		private final ForgeDomain thisDomain;
		private final UniqueList<ForgeDomain> os;
		
		DomainType(TypeFactory factory) {
			super(factory);
			thisDomain = (ForgeDomain)this;
			os = SDGCollections.singletonUniqueList(thisDomain);
		}
		
		public final Set<ForgeDomain> tupleTypes() {
			return os.asSet();
		}
		
		public final List<ForgeDomain> domains() {
			return os;
		}

		public final boolean isDomain() {
			return true;
		}

		public final ForgeDomain domain() {
			return thisDomain;
		}
		
		public final ForgeDomain range() {
			return thisDomain;
		}
		
		public final Tuple projectType(int... columns) {
			return factory.makeTuple(projectTuples(columns));
		}
		
		public Tuple product(ForgeType.Tuple tupleType) {
			return factory.makeTuple(productTuples(tupleType));
		}
		
		public final ForgeType.Unary union(ForgeType type) {
			return factory.makeUnary(unionTuples(type));
		}

		public final ForgeType.Unary difference(ForgeType type) {
			return factory.makeUnary(diffTuples(type));
		}

		public final ForgeType.Unary intersection(ForgeType type) {
			return factory.makeUnary(intersectTuples(type));
		}
	}
	
	/**
	 * Returns the relation of domains inside the type.
	 */
	private static TupleSet tupleSetOf(ForgeType type) {
		return ((AbstractType)type).tupleSet();
	}
	
	/**
	 * Returns the matrix domains inside the type.
	 */
	private static BooleanMatrix matrixOf(ForgeType type) {
		return ((AbstractType)type).matrix();
	}
	
	static final class DomainHolder {
		
		private ForgeDomain domain = null;
		
		private DomainHolder() {}
		
		private void initDomain(ForgeDomain domain) {
			assert domain != null;
			this.domain = domain;
		}
		
		private ForgeDomain domain() {
			assert domain != null;
			return domain;
		}
		
		public String toString() {
			return domain.toString();
		}
	}

}
