package forge.program;

/**
 * An expression composed of a unary operator and a subexpression.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class UnaryExpression extends AbstractExpression {

	private final Op op;
	private final ForgeExpression sub;
	private final ForgeType type;
	
	UnaryExpression(Op op, ForgeExpression sub) {
		super(sub.program());
		program().checkProgram(sub);
		this.op = op;
		this.sub = sub;
		op.checkArity(sub.arity());
		type = op.inferType(sub.type());
	}
	
	public Op op() {
		return op;
	}
	
	public ForgeExpression sub() {
		return sub;
	}
	
	public ForgeType type() {
		return type;
	}

	public <T> T accept(ExpressionVisitor<T> visitor) {
		return visitor.fullVisit(this);
	}
	
	@Override
	public ForgeExpression apply(Op op) {
		if (this.op.equals(Op.NOT) && op.equals(Op.NOT)) {
			Op.NOT.checkArity(arity());
			return sub;
		}
		return super.apply(op);
	}

	public String toString() {
		StringBuilder sb = new StringBuilder();
		switch(op) {
		case NO: case LONE: case ONE: case SOME:
			sb.append("(").append(op).append(" ").append(sub).append(")");
			break;
		default:
			sb.append(op).append(sub);
			break;
		}
		return sb.toString();
	}
	
	public static enum Op {
		
		/** translative closure */
		CLOSURE("^") {
			void checkArity(int sub) {
				IllegalArityException.checkTwo(sub);
			}
			ForgeType inferType(ForgeType sub) {
				return sub.type();
			}
		},
		
		/** no multiplicity */
		NO("no") {
			void checkArity(int sub) {}
			ForgeType inferType(ForgeType sub) {
				return sub.program().booleanDomain();
			}
		},
		
		/** lone multiplicity */
		LONE("lone") {
			void checkArity(int sub) {}
			ForgeType inferType(ForgeType sub) {
				return sub.program().booleanDomain();
			}
		},
		
		/** one multiplicity */
		ONE("one") {
			void checkArity(int sub) {}
			ForgeType inferType(ForgeType sub) {
				return sub.program().booleanDomain();
			}
		},
		
		/** some multiplicity */
		SOME("some") {
			void checkArity(int sub) {}
			ForgeType inferType(ForgeType sub) {
				return sub.program().booleanDomain();
			}
		},
		
		/** boolean value */
		BOOL("bool") {
			void checkArity(int sub) {
				IllegalArityException.checkUnary(sub);
			}
			ForgeType inferType(ForgeType sub) {
				return sub.program().booleanDomain();
			}
		},
		
		/** boolean negation */
		NOT("!") {
			void checkArity(int sub) {
				IllegalArityException.checkUnary(sub);
			}
			ForgeType inferType(ForgeType sub) {
				return sub.program().booleanDomain();
			}
		},
		
		/** integer negation */
		NEG("-") {
			void checkArity(int sub) {
				IllegalArityException.checkUnary(sub);
			}
			ForgeType inferType(ForgeType sub) {
				return sub.program().integerDomain();
			}
		},
		
		/** integer summation */
		SUM("sum") {
			void checkArity(int sub) {
				IllegalArityException.checkUnary(sub);
			}
			ForgeType inferType(ForgeType sub) {
				return sub.program().integerDomain();
			}
		},
		
		/** bitwise negation */
		BIT_NOT("bnot") {
			void checkArity(int sub) {
				IllegalArityException.checkUnary(sub);
			}
			ForgeType inferType(ForgeType sub) {
				return sub.program().integerDomain();
			}
		},
		
		/** cardinality */
		CARDINALITY("#") {
			void checkArity(int sub) {}
			ForgeType inferType(ForgeType sub) {
				return sub.program().integerDomain();
			}
		};
		
		private final String s;
		Op(String s) { this.s = s ; }
		public String toString() { return s; }
		
		abstract void checkArity(int sub);
		abstract ForgeType inferType(ForgeType sub);
	}
	
}
