package forge.solve;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import kodkod.ast.Expression;
import kodkod.ast.Formula;
import kodkod.ast.NaryFormula;
import kodkod.ast.Relation;
import kodkod.engine.Solver;
import kodkod.engine.config.Options;
import kodkod.instance.Bounds;
import kodkod.instance.TupleFactory;
import kodkod.instance.TupleSet;
import forge.cfg.CFGStmt;
import forge.cfg.ForgeCFG;
import forge.cfg.SpecStmt;
import forge.program.ForgeProcedure;
import forge.program.ForgeProgram;
import forge.program.ForgeVariable;
import forge.program.GlobalVariable;
import forge.program.InstanceDomain;
import forge.program.InstanceLiteral;
import forge.program.LocalVariable;
import forge.program.ForgeExpression.Modifiable;
import forge.solve.ConstantFactory.AbstractConst;
import forge.solve.SolveOptions.SatSolver;
import forge.translate.Environment;
import forge.translate.ExprTranslator;
import forge.translate.ForgeTranslation;
import forge.translate.RelationalModel;
import forge.translate.RelationalModel.DomainRelations;
import forge.translate.RelationalModel.StmtRelations;

/**
 * Constains the parameters to both the Forge and Kodkod analyses.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
class Analysis {

	// forge elems
	final ForgeCFG code, xform;
	final ForgeCFG.Spec spec;
	final SolveOptions options;
	final ForgeBounds bounds;
	
	// translation
	final ForgeTranslation transl;
	
	// kodkod elems
	final Formula kkSpec, kkFormula;
	final Bounds kkBounds;
	final Solver kkSolver;
	
	// * hack for JMLForge *
	final Map<LocalVariable, Relation> skolems;
	
	Analysis(ForgeCFG code, ForgeCFG xform, ForgeTranslation transl,
			SolveOptions options, ForgeCFG.Spec spec, ForgeBounds bounds,
			boolean check) {
		this.code    = code;
		this.xform   = xform;
		this.spec    = spec;
		this.options = options;
		this.bounds  = bounds;
		this.transl  = transl;
		
		// variables that will be re-used
		ForgeProcedure proc = code.procedure();
		Environment initEnv = transl.initialEnv();
		Environment finlEnv = transl.finalEnv();
		Environment specEnv = initEnv.copy();
		ExprTranslator specTransl = new ExprTranslator(specEnv);
		SpecStmt specStmt = spec.entry();
		Set<ForgeVariable> modified = specStmt.modified();
		
		// * hack for JMLForge *
		skolems = new HashMap<LocalVariable, Relation>();
		
		// set the old variables and the new variables
		for (ForgeVariable var : modified) {
			if (var.isGlobal() || proc.inParams().locals().contains(var)) {
				specTransl.pinOldExpr(var);
			}
			
			Expression varExpr = finlEnv.getExpr(var);
			
			// hack for JMLForge skolems
			if (spec.jmlHack() && varExpr == null) {
				LocalVariable skolem = (LocalVariable)var;
				Relation rel = Relation.nary(skolem.name(), skolem.arity());
				skolems.put(skolem, rel);
				varExpr = rel;
			}

			specEnv.putExpr(var, varExpr);
		}

		// make the Kodkod spec
		Formula kkSpec = specTransl.toForm(specStmt.condition());
		kkSpec = kkSpec.and(makeFrame(proc.program().globalVariables(), modified, initEnv, finlEnv));
		kkSpec = kkSpec.and(makeFrame(proc.outParams().locals().asSet(), modified, initEnv, finlEnv));
		this.kkSpec = check ? kkSpec.not() : kkSpec;
		
		kkFormula = kkFormula();
		kkBounds  = kkBounds();
		kkSolver  = kkSolver();
	}
	
	private Formula kkFormula() {
		// important kkSpec is on right, so the ucore will try to drop it first
		return NaryFormula.and(transl.assumes()).and(kkSpec);
	}

	private Bounds kkBounds() {
		// make the universe
		RelationalModel model = transl.model();
		ForgeProgram program = transl.program();
		Bounds kkBounds = new Bounds(bounds.universe());
		TupleFactory factory = bounds.universe().factory();
		
		// bound the boolean relations
		TupleSet trueSet = constTuples(bounds.boolAtom(true));
		TupleSet falseSet = constTuples(bounds.boolAtom(false));
		TupleSet boolSet = constTuples(bounds.extent(program.booleanDomain()));
		kkBounds.boundExactly(model.trueRelation(), trueSet);
		kkBounds.boundExactly(model.falseRelation(), falseSet);
		kkBounds.boundExactly(model.boolRelation(), boolSet);
		
		// bound the integer relations
		ForgeConstant.Unary ints = bounds.extent(program.integerDomain());
		TupleSet intSet = constTuples(ints);
		kkBounds.boundExactly(model.intRelation(), intSet);
		for (ForgeAtom atom : ints.tuples()) {
			int value = ((IntegerAtom)atom).value();
			kkBounds.boundExactly(value, constTuples(atom));
		}
		
		// bound the instance domain relations
		for (InstanceDomain idom : program.instanceDomains()) {
			DomainRelations domRels = model.domainRelations(idom);
			int numCreate = bounds.instanceBound(idom);
			
			// make first tuples
			TupleSet firstSet = numCreate == 0 ? factory.noneOf(1) :
				constTuples(bounds.instanceAtom(idom, 0));
			
			// make the create and order tuples
			TupleSet createSet = firstSet.clone();
			TupleSet orderSet = factory.noneOf(2);
			for (int i = 1; i < bounds.instanceBound(idom); i++) {
				ForgeAtom prev = bounds.instanceAtom(idom, i - 1);
				ForgeAtom next = bounds.instanceAtom(idom, i);
				orderSet.add(factory.tuple(prev, next));
				createSet.addAll(constTuples(next));
			}
			
			// bound the other relations
			kkBounds.boundExactly(domRels.firstRelation(), firstSet);
			kkBounds.boundExactly(domRels.createRelation(), createSet);
			kkBounds.boundExactly(domRels.orderRelation(), orderSet);
			kkBounds.bound(domRels.initialRelation(), createSet);
		}
		
		// bound the instance literal relations
		for (InstanceLiteral lit: program.instanceLiterals()) {
			Relation rel = model.literalRelation(lit);
			TupleSet litSet = constTuples(bounds.instanceAtom(lit));
			kkBounds.boundExactly(rel, litSet);
		}
		
		// bounds the parameters
		for (LocalVariable param : code.procedure().inParams().locals()) {
			Relation rel = model.paramRelation(param);
			TupleSet lowerTS = constTuples(bounds.initialLowerBound(param));
			TupleSet upperTS = constTuples(bounds.initialUpperBound(param));
			kkBounds.bound(rel, lowerTS, upperTS);
		}
		
		// bound the global variables
		for (GlobalVariable global : program.globalVariables()) {
			Relation rel = model.globalRelation(global);
			TupleSet lowerTS = constTuples(bounds.initialLowerBound(global));
			TupleSet upperTS = constTuples(bounds.initialUpperBound(global));
			kkBounds.bound(rel, lowerTS, upperTS);
		}
		
		// bound the existential variables
		for (Entry<CFGStmt,StmtRelations> e : model.stmtRelations()) {
			for (Entry<Modifiable,Relation> f : e.getValue()) {
				TupleSet ts = constTuples(bounds.extent(f.getKey().type()));
				kkBounds.bound(f.getValue(), ts);
			}
		}
		
		// * hack for JMLForge
		// bound skolems
		for (Entry<LocalVariable, Relation> e : skolems.entrySet()) {
			TupleSet ts = constTuples(bounds.extent(e.getKey().type()));
			kkBounds.bound(e.getValue(), ts);
		}
		
		return kkBounds;
	}
	
	private Solver kkSolver() {
		Options kkOptions = new Options();
		kkOptions.setBitwidth(bounds.bitWidth());
		kkOptions.setReporter(options.reporter().kkReporter);
		kkOptions.setFlatten(false);
		SatSolver solver = options.satSolver();
		kkOptions.setSolver(solver.kkSolver);
		kkOptions.setLogTranslation(options.coverage() ? 2 : 0);
		return new Solver(kkOptions);
	}
	
	private static TupleSet constTuples(ForgeConstant c) {
		return ((AbstractConst)c).tupleSet();
	}
	
	private static Formula makeFrame(Set<? extends ForgeVariable> vars,
			Set<? extends ForgeVariable> modified, Environment initEnv,
			Environment finlEnv) {
		Formula frame = Formula.TRUE;
		for (ForgeVariable v : vars) {
			if (!modified.contains(v)) {
				Expression initExpr = initEnv.getExpr(v);
				Expression finlExpr = finlEnv.getExpr(v);
				frame = frame.and(initExpr.eq(finlExpr));
			}
		}
		return frame;
	}
	
}
