package forge.solve;

import java.util.Collections;
import java.util.Comparator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import edu.mit.csail.sdg.annotations.Returns;
import forge.program.ForgeProgram;
import forge.program.ProgramElement;
import forge.program.ForgeExpression.Modifiable;

/**
 * Binds variables to constants.
 *
 * @specfield program: ForgeScene
 * @specfield mod2Value: Modifiable -> TupleSet
 * @invariant domain(mod2Value).program = this.program
 * @invariant range(mod2Value).factory = this.factory
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class Binding implements ProgramElement {
	
	private final ForgeBounds bounds;
	private final Map<Modifiable, ForgeConstant> mod2Value;
	
	/**
	 * Constructs an empty binding.
	 *
	 * @effect no this.var2Tuples'
	 */
	public Binding(ForgeBounds bounds) {
		this(bounds, new TreeMap<Modifiable, ForgeConstant>(MOD_COMP));
	}
	
	/**
	 * Constructs a copy of the specified binding.
	 *
	 * @effect this.var2Tuples' = binding.var2Tuples
	 */
	public Binding copy() {
		return new Binding(bounds,
				new TreeMap<Modifiable, ForgeConstant>(mod2Value));
	}
	
	/**
	 * Returns an unmodifiable view of this binding.
	 */
	public Binding unmodifiableView() {
		return new Binding(bounds,
				Collections.unmodifiableMap(mod2Value));
	}
	
	private Binding(ForgeBounds bounds,
			Map<Modifiable, ForgeConstant> mod2Value) {
		if (bounds == null) throw new NullPointerException("bounds");
		if (mod2Value == null) throw new NullPointerException("mod2Value");
		this.bounds = bounds;
		this.mod2Value = mod2Value;
	}

	public ForgeProgram program() {
		return bounds.program();
	}
	
	/**
	 * Returns the bounds to which this binding belongs.
	 */
	@Returns("this.bounds")
	public ForgeBounds bounds() {
		return bounds;
	}
	
	/**
	 * Returns an unmodifiable view of the set of bound modifiables.
	 */
	public Set<Modifiable> bound() {
		return Collections.unmodifiableSet(mod2Value.keySet());
	}
	
	/**
	 * Returns the constant to which the given modifiable is bound.
	 */
	@Returns("this.mod2Value.elems[mod]")
	public ForgeConstant getValue(Modifiable mod) {
		program().checkProgram(mod);
		return mod2Value.get(mod);
	}
	
	/**
	 * Maps the given modifiable to the given constant.
	 * 
	 * @effect this.dom2Value' = this.dom2Value ++ (domain -> value)
	 */
	public void putValue(Modifiable mod, ForgeConstant value) {
		program().checkProgram(mod);
		bounds.checkBounds(value);
		mod2Value.put(mod, value);
	}

	@Override
	public String toString() {
		return mod2Value.toString();
	}
	
	/**
	 * Comparator for modifiables
	 */
	private static final Comparator<Modifiable> MOD_COMP = new Comparator<Modifiable>() {
		public int compare(Modifiable e1, Modifiable e2) {
			int c = e1.name().compareTo(e2.name());
			if (c == 0) {
				c = System.identityHashCode(e1) - System.identityHashCode(e2);
			}
			return c;
		}
	};
}
