package forge.solve;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import kodkod.engine.bool.BooleanMatrix;
import kodkod.instance.Tuple;
import kodkod.instance.TupleSet;
import edu.mit.csail.sdg.util.bits.IntegerRepresentation;
import edu.mit.csail.sdg.util.bits.TwosComplement;
import edu.mit.csail.sdg.util.collections.SDGCollections;
import edu.mit.csail.sdg.util.collections.UniqueList;
import forge.program.ForgeDomain;
import forge.program.ForgeProgram;
import forge.program.ForgeType;
import forge.program.InstanceDomain;
import forge.program.InstanceLiteral;
import forge.program.IntegerDomain;
import forge.solve.ForgeConstant.Unary;
import forge.util.RelationFactory;

/**
 * Factory for constructing constants
 * 
 * @author Greg Dennis (gdennis@mit.edu)
 */
final class ConstantFactory extends
		RelationFactory<ForgeConstant, ForgeConstant.Unary, ForgeConstant.Tuple, ForgeAtom, ForgeAtom> {
	
	private final IntegerRepresentation intRep;
	private final BooleanAtom trueAtom, falseAtom;
	private final IntegerAtom zeroAtom;
	private final IntegerAtom[] posInts, negInts;
	private final Map<InstanceLiteral, InstanceAtom> ilit2Atom;
	private final Map<InstanceDomain, InstanceAtom[]> createAtoms;
	
	ConstantFactory(ForgeProgram program, int bitWidth,
			Map<InstanceDomain, Integer> dom2Bounds) {
		super(program);
		intRep = new TwosComplement(bitWidth);
		Set<ForgeAtom> atoms = new HashSet<ForgeAtom>();
		
		// boolean atoms
		trueAtom  = new BooleanAtom(this, true, program.booleanDomain());
		falseAtom = new BooleanAtom(this, false, program.booleanDomain());
		atoms.add(trueAtom);
		atoms.add(falseAtom);
		
		// zero atom
		IntegerDomain intDom = program.integerDomain();
		zeroAtom = new IntegerAtom(this, 0, intDom);
		atoms.add(zeroAtom);
		
		// positive int atoms
		posInts = new IntegerAtom[intRep.maxInt()];
		for (int i = 1; i <= intRep.maxInt(); i++) {
			IntegerAtom atom = new IntegerAtom(this, i, intDom);
			posInts[i - 1] = atom;
			atoms.add(atom);
		}
		
		// negative int atoms
		negInts = new IntegerAtom[-intRep.minInt()];
		for (int i = intRep.minInt(); i <= -1; i++) {
			IntegerAtom atom = new IntegerAtom(this, i, intDom);
			negInts[-i - 1] = atom;
			atoms.add(atom);
		}
		
		ilit2Atom = new HashMap<InstanceLiteral, InstanceAtom>();
		for (InstanceLiteral lit : program.instanceLiterals()) {
			InstanceAtom atom = new InstanceAtom(this, lit.name(), lit.type());
			ilit2Atom.put(lit, atom);
			atoms.add(atom);
		}
		
		createAtoms = new HashMap<InstanceDomain, InstanceAtom[]>();
		for (InstanceDomain idom : program.instanceDomains()) {
			Integer bound = dom2Bounds.get(idom);
			if (bound == null) throw new IllegalArgumentException(idom + " not bound");
			InstanceAtom[] instances = new InstanceAtom[bound];
			
			for (int i = 0; i < bound; i++) {
				InstanceAtom atom = new InstanceAtom(this, idom.name() + i, idom);
				instances[i] = atom;
				atoms.add(atom);
			}
			
			createAtoms.put(idom, instances);
		}
		
		setUniverse(atoms);
	}
	
	void checkFactory(ForgeConstant constant) {
		if (!this.equals(((AbstractConst)constant).factory)) {
			throw new IllegalArgumentException("factory mismatch");
		}
	}
	
	//
	// implement abstract stuff
	//
	
	@Override
	protected ForgeAtom toAtom(ForgeAtom o) {
		return o;
	}
	
	@Override
	protected ForgeConstant newBig(TupleSet tuples) {
		return new BigConst(tuples);
	}

	@Override
	protected forge.solve.ForgeConstant.Tuple newTuple(TupleSet tuples) {
		return new TupleConst(tuples);
	}

	@Override
	protected Unary newUnary(TupleSet tuples) {
		return new UnaryConst(tuples);
	}
	
	TupleSet emptyTuples(int arity) {
		return super.emptyTupleSet(arity);
	}
	
	ForgeConstant makeConst(TupleSet tupleSet) {
		return super.make(tupleSet);
	}
	
	ForgeConstant.Unary makeUnaryConst(TupleSet tupleSet) {
		return super.makeUnary(tupleSet);
	}
	
	IntegerRepresentation intRep() {
		return intRep;
	}
	
	BooleanAtom boolAtom(boolean value) {
		return value ? trueAtom : falseAtom;
	}
	
	IntegerAtom intAtom(int value) {
		return value == 0 ? zeroAtom :
			(value > 0 ? posInts[value - 1] : negInts[-value - 1]);
	}
	
	int instanceBound(InstanceDomain idom) {
		return createAtoms.get(idom).length;
	}
	
	InstanceAtom instanceAtom(InstanceDomain idom, int createIndex) {
		return createAtoms.get(idom)[createIndex];
	}
	
	InstanceAtom instanceAtom(InstanceLiteral ilit) {
		return ilit2Atom.get(ilit);
	}
	
	/**
	 * Common supertype for all Forge constants
	 */
	static abstract class AbstractConst implements ForgeConstant {
		
		final ConstantFactory factory;
		private TupleSet tupleSet = null;
		private BooleanMatrix matrix = null;
		
		private AbstractConst(ConstantFactory factory, TupleSet tupleSet) {
			this.factory = factory;
			this.tupleSet = tupleSet; 
		}
		
		private AbstractConst(ConstantFactory factory) {
			this.factory = factory;
		}
		
		public TupleSet tupleSet() {
			if (tupleSet == null) {
				tupleSet = factory.singletonTupleSet((ForgeAtom)this);
			}
			return tupleSet;
		}
		
		final BooleanMatrix matrix() {
			if (matrix == null) {
				matrix = factory.toBooleanMatrix(tupleSet());
			}
			return matrix;
		}
		
		public final ForgeProgram program() {
			return factory.program();
		}
		
		public final int arity() {
			return tupleSet().arity();
		}
		
		public final boolean isEmpty() {
			return tupleSet().isEmpty();
		}
		
		public final boolean isUnary() {
			return arity() == 1;
		}
		
		public boolean isTuple() {
			return tupleSet().size() == 1;
		}
		
		public final boolean subsetOf(ForgeConstant c) {
			return tupleSetOf(c).containsAll(tupleSet());
		}

		public final ForgeConstant join(ForgeConstant c) {
			return factory.make(factory.joinTupleSet(matrix(), matrixOf(c)));
		}

		public final ForgeConstant product(ForgeConstant c) {
			return factory.make(productTuples(c));
		}
		
		final TupleSet productTuples(ForgeConstant c) {
			return tupleSet().product(tupleSetOf(c));
		}
		
		final TupleSet projectTuples(int... columns) {
			return factory.projectTupleSet(matrix(), columns);
		}
		
		final TupleSet unionTuples(ForgeConstant c) {
			return factory.unionTupleSet(matrix(), matrixOf(c));
		}
		
		final TupleSet diffTuples(ForgeConstant c) {
			return factory.differenceTupleSet(matrix(), matrixOf(c));
		}
		
		final TupleSet intersectTuples(ForgeConstant c) {
			return factory.intersectTupleSet(matrix(), matrixOf(c));
		}
		
		// rest are overridden by subclasses
		
		public boolean isDomain() {
			return false;
		}
		
		public ForgeConstant.Unary domain() {
			return factory.makeUnary(tupleSet().project(0));
		}
		
		public ForgeConstant.Unary range() {
			return factory.makeUnary(tupleSet().project(arity() - 1));
		}
		
		public ForgeConstant projection(int... columns) {
			return factory.make(projectTuples(columns));
		}
		
		public ForgeConstant union(ForgeConstant c) {
			return factory.make(unionTuples(c));
		}

		public ForgeConstant difference(ForgeConstant c) {
			return factory.make(diffTuples(c));
		}

		public ForgeConstant intersection(ForgeConstant c) {
			return factory.make(intersectTuples(c));
		}
	}
	
	/**
	 * Implementation of ForgeConstnat (that are not unary or tuple).
	 */
	private final class BigConst extends AbstractConst {
		
		private final Set<ForgeConstant.Tuple> tuples;
		private final ForgeType type;
		
		private BigConst(TupleSet tupleSet) {
			super(ConstantFactory.this, tupleSet);
			assert tupleSet.arity() > 1 && tupleSet.size() != 1;
			tuples = setOfTuples(tupleSet);
			
			ForgeType t = program().empty(tupleSet.arity());
			for (kodkod.instance.Tuple tuple : tupleSet) {
				t = t.union(tupleType(tuple));
			}
			this.type = t;
		}
		
		public ForgeType type() {
			return type;
		}

		public Set<? extends Tuple> tuples() {
			return tuples;
		}
		
		public String toString() {
			return tupleSet().toString();
		}
	}
	
	/**
	 * Implementation of ForgeConstant.Unary (that are not tuple).
	 */
	private final class UnaryConst extends AbstractConst
			implements ForgeConstant.Unary {
		
		private final Set<ForgeAtom> atoms;
		private final ForgeType.Unary type;
		
		private UnaryConst(TupleSet tupleSet) {
			super(ConstantFactory.this, tupleSet);
			assert tupleSet.arity() == 1 && tupleSet.size() != 1;
			atoms = setOfAtoms(tupleSet);
			
			ForgeType.Unary t = program().empty();
			for (kodkod.instance.Tuple tuple : tupleSet) {
				t = t.union(atomType(tuple, 0));
			}
			this.type = t;
		}
		
		public ForgeType.Unary type() {
			return type;
		}
		
		public Set<ForgeAtom> tuples() {
			return atoms;
		}

		public boolean isAtom() {
			return false;
		}
		
		public ForgeConstant.Unary domain() {
			return this;
		}
		
		public ForgeConstant.Unary range() {
			return this;
		}
		
		public ForgeConstant.Unary union(ForgeConstant c) {
			return makeUnary(unionTuples(c));
		}

		public ForgeConstant.Unary difference(ForgeConstant c) {
			return makeUnary(diffTuples(c));
		}

		public ForgeConstant.Unary intersection(ForgeConstant c) {
			return makeUnary(intersectTuples(c));
		}
		
		public String toString() {
			return tupleSet().toString();
		}
	}
	
	/**
	 * Implementation of ForgeConstant.Tuple (that are not unary).
	 */
	private final class TupleConst extends AbstractConst
			implements ForgeConstant.Tuple {
		
		private final List<ForgeAtom> atoms;
		private final Set<ForgeConstant.Tuple> tuples;
		private final ForgeType.Tuple type;
		
		private TupleConst(TupleSet tupleSet) {
			super(ConstantFactory.this, tupleSet);
			assert tupleSet.size() == 1 && tupleSet.arity() > 1;
			atoms = listOfAtoms(tupleSet);
			tuples = Collections.<ForgeConstant.Tuple>singleton(this);
			type = tupleType(tupleSet.iterator().next());
		}
		
		public ForgeType.Tuple type() {
			return type;
		}

		public Set<? extends Tuple> tuples() {
			return tuples;
		}
		
		public List<ForgeAtom> atoms() {
			return atoms;
		}

		public ForgeAtom domain() {
			return atoms.get(0);
		}
		
		public ForgeAtom range() {
			return atoms.get(arity() - 1);
		}

		public Tuple projection(int... columns) {
			return makeTuple(projectTuples(columns));
		}
		
		public Tuple product(ForgeConstant.Tuple tuple) {
			return makeTuple(productTuples(tuple));
		}
			
		public String toString() {
			return tupleSet().toString();
		}
	}
	
	/**
	 * Implementation support for ForgeAtom.
	 */
	static abstract class AtomConst extends AbstractConst
			implements ForgeConstant.Unary, ForgeConstant.Tuple {
		
		private final ForgeAtom thisAtom;
		private final UniqueList<ForgeAtom> os;
		
		AtomConst(ConstantFactory factory) {
			super(factory);
			thisAtom = (ForgeAtom)this;
			os = SDGCollections.singletonUniqueList(thisAtom);
		}
		
		public abstract ForgeDomain type();
		
		public final Set<ForgeAtom> tuples() {
			return os.asSet();
		}
		
		public final List<ForgeAtom> atoms() {
			return os;
		}

		public final ForgeAtom domain() {
			return thisAtom;
		}
		
		public final ForgeAtom range() {
			return thisAtom;
		}
		
		public final Tuple projection(int... columns) {
			return factory.makeTuple(projectTuples(columns));
		}
		
		public Tuple product(ForgeConstant.Tuple tuple) {
			return factory.makeTuple(productTuples(tuple));
		}
		
		public final ForgeConstant.Unary union(ForgeConstant c) {
			return factory.makeUnary(unionTuples(c));
		}

		public final ForgeConstant.Unary difference(ForgeConstant c) {
			return factory.makeUnary(diffTuples(c));
		}

		public final ForgeConstant.Unary intersection(ForgeConstant c) {
			return factory.makeUnary(intersectTuples(c));
		}
	}
	
	/**
	 * Returns the relation of domains inside the type.
	 */
	private static TupleSet tupleSetOf(ForgeConstant c) {
		return ((AbstractConst)c).tupleSet();
	}
	
	/**
	 * Returns the matrix of atoms inside the constant.
	 */
	private static BooleanMatrix matrixOf(ForgeConstant c) {
		return ((AbstractConst)c).matrix();
	}
	
	/**
	 * Returns the type of the atom in the given tuple.
	 */
	private static ForgeDomain atomType(Tuple/*<ForgeAtom>*/ tuple, int i) {
		return ((ForgeAtom)tuple.atom(i)).type();
	}
	
	/**
	 * Returns the type of the given tuple.
	 */
	private static ForgeType.Tuple tupleType(Tuple/*<ForgeAtom>*/ tuple) {
		ForgeType.Tuple type = atomType(tuple, 0);
		for (int i = 1; i < tuple.arity(); i++) {
			type = type.product(atomType(tuple, i));
		}
		return type;
	}
	
}
