package forge.solve;

import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import kodkod.ast.Formula;
import edu.mit.csail.sdg.annotations.Effects;
import edu.mit.csail.sdg.annotations.Invariant;
import edu.mit.csail.sdg.annotations.Modifies;
import edu.mit.csail.sdg.annotations.Requires;
import edu.mit.csail.sdg.annotations.Returns;
import edu.mit.csail.sdg.annotations.Throws;
import forge.cfg.CFGStmt;
import forge.cfg.ExitStmt;
import forge.cfg.ForgeCFG;
import forge.util.BreadthFirstVisitor;

/**
 * Coverage of a 
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class Coverage {
	
	/**
	 * A coverage score given to a statement.
	 */
	public static final class Score {
		
		@Invariant("this.covered >= 0 && this.missed >= 0")
		private final int covered;
		private final int missed;
		
		@Modifies("this.covered, this.missed")
		@Effects("this.covered = 0 && this.missed = 0")
		Score() {
			this(0, 0);
		}
		
		@Effects({"return.covered = this.covered + 1",
			      "return.missed = this.missed"})
		@Throws("IllegalStateException : this.covered = Integer.MAX_VALUE")
		Score incCovered() {
			if (covered == Integer.MAX_VALUE) {
				throw new IllegalStateException();
			}
			return new Score(covered + 1, missed);
		}
		
		@Effects({"return.covered = this.covered",
	              "return.missed = this.missed + 1"})
	    @Throws("IllegalStateException : this.covered = Integer.MAX_VALUE")
		Score incMissed() {
			if (missed == Integer.MAX_VALUE) {
				throw new IllegalStateException();
			}
			return new Score(covered, missed + 1);
		}
		
		@Requires("covered >= 0 && missed >= 0")
		@Modifies("this.covered, this.missed")
		@Effects("this.(Coverage$Score@covered) = covered && this.(Coverage$Score@missed) = missed")
		private Score(int covered, int missed) {
			this.covered = covered;
			this.missed = missed;
		}
		
		@Returns("this.covered")
		public int timesCovered () {
			return covered;
		}

		@Returns("this.missed")
		public int timesMissed() {
			return missed;
		}
		
		@Returns("this.covered + this.missed")
		public int timesTotal() {
			return covered + missed;
		}
		
		public String toString() {
			return "(" + timesCovered() + "/" + timesTotal() + ")";
		}
		
		public int hashCode() {
			return covered + 31 * missed + 7;
		}
		
		public boolean equals(Object o) {
			if (!(o instanceof Score)) return false;
			Score s = (Score)o;
			return covered == s.covered && missed == s.missed;
		}
	}
	
	private final ForgeCFG code;
	private final ForgeCFG.Spec spec;
	private final Set<Formula> kkCore;
	private final Map<CFGStmt, Score> stmt2Score;
	private Set<CFGStmt> missed;
	
	Coverage(ForgeCFG code, ForgeCFG.Spec spec, Set<Formula> kkCore,
			Map<CFGStmt, Score> stmt2Score) {
		this.code = code;
		this.spec = spec;
		this.kkCore = kkCore;
		this.stmt2Score = stmt2Score;
		
		Set<CFGStmt> missed = new HashSet<CFGStmt>();
		for (Entry<CFGStmt, Score> e : stmt2Score.entrySet()) {
			Score score = e.getValue();
			if (score.missed > 0 && score.covered == 0) {
				missed.add(e.getKey());
			}
		}
		this.missed = Collections.unmodifiableSet(missed);
	}
	
	@Returns("this.code")
	public ForgeCFG code() {
		return code;
	}
	
	@Returns("this.spec")
	public ForgeCFG.Spec spec() {
		return spec;
	}
	
	/**
	 * Returns set of Kodkod formulas in the core.
	 */
	@Returns("this.kkCore")
	public Set<?> kkCore() {
		return kkCore;
	}
	
	@Returns("this.stmt2Score.elems[stmt]")
	public Score stmtScore(CFGStmt stmt) {
		return stmt2Score.get(stmt);
	}
	
	/**
	 * Returns the set of statements that could be covered at least once that
	 * are not covered at least once.
	 */
	public Set<CFGStmt> missed() {
		return missed;
	}
	
	public String toString () {
		StringBuilder sb = new StringBuilder();
		BreadthFirstVisitor printer = new PrintVisitor(sb);
		printer.traverse(code);
		printer.traverse(spec);
		return sb.toString();
	}
	
	private static final String NEW_LINE = System.getProperty("line.separator");
	
	private final class PrintVisitor extends BreadthFirstVisitor {
		
		private final StringBuilder sb;
		
		private PrintVisitor(StringBuilder sb) {
			this.sb = sb;
		}
		
		protected void visit(ForgeCFG cfg) {
			sb.append(cfg.procedure().toString()).append(" {");
		}
		
		protected void defaultVisit(CFGStmt stmt) {
			sb.append(NEW_LINE);
			
			if (missed.contains(stmt)) {
				sb.append("[MISSED] ");
			}
//			Score score = stmt2Score.get(stmt);
//			assert score != null;
//			sb.append(score).append(" ");
				
			sb.append(" ").append(stmt);
		}
		
		public void visit(ExitStmt stmt) {
			defaultVisit(stmt);
			sb.append(NEW_LINE).append("}").append(NEW_LINE);
		}
	}
}
