package forge.solve;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import kodkod.ast.Formula;
import kodkod.ast.Node;
import kodkod.engine.Proof;
import kodkod.engine.satlab.ReductionStrategy;
import kodkod.engine.ucore.RCEStrategy;
import edu.mit.csail.sdg.util.collections.Bag;
import edu.mit.csail.sdg.util.collections.HashBag;
import edu.mit.csail.sdg.util.collections.UnionSet;
import forge.cfg.CFGStmt;
import forge.cfg.ExitStmt;
import forge.cfg.SpecStmt;
import forge.cfg.StmtSet;
import forge.program.BinaryExpression;
import forge.program.ConditionalExpression;
import forge.program.ForgeLiteral;
import forge.program.ForgeProcedure;
import forge.program.ForgeType;
import forge.program.ForgeVariable;
import forge.program.LocalVariable;
import forge.program.OldExpression;
import forge.program.ProjectionExpression;
import forge.program.QuantifyExpression;
import forge.program.UnaryExpression;
import forge.solve.Coverage.Score;
import forge.transform.ExpressionDescender;
import forge.transform.Transformer;
import forge.util.BreadthFirstVisitor;
import forge.util.ExpressionUtil;

/**
 * Coverage information. Coverage always includes exit statements.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
final class CoverageMetric {
	
	/** hard coding this strategy for now. */
	private static ReductionStrategy reductionStrategy(Proof proof) {
		return new RCEStrategy(proof.log());
	}
	
	/**
	 * Measure the coverage.
	 */
	public static Coverage measure(Analysis analysis, Proof proof) {
		ForgeProcedure proc = analysis.code.procedure();
		
		// initialize the score of each code and spec statement
		Map<CFGStmt, Score> stmt2Score = new HashMap<CFGStmt, Score>();
		InitScorer initScorer = new InitScorer(stmt2Score);
		initScorer.traverse(analysis.code);
		initScorer.traverse(analysis.spec);
		
		// get the roots of the core and of the original formula
		proof.minimize(reductionStrategy(proof));
		Collection<Node> kkCoreNodes = proof.highLevelCore().values();
		Set<Formula> kkCore = new LinkedHashSet<Formula>();
		for (Node kkNode: kkCoreNodes) kkCore.add((Formula)kkNode);
		kkCore = Collections.unmodifiableSet(kkCore);
		Set<Formula> kkSpec = ExpressionUtil.roots(analysis.kkSpec);
		
		// determine the set of covered statements in the transformed procedure
		StmtSet xformCovered = StmtSet.hash(analysis.xform);
		boolean specCovered = false;
		for (Formula kkForm : kkCore) {
			if (kkSpec.contains(kkForm)) {
				specCovered = true;
			} else {
				StmtSet coreSlice = analysis.transl.slice(kkForm);
				xformCovered.addAll(coreSlice);
			}
		}
		
//		Set<Formula> roots = ExpressionUtil.roots(analysis.kkFormula);
//		for (Formula r : roots) {
//			System.out.println("root: " + r);
//			if (kkCore.contains(r)) {
//				System.out.println("in core");
//				if (kkSpec.contains(r)) {
//					System.out.println("from spec");
//				} else {
//					System.out.println("from " + analysis.transl.slice(r));
//				}
//			} else {
//				System.out.println("not in core");
//			}
//			System.out.println();
//		}
		
		// if the spec is covered
		SpecStmt specStmt = analysis.spec.entry();
		if (specCovered) {
			VarFinder varFinder = new VarFinder();
			specStmt.condition().accept(varFinder);
			Set<ForgeVariable> mods = specStmt.modified();
			Set<ForgeVariable> appear = varFinder.found;
			Set<ForgeVariable> outs = UnionSet.union(
					proc.program().globalVariables(),
					proc.outParams().locals().asSet());
			
			for (ForgeVariable var: outs) {
				if (!mods.contains(var) || appear.contains(var)) {
					StmtSet varSlice = analysis.transl.slice(var);
					xformCovered.addAll(varSlice);
				}
			}
			stmt2Score.put(specStmt, new Score().incCovered());
		} else {
			stmt2Score.put(specStmt, new Score().incMissed());
		}
		xformCovered = xformCovered.unmodifiable();
		
		// scores code statements based on covered xform statements
		new CodeScorer(stmt2Score, xformCovered,
				analysis.options.transformer()).traverse(analysis.xform);
		stmt2Score = Collections.unmodifiableMap(stmt2Score);

		// return the coverage
		return new Coverage(analysis.code, analysis.spec, kkCore, stmt2Score);
	}	

	private CoverageMetric() {}
	
	/**
	 * Initializes the score of each stmt to zero covers, zero misses.
	 */
	private static final class InitScorer extends BreadthFirstVisitor {
		
		private final Map<CFGStmt, Score> stmt2Score;
		
		private InitScorer(Map<CFGStmt, Score> stmt2Score) {
			this.stmt2Score = stmt2Score;
		}

		protected void defaultVisit(CFGStmt stmt) {
			stmt2Score.put(stmt, new Score());
		}
	}
	
	/**
	 * Traverses the transformed cfg and if the statement is covered, adds
	 * covered score to source, otherwise adds missed score to source.
	 */
	private static final class CodeScorer extends BreadthFirstVisitor {
		
		private final Map<CFGStmt, Score> stmt2Score;
		private final StmtSet xformCovered;
		private final Transformer xformer;
		
		private CodeScorer(Map<CFGStmt, Score> stmt2Score,
				StmtSet xformCovered, Transformer xformer) {
			this.stmt2Score = stmt2Score;
			this.xformCovered = xformCovered;
			this.xformer = xformer;
		}

		protected void defaultVisit(CFGStmt stmt) {
			CFGStmt source = xformer.sourceStmt(stmt);
			if (source != null) {
				Score score = stmt2Score.get(source);
				assert score != null : "no score for " + stmt;
				if (xformCovered.contains(stmt)) {
					score = score.incCovered();
				} else {
					score = score.incMissed();
				}
				stmt2Score.put(source, score);
				
			}
		}
		
		/** override to do nothing. */
		public void visit(ExitStmt stmt) {}
	}
	
	/**
	 * Finds all non-quantified, non-old appearances of variables
	 */
	private static final class VarFinder extends ExpressionDescender {
		
		private final Set<ForgeVariable> found;
		private final Bag<LocalVariable> quants;
		
		private VarFinder() {
			found = new HashSet<ForgeVariable>();
			quants = new HashBag<LocalVariable>();
		}
		
		@Override
		protected void descend(ForgeVariable expr) {
			if (!quants.contains(expr)) found.add(expr);
			super.putCache(expr, null);
		}
		
		@Override
		protected void descend(OldExpression expr) {
			super.putCache(expr, null);
		}
		
		@Override
		protected void descend(ForgeType expr) {
			super.putCache(expr, null);
		}
		
		@Override
		protected void descend(QuantifyExpression expr) {
			quants.addAll(expr.decls().locals());
			expr.sub().accept(this);
			quants.removeAll(expr.decls().locals());
			super.putCache(expr, null);
		}
		
		@Override
		protected void descend(BinaryExpression expr) {
			expr.left().accept(this);
			expr.right().accept(this);
			super.putCache(expr, null);
		}
		@Override
		protected void descend(ConditionalExpression expr) {
			expr.condition().accept(this);
			expr.thenExpr().accept(this);
			expr.elseExpr().accept(this);
			super.putCache(expr, null);
		}
		@Override
		protected void descend(ProjectionExpression expr) {
			expr.sub().accept(this);
			super.putCache(expr, null);
		}
		@Override
		protected void descend(UnaryExpression expr) {
			expr.sub().accept(this);
			super.putCache(expr, null);
		}
		@Override
		protected void descend(ForgeLiteral expr) {
			super.putCache(expr, null);
		}
	}
}
