package forge.solve;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import kodkod.instance.TupleSet;
import kodkod.instance.Universe;
import forge.program.ForgeDomain;
import forge.program.ForgeProgram;
import forge.program.ForgeType;
import forge.program.ForgeVariable;
import forge.program.InstanceDomain;
import forge.program.InstanceLiteral;
import forge.program.ProgramElement;

/**
 * Bounds type and the intial value of variables by constants.
 * 
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class ForgeBounds implements ProgramElement {

	private final ForgeProgram program;
	private final ConstantFactory factory;
	private final ForgeConstant.Unary bools, ints;
	private final Map<InstanceDomain, ForgeConstant.Unary> idom2Extent;
	private final Map<ForgeVariable, ForgeConstant> var2lower, var2upper;
	
	public ForgeBounds(ForgeProgram program, int bitWidth,
			Map<InstanceDomain, Integer> scopes) {
		this.program = program;
		factory = new ConstantFactory(program, bitWidth, scopes);
		bools = factory.boolAtom(true).union(factory.boolAtom(false));
		
		ForgeConstant.Unary i = factory.empty();
		for (int v = factory.intRep().minInt(); v <= factory.intRep().maxInt(); v++) {
			i = i.union(factory.intAtom(v));
		}
		ints = i;
		
		idom2Extent = new HashMap<InstanceDomain, ForgeConstant.Unary>();
		for (InstanceDomain idom : program.instanceDomains()) {
			ForgeConstant.Unary extent = factory.empty();
			
			for (int j = 0; j < scopes.get(idom); j++) {
				extent = extent.union(factory.instanceAtom(idom, j));
			}
			
			idom2Extent.put(idom, extent);
		}
		
		for (InstanceLiteral ilit : program.instanceLiterals()) {
			InstanceDomain idom = ilit.type();
			ForgeConstant.Unary extent = idom2Extent.get(idom);
			extent = extent.union(factory.instanceAtom(ilit));
			idom2Extent.put(idom, extent);
		}
		
		// initialize lower and upper maps
		var2lower = new HashMap<ForgeVariable, ForgeConstant>();
		var2upper = new HashMap<ForgeVariable, ForgeConstant>();
	}
	
	//
	// package-private
	//
	
	/**
	 * Returns the Kodkod universe.
	 */
	Universe universe() {
		return factory.universe();
	}
	
	/**
	 * Makes a constant from the given tuple set.
	 */
	ForgeConstant makeConstant(TupleSet tuples) {
		return factory.makeConst(tuples);
	}
	
	/**
	 * Checks that this constant belongs to this bounds.
	 */
	void checkBounds(ForgeConstant constant) {
		factory.checkFactory(constant);
	}
	
	//
	// public methods
	//
	
	public ForgeProgram program() {
		return program;
	}
	
	/**
	 * Return the empty unary constant.
	 */
	public ForgeConstant.Unary empty() {
		return factory.empty();
	}
	
	/**
	 * Return the empty constant of the given arity.
	 */
	public ForgeConstant empty(int arity) {
		return factory.empty(arity);
	}
	
	/**
	 * Return the boolean atom with the given value.
	 */
	public BooleanAtom boolAtom(boolean value) {
		return factory.boolAtom(value);
	}
	
	/**
	 * Return the bitwidth of integers in this bound.
	 */
	public int bitWidth() {
		return factory.intRep().width();
	}
	
	/**
	 * Return the minimum integer allowed by the bitwidth.
	 */
	public int minIntValue() {
		return factory.intRep().minInt();
	}
	
	/**
	 * Return the maximum integer allowed by the bitwidth.
	 */
	public int maxIntValue() {
		return factory.intRep().maxInt();
	}
	
	/**
	 * Return integer atom with the given value.
	 */
	public IntegerAtom intAtom(int value) {
		return factory.intAtom(value);
	}
	
	/**
	 * Return the number of atoms to which the given domain is limited.
	 */
	public int instanceBound(InstanceDomain idom) {
		return factory.instanceBound(idom);
	}
	
	/**
	 * Return the ith atom allocated for the given instance domain.
	 */
	public InstanceAtom instanceAtom(InstanceDomain idom, int atomIndex) {
		return factory.instanceAtom(idom, atomIndex);
	}
	
	/**
	 * Return the atom allocated for the given instance literal.
	 */
	public InstanceAtom instanceAtom(InstanceLiteral ilit) {
		return factory.instanceAtom(ilit);
	}
	
	/**
	 * Return the extent of the given domain type.
	 */
	public ForgeConstant.Unary extent(ForgeDomain type) {
		switch (type.kind()) {
		case BOOLEAN:  return bools;
		case INTEGER:  return ints;
		case INSTANCE: return idom2Extent.get(type);
		}
		assert false: "not such kind";
		return null;
	}
	
	/**
	 * Return the extent of the given unary type.
	 */
	public ForgeConstant.Unary extent(ForgeType.Unary type) {
		TupleSet ts = factory.emptyTuples(1);
		for (ForgeDomain dom : type.tupleTypes()) {
			ts.addAll(extent(dom).tupleSet());
		}
		return factory.makeUnaryConst(ts);
	}
	
	/**
	 * Return the extent of the given tuple type.
	 */
	public ForgeConstant extent(ForgeType.Tuple type) {
		return factory.makeConst(extentTS(type));
	}
	
	/**
	 * Return the extent of the given type.
	 */
	public ForgeConstant extent(ForgeType type) {
		TupleSet ts = factory.emptyTuples(type.arity());
		for (ForgeType.Tuple tupleType : type.tupleTypes()) {
			ts.addAll(extentTS(tupleType));
		}
		return factory.makeConst(ts);
	}
	
	private TupleSet extentTS(ForgeType.Tuple type) {
		Iterator<ForgeDomain> i = type.domains().iterator();
		TupleSet ts = extent(i.next()).tupleSet();
		while(i.hasNext()) {
			ts = ts.product(extent(i.next()).tupleSet());
		}
		return ts;
	}
	
	/**
	 * Bounds the initial value of the given variable anove and below.
	 */
	public void boundInitial(ForgeVariable var, ForgeConstant lower, ForgeConstant upper) {
		program.checkProgram(var);
		program.checkProgram(lower);
		program.checkProgram(upper);
		factory.checkFactory(lower);
		factory.checkFactory(upper);
		if (!upper.type().subtypeOf(var.type())) {
			throw new IllegalArgumentException("type mismatch: " + var + "\n  var.type():  " + var.type() + "\n  lower:     " + lower.type() + "\n  upper:     " + upper.type());
		}
		if (!lower.subsetOf(upper)) {
			throw new IllegalArgumentException("invalid bounds");
		}
		var2lower.put(var, lower);
		var2upper.put(var, upper);
	}
	
	/**
	 * Return the initial lower bound for the given variable.
	 */
	public ForgeConstant initialLowerBound(ForgeVariable var) {
		program.checkProgram(var);
		ForgeConstant lower = var2lower.get(var);
		if (lower == null) {
			lower = empty(var.arity());
			var2lower.put(var, lower);
		}
		return lower;
	}
	
	/**
	 * Return the initial upper bound for the given variable.
	 */
	public ForgeConstant initialUpperBound(ForgeVariable var) {
		program.checkProgram(var);
		ForgeConstant upper = var2upper.get(var);
		if (upper == null) {
			upper = extent(var.type());
			var2upper.put(var, upper);
		}
		return upper;
	}
}