package forge.solve;

import java.util.List;
import java.util.Set;

import kodkod.ast.Decl;
import kodkod.ast.Formula;
import kodkod.ast.Relation;
import kodkod.engine.bool.BooleanFormula;
import kodkod.engine.config.Reporter;
import kodkod.instance.Bounds;
import kodkod.util.ints.IntSet;

/**
 * Reports details of the Forge and underlying Kodkod analysis.
 * 
 * @author Greg Dennis (gdennis@mit.edu)
 */
public abstract class ForgeReporter {

	/**
	 * internal Kodkod reporter
	 */
	final Reporter kkReporter = new Reporter() {
		public void flattening(BooleanFormula circuit) {
			ForgeReporter.this.flattening(circuit);
		}
		public void generatingSBP() {
			ForgeReporter.this.generatingSBP();
		}
		public void detectingSymmetries(Bounds bounds) {
			ForgeReporter.this.detectingSymmetries(bounds);
		}
		public void detectedSymmetries(Set<IntSet> parts) {
			ForgeReporter.this.detectedSymmetries(parts);
		}
		public void optimizingBoundsAndFormula() {
			ForgeReporter.this.optimizingBoundsAndFormula();
		}
		public void skolemizing(Decl decl, Relation skolem, List<Decl> context) {
			ForgeReporter.this.skolemizing(decl, skolem, context);
		}
		public void solvingCNF(int primaryVars, int vars, int clauses) {
			ForgeReporter.this.solvingCNF(primaryVars, vars, clauses);
		}
		public void translatingToBoolean(Formula formula, Bounds bounds) {
			ForgeReporter.this.translatingToBoolean(formula, bounds);
		}
		public void translatingToCNF(BooleanFormula circuit) {
			ForgeReporter.this.translatingToCNF(circuit);
		}
	};
	
	/**
	 * Reports that Forge is transforming the procedure.
	 */
	public void transformingProcedure() {}
	
	/**
	 * Reports that Forge is translating the procedure to Kodkod.
	 */
	public void translatingToKodkod() {}
	
	/**
	 * Reports that Forge is building an analysis from a given specification
	 * and bounds.
	 */
	public void buildingAnalysis() {}
	
	/**
	 * Reports that Forge is about to invoke Kodkod to solve the analysis.
	 */
	public void solvingAnalysis() {}
	
	/**
	 * Reports that Kodkod SBP is in progress.
	 */
	public void generatingSBP() {}
	
	/**
	 * Reports that Kodkod is optimizing the bounds and formula.
	 */
	public void optimizingBoundsAndFormula() {}
	
	/**
	 * Reports that Kodkod is detecting symmetries in the given bounds
	 * (kodkod.instance.Bounds).
	 */
	public void detectingSymmetries(Object bounds) {}
	
	/**
	 * Reports the symmetries detected by Kodkod as a set of
	 * (kodkod.util.ints.IntSet) objects.
	 */
	public void detectedSymmetries(Set<?> parts) {}
	
	/**
	 * Reports that Kodkod flattering of the given boolean circuit
	 * (kodkod.engine.bool.BooleanFormula) is in progress.
	 */
	public void flattening(Object circuit) {}
	
	/**
	 * Reports that Kodkod is skolemizing the declaration (kodkod.ast.Decl)
	 * using the given skolem relation (kodkod.ast.Relation). The context list
	 * (java.util.List<kodkod.ast.Decl>) contains non-skolemizable quantified
	 * declarations on which the given decl depends, in the order of
	 * declaration, the most recent being the last in the list.
	 */
	public void skolemizing(Object decl, Object skolem, List<?> context) {}

	/**
	 * Reports that Kodkod generated a CNF consisting of the given number of
	 * variables and clauses, is being analyzed by a SAT solver.
	 */
	public void solvingCNF(int primaryVars, int vars, int clauses) {}

	/**
	 * Reports that Kodkod is translating the given formula (kodkod.ast.Formula)
	 * and bounds (kodkod.instance.Bounds) to a boolean circuit.
	 */
	public void translatingToBoolean(Object formula, Object bounds) {}

	/**
	 * Reports that Kodkod is translating the given boolean circuit
	 * (kodkod.engine.bool.BooleanFormula) to CNF.
	 */
	public void translatingToCNF(Object circuit) {}
	
}
