package forge.solve;

import kodkod.ast.Formula;
import edu.mit.csail.sdg.annotations.Returns;
import forge.cfg.ForgeCFG;
import forge.program.ForgeProgram;

/**
 * A solution that includes coverage information when it is unsatisfiable.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class ForgeSolution {
 
	private final ForgeCFG code;
	private final ForgeCFG.Spec spec;
	private final Formula kkFormula;
	private final Trace trace;
	private final Coverage coverage;
	
	ForgeSolution(Analysis analysis, Trace trace, Coverage coverage) {
		if ((trace != null) && (coverage != null)) {
			throw new IllegalArgumentException("can't have a trace and coverage");
		}
		code      = analysis.code;
		spec      = analysis.spec;
		kkFormula = analysis.kkFormula;
		this.trace    = trace;
		this.coverage = coverage;
	}
	
	public ForgeProgram program() {
		return code.program();
	}
	
	@Returns("this.code")
	public ForgeCFG code() {
		return code;
	}
	
	@Returns("this.spec")
	public ForgeCFG.Spec spec() {
		return spec;
	}
	
	/**
	 * Returns set of Kodkod formula solved.
	 */
	@Returns("this.kkFormula")
	public Object kkFormula() {
		return kkFormula;
	}
	
	/**
	 * Returns true iff the solver found a satisfying trace.
	 */
	@Returns("this.trace != null")
	public boolean satisfiable() {
		return trace != null;
	}
	
	/**
	 * Returns the satisfying trace if unsatisfiable; otherwise null.
	 */
	@Returns("this.trace")
	public Trace trace() {
		return trace;
	}
	
	/** 
	 * Return the statements covered by the analysis if unsatisfiable;
	 * otherwise null.
	 */
	@Returns("this.coverage")
	public Coverage coverage() {
		return coverage;
	}
	
	private static final String NEW_LINE = System.getProperty("line.separator");
	
	public String toString() {
		StringBuilder sb = new StringBuilder();
		if (satisfiable()) {
			sb.append("trace found").append(NEW_LINE).append(trace);
		} else if (coverage == null) {
			sb.append("no trace found");
		} else {
			sb.append("no trace found and coverage measured");
			sb.append(NEW_LINE).append(coverage);
		}
		return sb.toString();
	}
}
