package forge.solve;

import java.util.Iterator;
import java.util.NoSuchElementException;

import kodkod.engine.Solution;
import edu.mit.csail.sdg.annotations.Returns;
import forge.cfg.ForgeCFG;
import forge.program.ForgeProcedure;
import forge.program.ForgeProgram;
import forge.translate.ForgeTranslation;

/**
 * Binds instance literals to atoms.
 * All bindings belong to a ConstantScene.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class ForgeSolver {

	private final ForgeCFG code, transformed;
	private final SolveOptions options;
	private final ForgeTranslation transl;
	
	/** derived. */
	private final ForgeProcedure proc;
	
	/** lazily initialized. */
	private ForgeCFG.Spec trueSpec = null;
	
	/**
	 * Constructs a solver for analyzing the given code, where loops and
	 * recursion are unrolled the specified number of times.
	 */
	public ForgeSolver(ForgeCFG cfg, int unrolls) {
		this(cfg, new SolveOptions.Builder(unrolls).build());
	}
	
	/**
	 * Constructs a solver for analyzing the given code with the given options.
	 */
	public ForgeSolver(ForgeCFG code, final SolveOptions options) {
		if (code == null) throw new NullPointerException("code");
		if (options == null) throw new NullPointerException("options");
		this.code = code;
		this.proc = code.procedure();
		this.options = options;
		
		// perform the transformation
		options.reporter().transformingProcedure();
		transformed = options.transformer().transform(code);
//		System.out.println("transformed cfg:");
//		System.out.println(transformed);
//		new BreadthFirstVisitor() {
//			private void print(CFGStmt stmt) {
//				System.out.println();
//				System.out.println("TARGET = " + stmt);
//				System.out.println("SOURCE = " + options.transformer().sourceStmt(stmt));
//			}
//			public void visit(AssignStmt stmt) { print(stmt); }
//			public void visit(BranchStmt stmt) { print(stmt); }
//			public void visit(CallStmt stmt)   { print(stmt); }
//			public void visit(CreateStmt stmt) { print(stmt); }
//			public void visit(ExitStmt stmt)   { print(stmt); }
//			public void visit(SpecStmt stmt)   { print(stmt); }
//		}.traverse(transformed);
		
		// perform the translation
		options.reporter().translatingToKodkod();
		transl = ForgeTranslation.translate(options.strategy(),
				transformed, options.coverage());
	}
	
	@Returns("this.code.program")
	public ForgeProgram program() {
		return code.program();
	}
	
	@Returns("this.code")
	public ForgeCFG code() {
		return code;
	}
	
	@Returns("this.options")
	public SolveOptions options() {
		return options;
	}
	
	/**
	 * Finds a trace of the procedure within the given bounds.
	 */
	public ForgeSolution run(ForgeBounds bounds) {
		if (trueSpec == null) {
			trueSpec = ForgeCFG.specification(proc,
					program().globalVariables(), program().trueLiteral());
		}
		return run(trueSpec, bounds);
	}
	
	/**
	 * Finds a trace of the procedure within the given bounds that refines
	 * the given specification statement, determining coverage information
	 * in the case where no trace can be found if coverage flag is set to true.
	 * The selected solver must have proving capability if coverage is true. 
	 */
	public ForgeSolution run(ForgeCFG.Spec spec, ForgeBounds bounds) {
		return solve(spec, bounds, false);
	}
	
	/**
	 * Finds a trace of the procedure within the given bounds that does not
	 * refine the given specification statement, determining coverage information
	 * in the case where no trace can be found if coverage flag is set to true.
	 * The selected solver must have proving capability if coverage is true. 
	 */
	public ForgeSolution check(ForgeCFG.Spec spec, ForgeBounds bounds) {
		return solve(spec, bounds, true);
	}
	
	public Iterator<ForgeSolution> runAll(ForgeCFG.Spec spec, ForgeBounds bounds) {
		return solveAll(spec, bounds, false);
	}
	
	public Iterator<ForgeSolution> checkAll(ForgeCFG.Spec spec, ForgeBounds bounds) {
		return solveAll(spec, bounds, true);
	}
	
	private ForgeSolution solve(ForgeCFG.Spec spec, ForgeBounds bounds, boolean check) {
		Analysis analysis = analysis(spec, bounds, check);
		options.reporter().solvingAnalysis();
		Solution kkSol = analysis.kkSolver.solve(
				analysis.kkFormula, analysis.kkBounds);
		return solution(analysis, kkSol);
	}
	
	private Iterator<ForgeSolution> solveAll(ForgeCFG.Spec spec,
			ForgeBounds bounds, boolean check) {
		Analysis analysis = analysis(spec, bounds, check);
		Iterator<Solution> kkSols = analysis.kkSolver.solveAll(
				analysis.kkFormula, analysis.kkBounds);
		return new SolutionIter(analysis, kkSols);
	}
	
	private Analysis analysis(ForgeCFG.Spec spec, ForgeBounds bounds, boolean check) {
		program().checkProgram(spec);
		program().checkProgram(bounds);
		return new Analysis(code, transformed, transl, options, spec, bounds, check);
	}
	
	/**
	 * If satisfiable, translates back to a trace of the code;
	 * Otherwise, if there's a proof, translate back to coverage.
	 */
	private ForgeSolution solution(Analysis analysis, Solution kkSol) {
		Trace trace = null;
		Coverage coverage = null;
		if (kkSol.instance() != null) {
			trace = Interpreter.interpret(analysis, kkSol.instance());
		} else if (options.coverage()) {
			coverage = CoverageMetric.measure(analysis, kkSol.proof());
		}
		return new ForgeSolution(analysis, trace, coverage);
	}
	
	private final class SolutionIter implements Iterator<ForgeSolution> {
		
		private final Analysis analysis;
		private final Iterator<Solution> kkSolIter;
		
		private SolutionIter(Analysis analysis, Iterator<Solution> kkSolIter) {
			this.analysis = analysis;
			this.kkSolIter = kkSolIter;
		}
		
		public boolean hasNext() {
			return kkSolIter.hasNext();
		}
		
		public ForgeSolution next() {
			if (!kkSolIter.hasNext()) throw new NoSuchElementException();
			options.reporter().solvingAnalysis();
			return solution(analysis, kkSolIter.next());
		}
		
		public void remove() {
			throw new UnsupportedOperationException();
		}
	}
	
//	private static void printFormula(Formula formula) {
//	BufferedWriter writer = null;
//	
//	try {
//		writer = new BufferedWriter(new FileWriter("formula.txt"));
//		printFormula(formula, writer);
//	} catch (IOException e) {
//		System.out.println(e);
//	} finally {
//		if (writer != null) {
//			try { writer.close(); }
//			catch (IOException ignore) {}
//		}
//	}
//}
//
//private static void printFormula(Formula formula, BufferedWriter w)
//		throws IOException {
//	if (formula instanceof BinaryFormula) {
//		BinaryFormula bf = (BinaryFormula)formula;
//		if (bf.op().equals(BinaryFormula.Operator.AND)) {
//			printFormula(bf.left(), w);
//			printFormula(bf.right(), w);
//			return;
//		}
//	}
//	
//	w.write(formula.toString());
//	w.newLine();
//}
	
}
