package forge.solve;

import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import kodkod.ast.Expression;
import kodkod.engine.Evaluator;
import kodkod.instance.Instance;
import edu.mit.csail.sdg.util.collections.Iterators;
import edu.mit.csail.sdg.util.collections.Pair;
import forge.cfg.AssignStmt;
import forge.cfg.BranchStmt;
import forge.cfg.CFGStmt;
import forge.cfg.CFGVisitor;
import forge.cfg.CallStmt;
import forge.cfg.CreateStmt;
import forge.cfg.ExitStmt;
import forge.cfg.ForgeCFG;
import forge.cfg.SpecStmt;
import forge.cfg.UpdateStmt;
import forge.program.ForgeExpression;
import forge.program.ForgeVariable;
import forge.program.GlobalVariable;
import forge.program.LocalVariable;
import forge.program.ForgeExpression.Modifiable;
import forge.transform.Transformer;
import forge.translate.Environment;
import forge.translate.ExprTranslator;
import forge.translate.SymbolicExecutor;
import forge.translate.SEStrategy;

/**
 * Builds a trace based on a Kodkod instance.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
final class Interpreter {
	
	static Trace interpret(Analysis analysis, Instance kkInstance) {
		Environment initEnv = analysis.transl.initialEnv();
		Transformer xformer = analysis.options.transformer();
		SEStrategy strategy = analysis.options.strategy();
		ForgeBounds bounds  = analysis.bounds;
		Evaluator kkEval = new Evaluator(kkInstance, analysis.kkSolver.options());
		
		// construct interpreter and trace;
		Interpreter interpret = new Interpreter(kkEval, bounds, strategy, initEnv.copy(), xformer);
		interpret.xformVisitor.targetNext = analysis.xform.entry();
		return interpret.trace(analysis.code, initEnv, initEnv.copy());
	}
	
	private final Evaluator kkEval;
	private final ForgeBounds bounds;
	private final XformVisitor xformVisitor;
	
	private Interpreter(Evaluator kkEval, ForgeBounds bounds,
			SEStrategy strategy, Environment env, Transformer xformer) {
		this.kkEval = kkEval;
		this.bounds = bounds;
		xformVisitor = new XformVisitor(strategy, env, xformer);
	}
	
	private Trace trace(ForgeCFG code, Environment initEnv, Environment finlEnv) {
		// perform the trace
		CodeVisitor sourceVisitor = new CodeVisitor(finlEnv);
		Binding initBind = sourceVisitor.evaluateEnv();
		code.entry().accept(sourceVisitor);
		Binding finlBind = sourceVisitor.evaluateEnv();
		List<Step> steps = Collections.unmodifiableList(sourceVisitor.steps);
		
		// construct the spec evaluator
		Environment specEnv = initEnv.copy();
		ExprTranslator specTransl = new ExprTranslator(specEnv);
		for (GlobalVariable var : code.program().globalVariables()) {
			specTransl.pinOldExpr(var);
			specEnv.putExpr(var, finlEnv.getExpr(var));
		}
		for (LocalVariable var : code.procedure().outParams().locals()) {
			specEnv.putExpr(var, finlEnv.getExpr(var));
		}
		ForgeEvaluator specEval = new ForgeEvaluator(kkEval, bounds, specTransl);
		
		// return the trace
		return new Trace(code, initBind, finlBind, steps, specEval);
	}
	
	/**
	 * Visits the source CFG.
	 */
	private final class CodeVisitor implements CFGVisitor {
		
		private final ExprTranslator transl;
		private final ForgeEvaluator eval;
		private final List<Step> steps = new LinkedList<Step>();
		
		private CodeVisitor(Environment env) {
			transl = new ExprTranslator(env);
			eval = new ForgeEvaluator(kkEval, bounds, transl);
		}
		
		public void visit(BranchStmt stmt) {
			boolean truth = eval.evaluateBool(stmt.condition());
			steps.add(new Step.Branch(stmt, truth));
			if (truth) stmt.getThen().accept(this);
			else       stmt.getElse().accept(this);
		}

		public void visit(AssignStmt stmt) {
			makeUpdateStep(stmt);
		}
		
		public void visit(SpecStmt stmt) {
			makeUpdateStep(stmt);
		}
		
		public void visit(CreateStmt stmt) {
			makeUpdateStep(stmt);
		}
		
		private void makeUpdateStep(UpdateStmt update) {
			// translate until reach target for sourceStop
			xformVisitor.sourceStop = update;
			xformVisitor.targetNext.accept(xformVisitor);
			UpdateStmt target = xformVisitor.targetStop;
			assert target != null : "sourceStop = " + update;
			
			// set modifiables to expressions for corresponding modifiables
			for (Pair<? extends Modifiable, ? extends Modifiable> p :
					Iterators.pair(update.modified(), target.modified())) {
				Modifiable sourceMod = p.first();
				Modifiable targetMod = p.second();
				Expression targetExpr = xformVisitor.se.toExpr(targetMod);
				transl.env().putExpr(sourceMod, targetExpr);
			}
			
			// evaluate the environment
			Binding post = evaluateEnv();
			Step.Update updateStep = new Step.NonCall(update, post);
			steps.add(updateStep);
			update.getNext().accept(this);
		}
		
		public void visit(CallStmt stmt) {
			// create an environment for the call that contains exprs for
			// all instance domains and global variables
			Environment currEnv = transl.env();
			Environment callInitEnv = new Environment(transl.model());
			for (Modifiable mod : currEnv.mods()) {
				if (mod.isGlobal()) {
					callInitEnv.putExpr(mod, currEnv.getExpr(mod));
				}
			}
			
			// set the parameters to the argument expressions in the call env
			ForgeCFG callCFG = stmt.called();
			Iterator<ForgeExpression> inIter = stmt.inArgs().iterator();
			for (LocalVariable inParam : callCFG.procedure().inParams().locals()) {
				Expression inArg = transl.toExpr(inIter.next());
				callInitEnv.putExpr(inParam, inArg);
			}
			
			// trace the call
			Environment callFinlEnv = callInitEnv.copy();
			Trace callTrace = trace(callCFG, callInitEnv, callFinlEnv);
			
			// set the output arguments
			Iterator<ForgeVariable> outIter = stmt.outArgs().iterator();
			for (LocalVariable outParam : callCFG.procedure().outParams().locals()) {
				Expression outExpr = callFinlEnv.getExpr(outParam);
				currEnv.putExpr(outIter.next(), outExpr);
			}
			
			// set the global modifiables
			for (Modifiable mod : currEnv.mods()) {
				if (mod.isGlobal()) {
					currEnv.putExpr(mod, callFinlEnv.getExpr(mod));
				}
			}
			
			Binding post = evaluateEnv();
			Step.Call callStep = new Step.Call(stmt, post, callTrace);
			steps.add(callStep);
			stmt.getNext().accept(this);
		}
		
		private Binding evaluateEnv() {
			Binding binding = new Binding(bounds);
			for (Modifiable mod : transl.env().mods()) {
				binding.putValue(mod, eval.evaluate(mod));
			}
			return binding;
		}

		public void visit(ExitStmt stmt) {}
	}
	
	/**
	 * Visits the transformed CFG.
	 */
	private final class XformVisitor implements CFGVisitor {
		
		private final SymbolicExecutor se;
		private final ForgeEvaluator eval;
		private final Transformer xformer;
		
		// source statement through which this should translate
		// and corresponding target stmt in target cfg
		private UpdateStmt sourceStop = null, targetStop = null;
		
		// where the target translation should pick up next
		private CFGStmt targetNext;
		
		private XformVisitor(SEStrategy strategy, Environment env, Transformer xformer) {
			se = new SymbolicExecutor(strategy, env, null);
			eval = new ForgeEvaluator(kkEval, bounds, se);
			this.xformer = xformer;
		}
		
		public void visit(BranchStmt stmt) {
			boolean truth = eval.evaluateBool(stmt.condition());
			if (truth) stmt.getThen().accept(this);
			else       stmt.getElse().accept(this);
		}

		public void visit(AssignStmt stmt) {
			se.execute(stmt);
			makeUpdateStep(stmt);
		}
		
		public void visit(CreateStmt stmt) {
			se.execute(stmt);
			makeUpdateStep(stmt);
		}
		
		public void visit(SpecStmt stmt) {
			se.execute(stmt);
			makeUpdateStep(stmt);
		}
		
		public void visit(CallStmt stmt) {
			assert false : " all calls must have been inlined ";
		}
		
		private void makeUpdateStep(UpdateStmt update) {
			CFGStmt source = xformer.sourceStmt(update);
			targetNext = update.getNext();
			
			// if stopping, record stop point
			// else keep going
			if (sourceStop.equals(source)) {
				targetStop = update;
			} else {
				targetNext.accept(this);
			}
		}
		
		public void visit(ExitStmt stmt) {}
	}
}
