package forge.solve;

import java.util.ArrayList;
import java.util.List;

import kodkod.engine.satlab.SATFactory;
import edu.mit.csail.sdg.annotations.Returns;
import forge.transform.BatchTransformer;
import forge.transform.InlineTransformer;
import forge.transform.Transformer;
import forge.transform.UnrollTransformer;
import forge.translate.InlineStrategy;
import forge.translate.SEStrategy;

/**
 * Forge solving options. Uses the builder pattern.
 * 
 * @specfield transformer - applied to the procedure before analsis
 * @specfield reporter - reported to by the analysis
 * @specfield satSolver - sat solver used by the analysis
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class SolveOptions {
	
	/** Default transformer does nothing. */
	private static final Transformer DEFAULT_TRANSFORMER = Transformer.IDENTITY;
	
	/** Default reporter does nothing. */
	private static final ForgeReporter DEFAULT_REPORTER = new ForgeReporter() {};
	
	/** Default SAT solver is Java-based SAT. */
	private static final SatSolver DEFAULT_SATSOLVER = SatSolver.SAT4J;
	
	/** Default symbolic execution strategy. */
	private static final SEStrategy DEFAULT_STRATEGY = InlineStrategy.STRATEGY;
	
	/** Enumeration of possible SAT Solvers. */
	public static enum SatSolver {
		/** Pure Java, cannot prove. */
		SAT4J(SATFactory.DefaultSAT4J),
		/** Pure Java, cannot prove. */
		SAT4J_LIGHT(SATFactory.LightSAT4J),
		/** Platform-specific, cannot prove. */
		MINISAT(SATFactory.MiniSat),
		/** Platform-specific, prover. */
		MINISAT_PROVER(SATFactory.MiniSatProver),
		/** Platform-specific, cannot prove. */
		ZCHAFF(SATFactory.ZChaff),
		/** Platform-specific, cannot prove. */
		ZCHAFF_MINCOST(SATFactory.ZChaffMincost);
		
		final SATFactory kkSolver;
		
		private SatSolver(SATFactory kkSolver){
			this.kkSolver = kkSolver;
		}
		
		/** Returns true if solver can provide a proof. */
		public boolean prover() {
			return kkSolver.prover();
		}
	}
	
	/**
	 * Builder for SolveOptions
	 */
	public static class Builder {
		private Transformer transformer;
		private ForgeReporter reporter;
		private SatSolver satSolver;
		private SEStrategy strategy;
		
		public Builder(int unrolls) {
			this();
			List<Transformer> transformers = new ArrayList<Transformer>();
			transformers.add(new UnrollTransformer(unrolls));
			transformers.add(new InlineTransformer(unrolls));
			transformer = new BatchTransformer(transformers);
		}
		
		public Builder() {
			transformer = DEFAULT_TRANSFORMER;
			reporter    = DEFAULT_REPORTER;
			satSolver   = DEFAULT_SATSOLVER;
			strategy    = DEFAULT_STRATEGY;
		}
		
		/** Builds the SolveOptions. */
		public SolveOptions build() {
			return new SolveOptions(this);
		}
		
		/** Choose the transformer. */
		public Builder transformer(Transformer transformer) {
			if (transformer == null) throw new NullPointerException();
			this.transformer = transformer;
			return this;
		}
		
		/** Chooses the reporter. */
		public Builder reporter(ForgeReporter reporter) {
			if (reporter == null) throw new NullPointerException();
			this.reporter = reporter;
			return this;
		}
		
		/** Chooses the SAT solver. */
		public Builder satSolver(SatSolver satSolver) {
			if (satSolver == null) throw new NullPointerException();
			this.satSolver = satSolver;
			return this;
		}
		
		/** Choose the symbolic execution strategy. */
		public Builder strategy(SEStrategy strategy) {
			if (strategy == null) throw new NullPointerException();
			this.strategy = strategy;
			return this;
		}
	}
	
	private final Transformer transformer;
	private final ForgeReporter reporter;
	private final SatSolver satSolver;
	private final SEStrategy strategy;
	
	/**
	 * Constructs solver options with the given Builder.
	 */
	private SolveOptions(Builder builder) {
		transformer = builder.transformer;
		reporter    = builder.reporter;
		satSolver   = builder.satSolver;
		strategy    = builder.strategy;
	}
	
	@Returns("this.transformer")
	public Transformer transformer() {
		return transformer;
	}
	
	@Returns("this.reporter")
	public ForgeReporter reporter() {
		return reporter;
	}
	
	@Returns("this.satSolver")
	public SatSolver satSolver() {
		return satSolver;
	}
	
	@Returns("this.strategy")
	public SEStrategy strategy() {
		return strategy;
	}
	
	@Returns("this.satSolver.prover")
	public boolean coverage() {
		return satSolver.prover();
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("transformer: " + transformer + ", ");
		sb.append("reporter: "    + reporter    + ", ");
		sb.append("SAT solver: "  + satSolver   + ", ");
		sb.append("strategy: "    + strategy);
		return sb.toString();
	}
}
