package forge.solve;

import edu.mit.csail.sdg.annotations.Returns;
import edu.mit.csail.sdg.annotations.SpecField;
import edu.mit.csail.sdg.annotations.Throws;
import forge.cfg.BranchStmt;
import forge.cfg.CFGElement;
import forge.cfg.CFGStmt;
import forge.cfg.CallStmt;
import forge.cfg.ForgeCFG;
import forge.cfg.UpdateStmt;
import forge.program.ForgeProgram;
import forge.program.ForgeExpression.Modifiable;

/**
 * A step in a trace. There are exactly two types of steps: Update and Branch.
 * If a step is a update step, then <code>isBranch</code> returns false and the
 * return type of <code>stmt</code> is <code>UpdateStmt</code>. If a step is a
 * branch step, then <code>isBranch</code> returns true and the return type of
 * <code>stmt</code> is BranchStmt.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
@SpecField({"isBranch : boolean", "stmt : CFGStmt"})
public abstract class Step implements CFGElement {
	
	private static final String NEW_LINE = System.getProperty("line.separator");
	private static final String INDENT = "  ";
	
	private final ForgeCFG cfg;
	
	private Step(CFGStmt stmt) {
		cfg = stmt.cfg();
	}
	
	/* inherits specification */
	public final ForgeProgram program() {
		return cfg.program();
	}
	
	/* inherits specification */
	public final ForgeCFG cfg() {
		return cfg;
	}
	
	@Returns("this.isBranch")
	public abstract boolean isBranch();
	
	@Returns("this.stmt")
	public abstract CFGStmt stmt();

	/**
	 * An update step: an UpdateStmt and an effect.
	 */
	@SpecField({"isBranch : boolean | isBranch = false",
		        "stmt : CFGStmt | stmt = update",
		        "isCall : boolean"})
	public abstract static class Update extends Step {
		
		private final UpdateStmt update;
		private final Binding post;
		
		private Update(UpdateStmt update, Binding effect) {
			super(update);
			program().checkProgram(effect);
			this.update = update;
			this.post = effect;
		}

		/* inherits specification */
		public final boolean isBranch() {
			return false;
		}
		
		/* inherits specification */
		public UpdateStmt stmt() {
			return update;
		}
		
		@Returns("this.post")
		public final Binding post() {
			return post;
		}
		
		@Returns("this.isCall")
		public abstract boolean isCall();
		
		public String toString() {
			StringBuilder sb = new StringBuilder();
			sb.append(update).append(NEW_LINE);
			for (Modifiable mod : post.bound()) {
				sb.append(INDENT).append(mod).append(" = ");
				sb.append(post.getValue(mod)).append(NEW_LINE);
			}
			return sb.toString();
		}
	}
	
	static final class NonCall extends Update {
		
		NonCall(UpdateStmt update, Binding effect) {
			super(checkNotCall(update), effect);
		}
		
		/* inherits specification */
		public boolean isCall() {
			return false;
		}
		
		@Returns("update")
		@Throws("IllegalArgumentException :!(update in CallStmt)")
		private static UpdateStmt checkNotCall(UpdateStmt update) {
			if (update instanceof CallStmt) throw new IllegalArgumentException(
					"cannot construct NonCall with CallStmt");
			return update;
		}
	}
	
	/**
	 * A call step: an CallStmt with its own trace.
	 */
	public final static class Call extends Update {
		
		private final Trace trace;
		
		Call(CallStmt call, Binding effect, Trace trace) {
			super(call, effect);
			this.trace = trace;
		}

		@Override
		/* inherits specification */
		public boolean isCall() {
			return true;
		}

		@Override
		/* inherits specification */
		public CallStmt stmt() {
			return (CallStmt)super.stmt();
		}
		
		@Returns("this.trace")
		public Trace trace() {
			return trace;
		}

		@Override
		public String toString() {
			StringBuilder sb = new StringBuilder();
			sb.append(stmt()).append(NEW_LINE);
			for (Modifiable mod : post().bound()) {
				sb.append(INDENT).append(mod).append(" = ");
				sb.append(post().getValue(mod)).append(NEW_LINE);
			}
			return sb.toString();
		}
	}
	
	/**
	 * A branch step: a BranchStmt and a truth value.
	 */
	public final static class Branch extends Step {
		
		private final BranchStmt branch;
		private final boolean truth;
		
		Branch(BranchStmt branch, boolean truth) {
			super(branch);
			this.branch = branch;
			this.truth = truth;
		}

		/* inherits specification */
		public boolean isBranch() {
			return true;
		}

		/* inherits specification */
		public BranchStmt stmt() {
			return branch;
		}
		
		@Returns("this.truth")
		public boolean truth() {
			return truth;
		}
		
		public String toString() {
			return branch + NEW_LINE + INDENT + truth;
		}
	}
}