package forge.solve;

import java.util.Collection;
import java.util.List;

import edu.mit.csail.sdg.annotations.Returns;
import forge.cfg.CFGElement;
import forge.cfg.ForgeCFG;
import forge.program.ForgeExpression;
import forge.program.ForgeProcedure;
import forge.program.ForgeProgram;
import forge.program.LocalVariable;
import forge.program.ForgeExpression.Modifiable;

/**
 * A trace of a Forge procedure, a series of steps.
 *
 * @specfield steps: seq Step
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class Trace implements CFGElement {

	private final ForgeCFG cfg;
	private final Binding initBind, finlBind;
	private final List<Step> steps;
	private final ForgeEvaluator eval;
	
	Trace(ForgeCFG cfg, Binding initBind, Binding finlBind,
			List<Step> steps, ForgeEvaluator eval) {
		this.cfg = cfg;
		this.initBind = initBind;
		this.finlBind = finlBind;
		this.steps = steps;
		this.eval = eval;
	}
	
	/**
	 * Returns the Kodkod instance from which this trace was created.
	 */
	@Returns("this.kkEval.instance")
	public Object kkInstance() {
		return eval.kkEval.instance();
	}
	
	/* inherits specification */
	public ForgeProgram program() {
		return cfg.program();
	}
	
	/* inherits specification */
	public ForgeCFG cfg() {
		return cfg;
	}
	
	@Returns("this.initialBind")
	public Binding initialBind() {
		return initBind;
	}
	
	@Returns("this.finlBind")
	public Binding finalBind() {
		return finlBind;
	}
	
	@Returns("this.steps")
	public List<Step> steps() {
		return steps;
	}
	
	public ForgeConstant evaluate(ForgeExpression expr) {
		return eval.evaluate(expr);
	}
	
	public boolean evaluateBool(ForgeExpression expr) {
		return eval.evaluateBool(expr);
	}
	
	public int evaluateInt(ForgeExpression expr) {
		return eval.evaluateInt(expr);
	}
	
	private static final String NEW_LINE = System.getProperty("line.separator");
	private static final String INDENT = "  ";
	
	public String toString() {
		ForgeProcedure proc = cfg.procedure();
		StringBuilder sb = new StringBuilder();
		
		// print the initial state
		List<LocalVariable> inLocals = proc.inParams().locals();
		Binding emptyBind = new Binding(eval.bounds);
		sb.append("initial state:").append(NEW_LINE);
		appendBind(initBind, inLocals, emptyBind, INDENT, sb);
		
		// print the steps
		appendSteps(steps, initBind, "", sb);
		
		// print the final state
		List<LocalVariable> outLocals = cfg.procedure().outParams().locals();
		sb.append("final state:").append(NEW_LINE);
		appendBind(finlBind, outLocals, initBind, INDENT, sb);
		
		// return the string
		return sb.toString();
	}
	
	private static void appendSteps(List<Step> steps, Binding prev,
			String prefix, StringBuilder sb) {
		String prefixNext = prefix + INDENT;
		
		for (Step step : steps) {
			sb.append(prefix).append(step.stmt()).append(NEW_LINE);
			
			if (step.isBranch()) {
				Step.Branch branch = (Step.Branch)step;
				sb.append(prefixNext).append(branch.truth()).append(NEW_LINE);
			} else {
				Step.Update stmt = (Step.Update)step;
				if (stmt.isCall()) {
					Step.Call call = (Step.Call)stmt;
					Trace trace = call.trace();
					List<LocalVariable> inParams = call.stmt().called().procedure().inParams().locals();
					String prefixCall = prefixNext + "| ";
					appendBind(trace.initBind, inParams, prev, prefixCall, sb);
					appendSteps(trace.steps(), trace.initBind, prefixCall, sb);
				}
				
				Binding bind = stmt.post();
				appendBind(bind, stmt.stmt().modified(), prev, prefixNext, sb);
				prev = bind;
			}
		}
	}
	
	/**
	 * Appends all the mod-value pairs for modifiables in must, followed by all
	 * the pairs in bind that are different from those in prev
	 */
	private static void appendBind(Binding bind, Collection<? extends Modifiable> locals,
			Binding prev, String prefix, StringBuilder sb) {
		for (Modifiable mod : locals) {
			if (!mod.isGlobal()) {
				appendPair(mod, bind.getValue(mod), prefix, sb);
			}
		}
		for (Modifiable mod : bind.bound()) {
			if (mod.isGlobal()) {
				ForgeConstant value = bind.getValue(mod);
				if (value != prev.getValue(mod)) {
					appendPair(mod, value, prefix, sb);
				}
			}
		}
	}
	
	private static void appendPair(Modifiable mod, ForgeConstant value,
			String prefix, StringBuilder sb) {
		sb.append(prefix).append(mod).append(" = ");
		sb.append(value).append(NEW_LINE);
	}
	
}