package forge.tests;

import java.util.Collections;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import forge.cfg.ForgeCFG;
import forge.program.ForgeExpression;
import forge.program.GlobalVariable;
import forge.program.InstanceDomain;
import forge.solve.ForgeBounds;
import forge.solve.ForgeSolution;
import forge.solve.ForgeSolver;
import forge.solve.SolveOptions;
import forge.translate.ConstrainStrategy;
import forge.translate.HybridStrategy;
import forge.translate.InlineStrategy;
import forge.translate.SEStrategy;

/**
 * A generic Forge Test to be extended
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public abstract class ForgeTest {
	
	protected final static Map<InstanceDomain, Integer> NO_SCOPES = Collections.emptyMap();
	protected final static Set<GlobalVariable> NO_GLOBALS = Collections.emptySet();
	
	protected final static SEStrategy INLINE_STRATEGY    = InlineStrategy.STRATEGY;
	protected final static SEStrategy CONSTRAIN_STRATEGY = ConstrainStrategy.STRATEGY;
	protected final static SEStrategy HYBRID_STRATEGY    = HybridStrategy.STRATEGY;

	private final String name;
	private final ForgeCFG code;
	private final ForgeExpression preCond, postCond;
	private final Set<GlobalVariable> modGlobals;
	private final ForgeCFG.Spec checkSpec, runSpec;
	
	private SolveOptions.Builder optionsBuilder = null;
	private ForgeBounds bounds = null;
	
	protected ForgeTest(String name, ForgeCFG code, ForgeExpression preCond,
			ForgeExpression postCond, Set<GlobalVariable> modGlobals) {
		this.name = name;
		this.code = code;
		this.preCond = preCond;
		this.postCond = postCond;
		this.modGlobals = modGlobals;
		checkSpec = ForgeCFG.specification(code.procedure(), modGlobals,
				preCond.implies(postCond));
		runSpec = ForgeCFG.specification(code.procedure(), modGlobals,
				preCond.and(postCond));
	}
	
	public final String name() {
		return name;
	}
	
	public final ForgeCFG code() {
		return code;
	}
	
	public final ForgeCFG.Spec checkSpec() {
		return checkSpec;
	}
	
	public final ForgeCFG.Spec runSpec() {
		return runSpec;
	}
	
	public final SolveOptions.Builder newOptions(int unrolls) {
		optionsBuilder = new SolveOptions.Builder(unrolls);
		return optionsBuilder;
	}
	
	public final SolveOptions.Builder currOptions() {
		return optionsBuilder;
	}
	
	public final void setBounds(int bitWidth,
			Map<InstanceDomain, Integer> scopes) {
		bounds = new ForgeBounds(code.program(), bitWidth, scopes);
	}
	
	public final ForgeSolution checkTest() {
		return makeSolver().check(checkSpec, bounds);
	}
	
	public final ForgeSolution runTest() {
		return makeSolver().run(runSpec, bounds);
	}
	
	public final Iterator<ForgeSolution> checkAllTest() {
		return makeSolver().checkAll(checkSpec, bounds);
	}
	
	public final Iterator<ForgeSolution> runAllTest() {
		return makeSolver().runAll(runSpec, bounds);
	}
	
	private ForgeSolver makeSolver() {
		if (optionsBuilder == null || bounds == null) {
			throw new IllegalStateException();
		}
		return new ForgeSolver(code, optionsBuilder.build());
	}
	
	public final void standardTest() {
		ForgeSolver solver = makeSolver();
		System.out.println("===== TESTING THE FOLLOWING CODE =====");
		System.out.println(code);
		
		// check the cond
		System.out.println("===== CHECKING AGAINST THE FOLLOWING SPEC =====");
		System.out.println(checkSpec);
		long checkStart = System.currentTimeMillis();
		ForgeSolution checkSol = solver.check(checkSpec, bounds);
		long checkDuration = System.currentTimeMillis() - checkStart;
		printSolution(checkSol, checkDuration);
		
		// run the code
		System.out.println("===== RUNNING AGAINST THE FOLLOWING SPEC =====");
		System.out.println(runSpec);
		long runStart = System.currentTimeMillis();
		ForgeSolution runSol = solver.run(runSpec, bounds);
		long runDuration = System.currentTimeMillis() - runStart;
		printSolution(runSol, runDuration);
		
		System.out.println("===== TEST COMPLETED =====");
	}
	
	private static final String NEW_LINE = System.getProperty("line.separator");
	
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append(name()).append(" test").append(NEW_LINE);
		sb.append("code: ").append(code);
		sb.append("pre condition: ").append(preCond).append(NEW_LINE);
		sb.append("post condition: ").append(postCond).append(NEW_LINE);
		sb.append("mod globals: ").append(modGlobals);
		return sb.toString();
	}
	
	private static void printSolution(ForgeSolution sol, long duration) {
		long seconds = (long)(duration / 1000.0);
		System.out.println("analysis completed in " + seconds + " seconds");
		System.out.println(sol);
	}
}
