package forge.tests;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import forge.cfg.AssignStmt;
import forge.cfg.BranchStmt;
import forge.cfg.CallStmt;
import forge.cfg.ForgeCFG;
import forge.program.ForgeExpression;
import forge.program.ForgeProcedure;
import forge.program.ForgeProgram;
import forge.program.ForgeVariable;
import forge.program.GlobalVariable;
import forge.program.InstanceDomain;
import forge.program.LocalVariable;
import forge.solve.SolveOptions.SatSolver;

/**
 * Constructs a FIR program consisting of a list data structure and two
 * implementations of the contains method, one that contains a loop and
 * the other that contains a recursive call.
 */
public abstract class ListTest extends ForgeTest {
	
	private static final int LIST_SIZE   = 3;
	private static final int BIT_WIDTH   = 3;
	private static final int NUM_UNROLLS = 3;
	
	/**
	 * Constructs the FIR and checks both of the contains procedures for
	 * any counterexamples within the bounds described above.
	 */
	public static void main(String[] args) {
		Map<InstanceDomain, Integer> listScope = Collections.singletonMap(list, LIST_SIZE);
		
		ForgeTest loopTest = new Looping();
		loopTest.newOptions(NUM_UNROLLS).satSolver(SatSolver.MINISAT_PROVER);
		loopTest.setBounds(BIT_WIDTH, listScope);
		loopTest.standardTest();
		
		System.out.println();
		
		ForgeTest recTest = new Recursive();
		recTest.newOptions(NUM_UNROLLS).satSolver(SatSolver.MINISAT_PROVER);
		recTest.setBounds(BIT_WIDTH, listScope);
		recTest.standardTest();
	}
	
	private ListTest(String name, ForgeCFG code) {
		super(name, code, preCond, postCond, NO_GLOBALS);
	}
	
	public InstanceDomain listDomain() {
		return list;
	}
	
	private static final ForgeProgram program;
	private static final InstanceDomain list;
	private static final GlobalVariable next, elem;
	private static final LocalVariable lst, val, ret;
	private static final ForgeProcedure contains;
	private static final ForgeExpression preCond, postCond;
	
	static {
		program = new ForgeProgram();
		list = program.newInstanceDomain("List");
		next = program.newGlobalVariable("next", list.product(list));
		elem = program.newGlobalVariable("elem", list.product(program.integerDomain()));
		lst = program.newLocalVariable("lst", list);
		val = program.newLocalVariable("val", program.integerDomain());
		ret = program.newLocalVariable("return", program.booleanDomain());
		contains = program.newProcedure("contains", lst.and(val), ret);
		ForgeExpression nexts = next.closure();
		
		// make a constraint for valid arguments and fields
		// the list variable is lone (singleton element or the empty set),
		// and the value we are searching for must be a singleton
		// next is acyclic and a partial function
		// elem is a total function
		ForgeExpression validArgs = lst.lone().and(val.one());
		LocalVariable x = program.newLocalVariable("x", list);
		ForgeExpression acyclic = x.in(x.join(nexts)).not();
		ForgeExpression fieldFunctions = x.join(next).lone().and(x.join(elem).one());
		ForgeExpression validFields = acyclic.and(fieldFunctions).forAll(x);
		preCond = validArgs.and(validFields);
		
		// reachable is the set of all reachable lists from lst, including lst itself
		// correctReturn is the constraint that the return variable is true iff
		// given val is in the elem field of the reachable lists.
		ForgeExpression reachable = lst.join(nexts).union(lst);
		postCond = ret.eq(val.in(reachable.join(elem)));
	}
	
	public static final class Looping extends ListTest {
		
		private static final ForgeCFG.Impl code;
		
		static {
			code = ForgeCFG.implementation(contains);
			
			// construct the statements in the procedure.
			// initially, the exit stmt is the only each successor each stmt
			// the implementation presumes the lst variable can be the empty set,
			// which we will use to denote the empty list
			BranchStmt isEmpty   = code.newBranch(lst.no());
			AssignStmt notFound  = code.newAssign(ret, program.falseLiteral());
			BranchStmt checkElem = code.newBranch(lst.join(elem).eq(val));
			AssignStmt foundElem = code.newAssign(ret, program.trueLiteral());
			AssignStmt nextList  = code.newAssign(lst, lst.join(next));
			
			// make the isEmpty statement the entry point of the procedure
			// set the successors of the statements, the successors not set
			// are left to be the exit statement of the procedure
			code.setEntry(isEmpty);
			isEmpty.setThen(notFound);
			isEmpty.setElse(checkElem);
			checkElem.setThen(foundElem);
			checkElem.setElse(nextList);
			nextList.setNext(isEmpty);
		}
		
		public Looping() {
			super("Looping List", code);
		}
	}
	
	public static final class Recursive extends ListTest {
		
		private static final ForgeCFG.Impl code;
		
		static {
			code = ForgeCFG.implementation(contains);
			
			BranchStmt isEmpty = code.newBranch(lst.no());
			AssignStmt notFound = code.newAssign(ret, program.falseLiteral());
			BranchStmt checkElem = code.newBranch(lst.join(elem).eq(val));
			AssignStmt foundElem = code.newAssign(ret, program.trueLiteral());
			
			// a call statement must include a number of in- and out-arguments
			// equal to the called procedures in- and out-parameters, respectively
			List<ForgeExpression> inArgs = Arrays.asList(lst.join(next), val);
			List<ForgeVariable> outArgs = Arrays.<ForgeVariable>asList(ret);
			CallStmt callContains = code.newCall(code, inArgs, outArgs);

			code.setEntry(isEmpty);
			isEmpty.setThen(notFound);
			isEmpty.setElse(checkElem);
			checkElem.setThen(foundElem);
			checkElem.setElse(callContains);
		}
		
		public Recursive() {
			super("Recursive List", code);
		}
	}
}