package forge.tests;

import forge.cfg.AssignStmt;
import forge.cfg.BranchStmt;
import forge.cfg.ForgeCFG;
import forge.program.ForgeExpression;
import forge.program.ForgeProcedure;
import forge.program.ForgeProgram;
import forge.program.LocalVariable;
import forge.solve.ForgeSolution;
import forge.solve.SolveOptions.SatSolver;

/**
 * Good for testing nested loop unrolling.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class MultiplyTest extends ForgeTest {
	
	public static void main(String[] args) {
		MultiplyTest test = new MultiplyTest();
		test.newOptions(3).satSolver(SatSolver.MINISAT_PROVER);
		test.setBounds(7, NO_SCOPES);
		ForgeSolution sol = test.checkTest();
		System.out.println(test);
		System.out.println(sol);
	}
	
	public MultiplyTest() {
		super("Multiply", code, preCond, postCond, NO_GLOBALS);
	}
	
	private static final ForgeCFG.Impl code;
	private static final ForgeExpression preCond, postCond;
	
	static {
		// make the procedure
		ForgeProgram prog = new ForgeProgram();
		LocalVariable x = prog.newLocalVariable("x", prog.integerDomain());
		LocalVariable y = prog.newLocalVariable("y", prog.integerDomain());
		LocalVariable product = prog.newLocalVariable("product", prog.integerDomain());
		ForgeProcedure proc = prog.newProcedure("multiply", x.and(y), product);
		
		// code the multiply procedure
		code = ForgeCFG.implementation(proc);
		ForgeExpression zero = prog.integerLiteral(0);
		ForgeExpression one = prog.integerLiteral(1);
		LocalVariable xSign = prog.newLocalVariable("xSign", prog.booleanDomain());
		LocalVariable ySign = prog.newLocalVariable("ySign", prog.booleanDomain());
		LocalVariable productSign = prog.newLocalVariable("productSign", prog.booleanDomain());
		LocalVariable i = prog.newLocalVariable("i", prog.integerDomain());
		LocalVariable j = prog.newLocalVariable("j", prog.integerDomain());
		
		// make the statements
		AssignStmt setXSign = code.newAssign(xSign, x.gt(zero));
		AssignStmt setYSign = code.newAssign(ySign, y.gt(zero));
		AssignStmt setProductSign = code.newAssign(productSign, xSign.eq(ySign));
		AssignStmt initProduct = code.newAssign(product, zero);
		AssignStmt initI = code.newAssign(i, xSign.thenElse(x, zero.minus(x)));
		BranchStmt testI = code.newBranch(i.gt(zero));
		AssignStmt initJ = code.newAssign(j, ySign.thenElse(y, zero.minus(y)));
		BranchStmt testJ = code.newBranch(j.gt(zero));
		AssignStmt incProduct = code.newAssign(product, product.plus(one));
		AssignStmt decI	= code.newAssign(i, i.minus(one));
		AssignStmt decJ	= code.newAssign(j, j.minus(one));
		AssignStmt finalProduct = code.newAssign(product, productSign.thenElse(product, zero.minus(product)));
		
		// connect the statements
		code.setEntry(setXSign);
		setXSign.setNext(setYSign);
		setYSign.setNext(setProductSign);
		setProductSign.setNext(initProduct);
		initProduct.setNext(initI);
		initI.setNext(testI);
		testI.setThen(initJ);
		testI.setElse(finalProduct);
		initJ.setNext(testJ);
		testJ.setThen(incProduct);
		testJ.setElse(decI);
		incProduct.setNext(decJ);
		decJ.setNext(testJ);
		decI.setNext(testI);
		
		// make the precondition
		// args are not the minimum value
		LocalVariable z = prog.newLocalVariable("z", prog.integerDomain());
		preCond = x.one().and(y.one()).and(x.gt(z).and(y.gt(z)).forSome(z));
		postCond = product.eq(x.times(y));
	}
}