package forge.tests;

import static java.util.Collections.singleton;

import java.util.HashMap;
import java.util.Map;

import forge.cfg.AssignStmt;
import forge.cfg.BranchStmt;
import forge.cfg.ForgeCFG;
import forge.cfg.SpecStmt;
import forge.program.ForgeExpression;
import forge.program.ForgeProcedure;
import forge.program.ForgeProgram;
import forge.program.InstanceDomain;
import forge.program.IntegerLiteral;
import forge.program.LocalVariable;
import forge.solve.ForgeBounds;
import forge.solve.ForgeSolver;
import forge.solve.Trace;

/**
 * Analyzes TopologicalFIR
 * 
 * @author Greg Dennis (gdennis@mit.edu)
 */
public class TopologicalSortTest {

	public static void main(String[] args) {
		TopologicalSortTest fir = new TopologicalSortTest();
		System.out.println(fir.kahn);
		
		ForgeSolver solver = new ForgeSolver(fir.kahn, 3);
		Map<InstanceDomain, Integer> dom2Scope = new HashMap<InstanceDomain, Integer>();
		dom2Scope.put(fir.node, 3);
		ForgeBounds bounds = new ForgeBounds(fir.program, 4, dom2Scope);
		
		ForgeCFG.Spec spec = ForgeCFG.specification(fir.kahn.procedure(),
				fir.program.globalVariables(),
				fir.node.difference(fir.node.join(fir.dag)).some());
		Trace trace = solver.run(spec, bounds).trace();
		System.out.println(trace);
	}
	
	final ForgeProgram program;
	final IntegerLiteral zero, one;
	final InstanceDomain node;
	final LocalVariable dag, order;
	final ForgeProcedure sort;
	final ForgeCFG kahn;
	
	private final LocalVariable n, i;
	
	TopologicalSortTest() {
		program = new ForgeProgram();
		zero = program.integerLiteral(0);
		one  = program.integerLiteral(1);
		node = program.newInstanceDomain("Node");
		dag   = program.newLocalVariable("dag", node.product(node));
		order = program.newLocalVariable("order", node.product(node));
		sort  = program.newProcedure("sort", dag, order);
		
		n = program.newLocalVariable("n", node);
		i = program.newLocalVariable("i", program.integerDomain());
		
		kahn = kahnCFG();
	}
	
	private ForgeCFG kahnCFG() {
		ForgeCFG.Impl cfg = ForgeCFG.implementation(sort);
		LocalVariable queue = program.newLocalVariable("queue",
				program.integerDomain().product(node));
		LocalVariable graph = program.newLocalVariable("graph", node.product(node));
		LocalVariable roots = program.newLocalVariable("roots", node);
		
		// initialize the order, queue, and graph
		AssignStmt initOrder = cfg.newAssign(order, program.empty(2));
		AssignStmt initQueue = cfg.newAssign(queue, program.empty(2));
		AssignStmt initGraph = cfg.newAssign(graph, dag);
		
		// enqueue the roots in some order
		AssignStmt setRoots = cfg.newAssign(roots, node.difference(dag.range()));
		SpecStmt enqueueRoots = enqueue(cfg, queue, roots);
		
		// if queue nonempty, add first element to order
		BranchStmt queueEmpty = cfg.newBranch(queue.no());
		ForgeExpression next = zero.join(queue);
		ForgeExpression lastInOrder = node.difference(order.join(node));
		ForgeExpression newOrder = order.union(lastInOrder.product(next));
		AssignStmt addOrder = cfg.newAssign(order, newOrder);
		
		// dequeue first element
		ForgeExpression newQueue = i.plus(one).join(queue).eq(n).comprehension(i.and(n));
		AssignStmt dequeue = cfg.newAssign(queue, newQueue);
		
		// remove incoming edges from graph
		ForgeExpression newGraph = graph.difference(next.product(node));
		AssignStmt removeEdges = cfg.newAssign(graph, newGraph);
		
		// enqueue the adjacents that no longer have incoming.
		AssignStmt setNewRoots = cfg.newAssign(roots, next.join(graph).difference(graph.range()));
		SpecStmt enqueueNewRoots = enqueue(cfg, queue, roots);
		
		// make the graph
		cfg.setEntry(initOrder);
		initOrder.setNext(initQueue);
		initQueue.setNext(initGraph);
		initGraph.setNext(setRoots);
		setRoots.setNext(enqueueRoots);
		enqueueRoots.setNext(queueEmpty);
		queueEmpty.setElse(addOrder);
		addOrder.setNext(dequeue);
		dequeue.setNext(removeEdges);
		removeEdges.setNext(setNewRoots);
		setNewRoots.setNext(enqueueNewRoots);
		enqueueNewRoots.setNext(queueEmpty);
		return cfg;
	}
	
	private SpecStmt enqueue(ForgeCFG.Impl cfg, LocalVariable queue, ForgeExpression elems) {
//		ForgeExpression indices = queue.domain();
		ForgeExpression length = queue.size().eq(queue.old().size().plus(elems.size()));
//		ForgeExpression prefix = i.gte(zero).and(i.in(indices.difference(zero)).implies(i.minus(one).in(indices))).forAll(i);
//		ForgeExpression contains = queue.old().in(queue).and(elems.in(queue.range()));
//		ForgeExpression enqueued = length.and(prefix).and(contains);
//		return cfg.newSpec(singleton(queue), enqueued);
		return cfg.newSpec(singleton(queue), length);
	}
	

//	private ForgeExpression totalOrder(ForgeVariable rel, ForgeExpression elems) {
//		ForgeExpression functional = n.join(rel).lone().forAll(n);
//		ForgeExpression reachable = n.union(n.join(rel.closure())).eq(elems).forSome(n);
//		return functional.and(reachable);		
//	}
	
}
