package forge.tests;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import forge.cfg.AssignStmt;
import forge.cfg.BranchStmt;
import forge.cfg.ForgeCFG;
import forge.cfg.UpdateStmt;
import forge.program.ForgeExpression;
import forge.program.ForgeProcedure;
import forge.program.ForgeProgram;
import forge.program.GlobalVariable;
import forge.program.InstanceDomain;
import forge.program.LocalVariable;
import forge.solve.SolveOptions.SatSolver;


/**
 * Insertion into a binary tree integer set.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class TreeIntSetTest extends ForgeTest {
	
	public static void main(String[] args) {
		TreeIntSetTest test = new TreeIntSetTest();
		Map<InstanceDomain, Integer> scopes = new HashMap<InstanceDomain, Integer>();
		scopes.put(test.treeDomain(), 1);
		scopes.put(test.nodeDomain(), 2);
		test.newOptions(2).satSolver(SatSolver.MINISAT_PROVER).strategy(HYBRID_STRATEGY);
		test.setBounds(4, scopes);
		test.standardTest();
	}

	public TreeIntSetTest() {
		super("Tree Int Set Insert", code, preCond, postCond, modGlobals);
	}
	
	public InstanceDomain treeDomain() {
		return tree;
	}
	
	public InstanceDomain nodeDomain() {
		return node;
	}
	
	private static final ForgeCFG.Impl code;
	private static final ForgeExpression preCond, postCond;
	private static final Set<GlobalVariable> modGlobals;
	private static final InstanceDomain tree, node;
	
	static {
		ForgeProgram prog = new ForgeProgram();
		tree = prog.newInstanceDomain("Tree");
		node = prog.newInstanceDomain("Node");
		GlobalVariable root = prog.newGlobalVariable("root", tree.product(node));
		GlobalVariable left = prog.newGlobalVariable("left", node.product(node));
		GlobalVariable right = prog.newGlobalVariable("right", node.product(node));
		GlobalVariable key = prog.newGlobalVariable("key", node.product(prog.integerDomain()));
		
		// make the procedure
		LocalVariable t = prog.newLocalVariable("t", tree);
		LocalVariable k = prog.newLocalVariable("k", prog.integerDomain());
		LocalVariable r = prog.newLocalVariable("r", prog.booleanDomain());
		ForgeProcedure proc = prog.newProcedure("add", t.and(k), r);
		
		// make the implementation stmts
		code = ForgeCFG.implementation(proc);
		LocalVariable y = prog.newLocalVariable("y", node);
		LocalVariable x = prog.newLocalVariable("x", node);
		UpdateStmt initY    = code.newAssign(y, prog.empty());
		UpdateStmt initX    = code.newAssign(x, t.join(root));
		BranchStmt someX    = code.newBranch(x.some());
		UpdateStmt assignY  = code.newAssign(y, x);
		BranchStmt kLtX     = code.newBranch(k.lt(x.join(key)));
		BranchStmt kGtX     = code.newBranch(k.gt(x.join(key)));
		AssignStmt rFalse   = code.newAssign(r, prog.falseLiteral());
		UpdateStmt xLeft    = code.newAssign(x, x.join(left));
		UpdateStmt xRight   = code.newAssign(x, x.join(right));
		BranchStmt noY      = code.newBranch(y.no());
		UpdateStmt kNode    = code.newCreate(x, node);
		UpdateStmt setKey   = code.newAssign(key, key.union(x.product(k)));
		UpdateStmt setRoot  = code.newAssign(root, root.override(t.product(x)));
		BranchStmt kLessY   = code.newBranch(k.lt(y.join(key)));
		UpdateStmt setLeft  = code.newAssign(left, left.override(y.product(x)));
		UpdateStmt setRight = code.newAssign(right, right.override(y.product(x)));
		AssignStmt rTrue    = code.newAssign(r, prog.trueLiteral());
		
		/**
		 insert(t, k) : (r)
		 
		 y := {};
		 x := t.root;
		 
		 while(some x) {
		   y := x;
		   if (k < x.key) {
		     x := x.left;
		   } else {
		     if (k > x.key) {
		       x := x.right;
		     } else {
		       r = false;
		       exit;
		     }
		   }
		 }
		 
		 x = new Node;
		 key := key ++ x->k;
		 
		 if (no y) {
		   root := root ++ t->x;
		 } else {
		   if (k < y.key) {
		     left := left ++ y->x;
		   } else {
		     right := right ++ y->x;
		   }
		 }
		 
		 r = true;
		 exit;
		 */
		
		// connect the stmts
		
		code.setEntry(initY);
		initY.setNext(initX);
		initX.setNext(someX);
		someX.setThen(assignY);
		someX.setElse(kNode);
		assignY.setNext(kLtX);
		kLtX.setThen(xLeft);
		kLtX.setElse(kGtX);
		kGtX.setThen(xRight);
		kGtX.setElse(rFalse);
		xLeft.setNext(someX);
		xRight.setNext(someX);
		kNode.setNext(setKey);
		setKey.setNext(noY);
		noY.setThen(setRoot);
		noY.setElse(kLessY);
		kLessY.setThen(setLeft);
		kLessY.setElse(setRight);
		setRoot.setNext(rTrue);
		setLeft.setNext(rTrue);
		setRight.setNext(rTrue);
		
		ForgeExpression rootOld  = root.old();
		ForgeExpression leftOld  = left.old();
		ForgeExpression rightOld = right.old();
		ForgeExpression keyOld   = key.old();
		ForgeExpression descsOld = leftOld.union(rightOld).closure();
		
		ForgeExpression validArgs = t.one().and(k.one());
		ForgeExpression validTree = t.join(rootOld).lone();
		LocalVariable n = prog.newLocalVariable("n", node);
		ForgeExpression nKey   = n.join(keyOld);
		ForgeExpression nLeft  = n.join(leftOld);
		ForgeExpression nRight = n.join(rightOld);
		ForgeExpression nodesPre  = t.join(rootOld.union(rootOld.join(descsOld)));
		ForgeExpression nodeMults = nLeft.lone().and(nRight.lone()).and(nKey.one());
		
		ForgeExpression leftKeys = nLeft.union(nLeft.join(descsOld)).join(keyOld);
		ForgeExpression rightKeys = nRight.union(nRight.join(descsOld)).join(keyOld);
		LocalVariable i = prog.newLocalVariable("i", prog.integerDomain());
		ForgeExpression ordered = i.in(leftKeys).implies(i.lt(nKey)).and(i.in(rightKeys).implies(i.gt(nKey))).forAll(i);
		
		ForgeExpression validNode = n.in(nodesPre).implies(nodeMults.and(ordered)).forAll(n);
		preCond = validArgs.and(validTree.and(validNode));
		
		ForgeExpression keysPre  = nodesPre.join(keyOld);
		ForgeExpression keysPost = t.join(root.union(root.join(left.union(right).closure()))).join(key);
		ForgeExpression keysSpec = keysPost.eq(keysPre.union(k));
		ForgeExpression retSpec  = r.eq(k.in(keysPre).not());
		postCond = keysSpec.and(retSpec);
		
		Set<GlobalVariable> mg = new HashSet<GlobalVariable>();
		mg.add(root); mg.add(left); mg.add(right); mg.add(key);
		modGlobals = Collections.unmodifiableSet(mg);
	}

}
