package forge.tests;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import forge.cfg.AssignStmt;
import forge.cfg.BranchStmt;
import forge.cfg.ForgeCFG;
import forge.cfg.SpecStmt;
import forge.cfg.UpdateStmt;
import forge.program.ForgeExpression;
import forge.program.ForgeProcedure;
import forge.program.ForgeProgram;
import forge.program.ForgeVariable;
import forge.program.GlobalVariable;
import forge.program.InstanceDomain;
import forge.program.LocalVariable;
import forge.solve.SolveOptions.SatSolver;
import forge.translate.ConstrainStrategy;


/**
 * Insertion into a binary tree integer set.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class TreeIntSetTest2 extends ForgeTest {
	
	public static void main(String[] args) {
		TreeIntSetTest2 test = new TreeIntSetTest2();
		Map<InstanceDomain, Integer> scopes = Collections.singletonMap(test.nodeDomain(), 5);
		test.newOptions(2).satSolver(SatSolver.MINISAT_PROVER).strategy(ConstrainStrategy.STRATEGY);
		test.setBounds(4, scopes);
		test.standardTest();
	}

	public TreeIntSetTest2() {
		super("Tree Int Set Insert2", code, preCond, postCond, modGlobals);
	}
	
	public InstanceDomain nodeDomain() {
		return node;
	}
	
	private static final ForgeCFG.Impl code;
	private static final ForgeExpression preCond, postCond;
	private static final Set<GlobalVariable> modGlobals;
	private static final InstanceDomain node;
	
	static {
		ForgeProgram prog = new ForgeProgram();
		node = prog.newInstanceDomain("Node");
		GlobalVariable root = prog.newGlobalVariable("root", node);
		GlobalVariable left = prog.newGlobalVariable("left", node.product(node));
		GlobalVariable right = prog.newGlobalVariable("right", node.product(node));
		GlobalVariable value = prog.newGlobalVariable("value", node.product(prog.integerDomain()));
		
		// make the add procedure
		LocalVariable v = prog.newLocalVariable("v", prog.integerDomain());
		LocalVariable r = prog.newLocalVariable("r", prog.booleanDomain());
		ForgeProcedure add = prog.newProcedure("add", v, r);
		
		// make the getChild procedure
		LocalVariable m = prog.newLocalVariable("m", node);
		LocalVariable isLeft = prog.newLocalVariable("isLeft", prog.booleanDomain());
		ForgeProcedure getChild = prog.newProcedure("getChild", m.and(isLeft), m);
		
		/**
		 add(v) : (r) {
			 p := {};
			 n := root;
			 
			 while(some n) {
			   p := n;
			   if (v = n.value) {
			     r := false;
			     exit;
			   }
			   n := getChild(n, v < n.value);
			 }
			 
			 n = new Node;
			 value := value + n->v;
			 
			 if (no p) {
			   root := n;
			 } else {
			   if (v < p.value) {
			     left := left ++ p->n;
			   } else {
			     right := right ++ p->n;
			   }
			 }
			 
			 r = true;
			 exit;
		 }
		 
		 getChild(m, isLeft) : m {
		 	m := spec((isLeft => m = old(m).left) && (!isLeft => m = old(m).right))
		 }
		 */
		
		// make the add implementations
		{
			ForgeCFG.Impl cfg = ForgeCFG.implementation(getChild);
			SpecStmt setM = cfg.newSpec(Collections.singleton(m),
					isLeft.implies(m.eq(m.old().join(left))).and(
				    isLeft.not().implies(m.eq(m.old().join(right)))));
			cfg.setEntry(setM);
			
			code = ForgeCFG.implementation(add);
			LocalVariable p = prog.newLocalVariable("p", node);
			LocalVariable n = prog.newLocalVariable("n", node);
			UpdateStmt initP    = code.newAssign(p, prog.empty());
			UpdateStmt initN    = code.newAssign(n, root);
			BranchStmt someN    = code.newBranch(n.some());
			UpdateStmt assignP  = code.newAssign(p, n);
			BranchStmt vEqNVal  = code.newBranch(v.eq(n.join(value)));
			AssignStmt rFalse   = code.newAssign(r, prog.falseLiteral());
			UpdateStmt callChild  = code.newCall(cfg,
					Arrays.asList(n, v.lt(n.join(value))),
					Collections.<ForgeVariable>singletonList(n));
			UpdateStmt newNode  = code.newCreate(n, node);
			UpdateStmt setValue = code.newAssign(value, value.union(n.product(v)));
			BranchStmt noP      = code.newBranch(p.no());
			UpdateStmt setRoot  = code.newAssign(root, n);
			BranchStmt vLtPVal  = code.newBranch(v.lt(p.join(value)));
			UpdateStmt setLeft  = code.newAssign(left, left.override(p.product(n)));
			UpdateStmt setRight = code.newAssign(right, right.override(p.product(n)));
			AssignStmt rTrue    = code.newAssign(r, prog.trueLiteral());
			
			code.setEntry(initP);
			initP.setNext(initN);
			initN.setNext(someN);
			someN.setThen(assignP);
			someN.setElse(newNode);
			assignP.setNext(vEqNVal);
			vEqNVal.setThen(rFalse);
			vEqNVal.setElse(callChild);
			callChild.setNext(someN);
			newNode.setNext(setValue);
			setValue.setNext(noP);
			noP.setThen(setRoot);
			noP.setElse(vLtPVal);
			vLtPVal.setThen(setLeft);
			vLtPVal.setElse(setRight);
			setRoot.setNext(rTrue);
			setLeft.setNext(rTrue);
			setRight.setNext(rTrue);
		}
		
		// make the specification
		{
			ForgeExpression rootOld  = root.old();
			ForgeExpression leftOld  = left.old();
			ForgeExpression rightOld = right.old();
			ForgeExpression valueOld = value.old();
			ForgeExpression descsOld = leftOld.union(rightOld).closure();
			
			ForgeExpression validArg = v.one();
			ForgeExpression validRoot = rootOld.lone();
			LocalVariable n = prog.newLocalVariable("n", node);
			ForgeExpression nKey   = n.join(valueOld);
			ForgeExpression nLeft  = n.join(leftOld);
			ForgeExpression nRight = n.join(rightOld);
			ForgeExpression nodesPre  = rootOld.union(rootOld.join(leftOld.union(rightOld).closure()));
			ForgeExpression nodeMults = nLeft.lone().and(nRight.lone()).and(nKey.one());
			
			ForgeExpression leftValues = nLeft.union(nLeft.join(descsOld)).join(valueOld);
			ForgeExpression rightValues = nRight.union(nRight.join(descsOld)).join(valueOld);
			LocalVariable i = prog.newLocalVariable("i", prog.integerDomain());
			ForgeExpression ordered = i.in(leftValues).implies(i.lt(nKey)).and(i.in(rightValues).implies(i.gt(nKey))).forAll(i);

			ForgeExpression validNodes = n.in(nodesPre).implies(nodeMults.and(ordered)).forAll(n);
			preCond = validArg.and(validRoot.and(validNodes));
			
			ForgeExpression valuesPre  = nodesPre.join(valueOld);
			ForgeExpression valuesPost = root.union(root.join(left.union(right).closure())).join(value);
			ForgeExpression valuesSpec = valuesPost.eq(valuesPre.union(v));
			ForgeExpression retSpec  = r.eq(v.in(valuesPre).not());
			postCond = valuesSpec.and(retSpec);
			
			Set<GlobalVariable> mg = new HashSet<GlobalVariable>();
			mg.add(root); mg.add(left); mg.add(right); mg.add(value);
			modGlobals = Collections.unmodifiableSet(mg);
		}
	}

}
