package forge.transform;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import edu.mit.csail.sdg.util.collections.ArrayHashSetUniqueList;
import edu.mit.csail.sdg.util.collections.UniqueList;
import forge.cfg.AssignStmt;
import forge.cfg.BranchStmt;
import forge.cfg.CFGStmt;
import forge.cfg.CFGVisitor;
import forge.cfg.CallStmt;
import forge.cfg.CreateStmt;
import forge.cfg.ExitStmt;
import forge.cfg.ForgeCFG;
import forge.cfg.SpecStmt;
import forge.cfg.UpdateStmt;
import forge.program.ForgeExpression;
import forge.program.ForgeVariable;
import forge.program.InstanceDomain;

/**
 * Implementation support for a transformer of a single call graph.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public abstract class AbstractTransformer implements Transformer {

	private final Map<ForgeCFG, ForgeCFG.Impl> source2TargetCFG;
	final Map<CFGStmt, CFGStmt> target2SourceStmt;
	
	/**
	 * Constructs a transformer with a tag that will be appended to each
	 * transformed procedure. Specifically, if the original procedure is
	 * named "x", the transformed procedure will be named "x_tag".
	 */
	protected AbstractTransformer() {
		source2TargetCFG   = new HashMap<ForgeCFG, ForgeCFG.Impl>();
		target2SourceStmt  = new HashMap<CFGStmt, CFGStmt>();
	}
	
	/**
	 * Checks in the cache whether or not the procedure has already been
	 * transformed. If it has, returns the transformed version. If not, calls
	 * makeTarget, caches the source procedure to the target, and then calls
	 * transform(source, target).
	 */
	public final ForgeCFG transform(ForgeCFG source) {
		ForgeCFG.Impl target = source2TargetCFG.get(source);
		if (target == null) {
			TransformVisitor vis = visitor(source);
			target = vis.target();
			source2TargetCFG.put(source, target);
			CFGStmt entry = vis.apply(source.entry());
			target.setEntry(entry);
		}
		return target;
	}

	public final CFGStmt sourceStmt(CFGStmt targetStmt) {
		return target2SourceStmt.get(targetStmt);
	}

	/**
	 * Returns a TransformVisitor that converts source.
	 */
	protected abstract TransformVisitor visitor(ForgeCFG source);
	
	/**
	 * A TransformVisitor
	 */
	protected static abstract class TransformVisitor implements CFGVisitor {
		
		private final AbstractTransformer transformer;
		private final ForgeCFG source;
		private final ForgeCFG.Impl target;
		
		/** maps each target stmt to the source from which it was replaced. */
		private final Map<CFGStmt, CFGStmt> replaced;
		
		/**
		 * Constructs a TransformVisitor belonging to the given transformer,
		 * 
		 */
		protected TransformVisitor(AbstractTransformer transformer, ForgeCFG source) {
			this(transformer, source,
					ForgeCFG.implementation(source.procedure()));
		}
		
		protected TransformVisitor(AbstractTransformer transformer,
				ForgeCFG source, ForgeCFG.Impl target) {
			this.transformer = transformer;
			this.source = source;
			this.target = target;
			replaced = new HashMap<CFGStmt, CFGStmt>();
		}
		
		public final ForgeCFG source() {
			return source;
		}
		
		public final ForgeCFG.Impl target() {
			return target;
		}
		
		public final CFGStmt apply(CFGStmt stmt) {
			stmt.accept(this);
			return replaced.get(stmt);
		}
		
		/**
		 * Sets the result of applying the visitor to the given source.
		 */
		protected final void replace(CFGStmt source, CFGStmt target) {
			source().checkCFG(source);
			if (target != null) target().checkCFG(target);
			replaced.put(source, target);
		}

		/**
		 * Transforms the statement to a new statement or null, but must also
		 * call replace to set the result of an application to this statement.
		 * <p>
		 * This implementation copies the stmt, replaces the stmt for the copy,
		 * applies the visitor to the next stmt, sets the next stmt of the copy
		 * to the result of that application, and returns the copy.
		 */
		protected UpdateStmt transform(AssignStmt stmt) {
			return applyNext(stmt, copy(stmt));
		}

		/**
		 * Transforms the statement to a new statement or null, but must also
		 * call replace to set the result of an application to this statement.
		 * <p>
		 * This implementation copies the stmt, replaces the stmt for the copy,
		 * applies the visitor to the next stmt, sets the next stmt of the copy
		 * to the result of that application, and returns the copy.
		 */
		protected UpdateStmt transform(CreateStmt stmt) {
			return applyNext(stmt, copy(stmt));
		}
		
		/**
		 * Transforms the statement to a new statement or null, but must also
		 * call replace to set the result of an application to this statement.
		 * <p>
		 * This implementation copies the stmt, replaces the stmt for the copy,
		 * applies the visitor to the next stmt, sets the next stmt of the copy
		 * to the result of that application, and returns the copy.
		 */
		protected UpdateStmt transform(SpecStmt stmt) {
			return applyNext(stmt, copy(stmt));
		}
		
		/**
		 * Transforms the statement to a new statement or null, but must also
		 * call replace to set the result of an application to this statement.
		 * <p>
		 * This implementation copies the stmt, replaces the stmt for the copy,
		 * applies the visitor to the next stmt, sets the next stmt of the copy
		 * to the result of that application, and returns the copy.
		 */
		protected CFGStmt transform(CallStmt stmt) {
			return applyNext(stmt, copy(stmt));
		}
		
		private UpdateStmt applyNext(UpdateStmt stmt, UpdateStmt copy) {
			replace(stmt, copy);
			copy.setNext(apply(stmt.getNext()));
			return copy;
		}

		/**
		 * Transforms the statement to a new statement or null, but must also
		 * call replace to set the result of an application to this statement.
		 * <p>
		 * This implementation copies the stmt, replaces the stmt for the copy,
		 * applies the visitor to the then stmt, sets the then branch of the
		 * copy to the result of that application, applies the visitor to the
		 * else stmt, sets the else branch to the result of that application.
		 */
		protected CFGStmt transform(BranchStmt stmt) {
			BranchStmt copy = copy(stmt);
			replace(stmt, copy);
			copy.setThen(apply(stmt.getThen()));
			copy.setElse(apply(stmt.getElse()));
			return copy;
		}

		/**
		 * Transforms the statement to a new statement or null, but must also
		 * call replace to set the result of an application to this statement.
		 * <p>
		 * This implementation replaces the stmt for the exit stmt of the
		 * target cfg and returns the target's exit.
		 */
		protected CFGStmt transform(ExitStmt stmt) {
			ExitStmt copy = target.exit();
			replace(stmt, copy);
			return copy;
		}
		
		/**
		 * Called during the copy methods on any expression that appears.
		 * Given a ForgeVariable, must return a ForgeVariable and given an
		 * InstanceDomain, must return an InstanceDomain.
		 */
		protected ForgeExpression replaceExpr(ForgeExpression expr) {
			return expr;
		}
		
		protected final AssignStmt copy(AssignStmt stmt) {
			return target.newAssign(
					replaceVar(stmt.var()),
					replaceExpr(stmt.expr()));
		}
		
		protected final CallStmt copy(CallStmt stmt) {
			return target.newCall(
					transformer.transform(stmt.called()),
					replaceExprs(stmt.inArgs()),
					replaceVars(stmt.outArgs()));
		}
		
		protected final CreateStmt copy(CreateStmt stmt) {
			return target.newCreate(
				replaceVar(stmt.var()), 
				replaceDom(stmt.type()));
		}
		
		protected final BranchStmt copy(BranchStmt stmt) {
			return target.newBranch(
					replaceExpr(stmt.condition()));
		}
		
		protected final SpecStmt copy(SpecStmt stmt) {
			return target.newSpec(
					replaceVars(stmt.modified()).asSet(),
					replaceExpr(stmt.condition()));
		}
		
		protected final ForgeVariable replaceVar(ForgeVariable var) {
			return (ForgeVariable)replaceExpr(var);
		}
		
		protected final InstanceDomain replaceDom(InstanceDomain dom) {
			return (InstanceDomain)replaceExpr(dom);
		}
		
		protected final List<ForgeExpression> replaceExprs(List<ForgeExpression> exprs) {
			List<ForgeExpression> replaced = new ArrayList<ForgeExpression>(exprs.size());
			for (ForgeExpression expr : exprs) {
				replaced.add(replaceExpr(expr));
			}
			return replaced;
		}
		
		protected final UniqueList<ForgeVariable> replaceVars(Collection<ForgeVariable> vars) {
			UniqueList<ForgeVariable> replaced = new ArrayHashSetUniqueList<ForgeVariable>();
			for (ForgeVariable var : vars) {
				replaced.add(replaceVar(var));
			}
			return replaced;
		}
		
		public final void visit(AssignStmt stmt) {
			if (!replaced.containsKey(stmt)) record(stmt, transform(stmt));
		}
		public final void visit(BranchStmt stmt) {
			if (!replaced.containsKey(stmt)) record(stmt, transform(stmt));
		}
		public final void visit(CallStmt stmt) {
			if (!replaced.containsKey(stmt)) record(stmt, transform(stmt));
		}
		public final void visit(CreateStmt stmt) {
			if (!replaced.containsKey(stmt)) record(stmt, transform(stmt));
		}
		public final void visit(ExitStmt stmt) {
			if (!replaced.containsKey(stmt)) record(stmt, transform(stmt));
		}
		public final void visit(SpecStmt stmt) {
			if (!replaced.containsKey(stmt)) record(stmt, transform(stmt));
		}
		
		private void record(CFGStmt source, CFGStmt target) {
			source().checkCFG(source);
			if (target != null) {
				target().checkCFG(target);
				// map target to source, target should not map to more than one
				CFGStmt temp = transformer.target2SourceStmt.put(target, source);
				assert temp == null : "SOURCE " + source + "; TARGET " + target + "; TEMP " + temp;
			}
		}
	}
	
}
