package forge.transform;

import forge.program.BinaryExpression;
import forge.program.ConditionalExpression;
import forge.program.ExpressionVisitor;
import forge.program.ForgeLiteral;
import forge.program.ForgeType;
import forge.program.ForgeVariable;
import forge.program.OldExpression;
import forge.program.ProjectionExpression;
import forge.program.QuantifyExpression;
import forge.program.UnaryExpression;

/**
 * Descends through a Forge expression graph without caching.
 * Return variables of the visit methods are ignores, so they should
 * all be null.
 * 
 * @author Greg Dennis (gdennis@mit.edu)
 */
public abstract class ExpressionDescender extends ExpressionVisitor<Object> {
	
	protected ExpressionDescender() {}
	
	/**
	 * Applies visitor to left and right subexpression.
	 */
	protected void descend(BinaryExpression expr) {
		expr.left().accept(this);
		expr.right().accept(this);
	}
	
	/**
	 * Applies visitor to subexpression.
	 */
	protected void descend(UnaryExpression expr) {
		expr.sub().accept(this);
	}
	
	/**
	 * Applies visitor to variable and subexpression.
	 */
	protected void descend(QuantifyExpression expr) {
		expr.sub().accept(this);
	}
	
	/**
	 * Applies visitor to condition, then, and else expressions.
	 */
	protected void descend(ConditionalExpression expr) {
		expr.condition().accept(this);
		expr.thenExpr().accept(this);
		expr.elseExpr().accept(this);
	}
	
	/**
	 * Applies visitor to subexpression
	 */
	protected void descend(ProjectionExpression expr) {
		expr.sub().accept(this);
	}
	
	/**
	 * Applies visitor to variable.
	 */
	protected void descend(OldExpression expr) {
		expr.variable().accept(this);
	}
	
	/**
	 * Does nothing.
	 */
	protected void descend(ForgeVariable expr) {}

	/**
	 * Does nothing.
	 */
	protected void descend(ForgeLiteral expr) {}
	
	/**
	 * Does nothing.
	 */
	protected void descend(ForgeType expr) {}
	
	//
	// these just call the descend methods
	//
	
	@Override
	protected final Object visit(BinaryExpression expr) {
		descend(expr);
		return null;
	}
	
	@Override
	protected final Object visit(UnaryExpression expr) {
		descend(expr);
		return null;
	}
	
	@Override
	protected final Object visit(ConditionalExpression expr) {
		descend(expr);
		return null;
	}
	
	@Override
	protected final Object visit(QuantifyExpression expr) {
		descend(expr);
		return null;
	}
	
	@Override
	protected final Object visit(ProjectionExpression expr) {
		descend(expr);
		return null;
	}
	
	@Override
	protected final Object visit(OldExpression expr) {
		descend(expr);
		return null;
	}
	
	@Override
	protected final Object visit(ForgeLiteral expr) {
		descend(expr);
		return null;
	}
	
	@Override
	protected final Object visit(ForgeType expr) {
		descend(expr);
		return null;
	}
	
	@Override
	protected final Object visit(ForgeVariable expr) {
		descend(expr);
		return null;
	}

}
