package forge.transform;

import forge.program.BinaryExpression;
import forge.program.ConditionalExpression;
import forge.program.ExpressionVisitor;
import forge.program.ForgeExpression;
import forge.program.ForgeLiteral;
import forge.program.ForgeType;
import forge.program.ForgeVariable;
import forge.program.LocalDecls;
import forge.program.OldExpression;
import forge.program.ProjectionExpression;
import forge.program.QuantifyExpression;
import forge.program.UnaryExpression;

/**
 * A depth-first visitor that replaces each node in the expression.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public abstract class ExpressionReplacer
		extends ExpressionVisitor<ForgeExpression> {

	protected ExpressionReplacer() {
		super();
	}
	
	protected ForgeExpression visit(UnaryExpression expr) {
		ForgeExpression replacedSub = expr.sub().accept(this);
		ForgeExpression replaced = replacedSub == expr.sub() ?
				expr : replacedSub.apply(expr.op());
		super.putCache(expr, replaced);
		return replaced;
	}

	protected ForgeExpression visit(BinaryExpression expr) {
		ForgeExpression replacedLeft = expr.left().accept(this);
		ForgeExpression replacedRight = expr.right().accept(this);
		ForgeExpression replaced =
			(replacedLeft == expr.left() && replacedRight == expr.right()) ?
					expr : replacedLeft.compose(expr.op(), replacedRight);
		super.putCache(expr, replaced);
		return replaced;
	}
	
	protected ForgeExpression visit(ConditionalExpression expr) {
		ForgeExpression replacedCond = expr.condition().accept(this);
		ForgeExpression replacedThen = expr.thenExpr().accept(this);
		ForgeExpression replacedElse = expr.elseExpr().accept(this);
		ForgeExpression replaced =
			(replacedCond == expr.condition() &&
					replacedThen == expr.thenExpr() &&
					replacedElse == expr.elseExpr()) ? expr :
					replacedCond.thenElse(replacedThen, replacedElse);
		super.putCache(expr, replaced);
		return replaced;
	}
	
	protected ForgeExpression visit(ProjectionExpression expr) {
		ForgeExpression replacedSub = expr.sub().accept(this);
		
		final ForgeExpression replaced;
		if (replacedSub == expr.sub()) {
			replaced = expr;
		} else {
			int[] columns = new int[expr.columns().size()];
			int i = 0;
			for (int c : expr.columns()) columns[i++] = c;
			replaced = replacedSub.projection(columns);
			
		}

		super.putCache(expr, replaced);
		return replaced;
	}
	
	protected ForgeExpression visit(QuantifyExpression expr) {
		LocalDecls replacedDecls = replaceDecls(expr.decls());
		ForgeExpression replacedForm = expr.sub().accept(this);
		ForgeExpression replaced = (replacedDecls == expr.decls() && replacedForm == expr.sub()) ?
				expr : replacedForm.quantify(expr.op(), replacedDecls);
		super.putCache(expr, replaced);
		return replaced;
	}
	
	protected ForgeExpression visit(OldExpression expr) {
		ForgeVariable replacedVar = (ForgeVariable)expr.variable().accept(this);
		ForgeExpression replaced = replacedVar == expr.variable() ?
				expr : replacedVar.old();
		super.putCache(expr, replaced);
		return replaced;
	}

	protected ForgeExpression visit(ForgeLiteral expr) {
		super.putCache(expr, expr);
		return expr;
	}

	protected ForgeExpression visit(ForgeType expr) {
		super.putCache(expr, expr);
		return expr;
	}
	
	protected ForgeExpression visit(ForgeVariable expr) {
		super.putCache(expr, expr);
		return expr;
	}
	
	protected LocalDecls replaceDecls(LocalDecls decls) {
		return decls;
	}
}
