package forge.transform;

import java.util.LinkedList;
import java.util.List;

import edu.mit.csail.sdg.util.collections.Bag;
import edu.mit.csail.sdg.util.collections.HashBag;
import edu.mit.csail.sdg.util.collections.SDGCollections;
import forge.cfg.AssignStmt;
import forge.cfg.BranchStmt;
import forge.cfg.CFGStmt;
import forge.cfg.CallStmt;
import forge.cfg.CreateStmt;
import forge.cfg.ExitStmt;
import forge.cfg.ForgeCFG;
import forge.cfg.SpecStmt;
import forge.cfg.UpdateStmt;
import forge.program.ForgeExpression;
import forge.program.ForgeVariable;
import forge.program.LocalVariable;
import forge.program.QuantifyExpression;

/**
 * Inlines all the calls in a call graph into a procedure.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class InlineTransformer extends AbstractTransformer {
	
	private final int numUnrolls;

	/**
	 * Constructs a call inliner that unrolls recursion the given number of times.
	 * 
	 * @throws IllegalArgumentException - numUnrolls < 1
	 */
	public InlineTransformer(int numUnrolls) {
		if (numUnrolls < 1) {
			throw new IllegalArgumentException("numUnrolls = " + numUnrolls);
		}
		this.numUnrolls = numUnrolls;
	}

	@Override
	protected TransformVisitor visitor(ForgeCFG source) {
		return new TopInliner(source);
	}
	
	/**
	 * Visitor that inlines procedure calls.
	 * The visit methods return null if the stmt should be unreachable
	 * as determined by the maximum number of loop unrollings.
	 */
	private final class TopInliner extends TransformVisitor {
		
		// this only helps generate somewhat unique names
		private int inlineCount = 0;
		
		private TopInliner(ForgeCFG source) {
			super(InlineTransformer.this, source);
		}
		
		@Override
		protected CFGStmt transform(CallStmt stmt) {
			CFGStmt inlinedNext = apply(stmt.getNext());
			Bag<CallStmt> called = SDGCollections.singletonBag(stmt);
			CFGStmt inlinedCall = inlineCall(stmt.called(), stmt.inArgs(),
					stmt.outArgs(), inlinedNext, called);
			super.replace(stmt, inlinedCall);
			return null;
		}
		
		/**
		 * For a procedure declared as:
		 *   p(T i1, T i2) : o1, o2 { body }
		 * 
		 * A call to that procedure of the form
		 *   v1, v2 = p(e1, e2)
		 *   
		 * becomes
		 *   ifresh1 = e1
		 *   ifresh2 = e2
		 *   body [ifresh1/l1, ifresh2/l2, ofresh1/o1, ofresh2/o2]
		 *   v1 = ofresh1
		 *   v2 = ofresh2
		 */
		private CFGStmt inlineCall(ForgeCFG calledCFG,
				List<ForgeExpression> inArgs, List<ForgeVariable> outArgs,
				CFGStmt inlinedNext, Bag<CallStmt> called) {
			List<LocalVariable> ins = calledCFG.procedure().inParams().locals();
			List<LocalVariable> outs = calledCFG.procedure().outParams().locals();			
			LocalReplacer localReplacer = new LocalReplacer(inlineCount++);
			LinkedList<AssignStmt> inAssigns = new LinkedList<AssignStmt>();
			
			// create the fresh param for each of the in-parameters
			// create assign stmts of each fresh param to its argument
			for (int i = 0; i < ins.size(); i++) {
				ForgeVariable freshParam = (LocalVariable)ins.get(i).accept(localReplacer);				
				AssignStmt assign = target().newAssign(freshParam, inArgs.get(i));
				inAssigns.addFirst(assign);
			}
			
			// create a fresh param for each of the out-parameters
			// create assign stmts of each out-parameter to its fresh version
			// prepend the assignments to the rest of the CFG
			for (int i = 0; i < outs.size(); i++) {
				ForgeVariable freshParam = (LocalVariable)outs.get(i).accept(localReplacer);
				AssignStmt assign = target().newAssign(outArgs.get(i), freshParam);
				assign.setNext(inlinedNext);
				inlinedNext = assign;
			}
			
			// inline the call
			RestInliner inliner = new RestInliner(calledCFG, called, inlinedNext, localReplacer);
			CFGStmt inlined = inliner.apply(calledCFG.entry());
			assert inlined != null;
			
			// add assign statements
			for (AssignStmt assign : inAssigns) {
				assign.setNext(inlined);
				inlined = assign;
			}
			
			return inlined;
		}
		
		private final class RestInliner extends TransformVisitor {
			
			private final Bag<CallStmt> called;
			private final CFGStmt callExit;
			private final ExpressionReplacer localReplacer;
			
			private RestInliner(ForgeCFG calledCFG, Bag<CallStmt> called,
					CFGStmt callExit, LocalReplacer localReplacer) {
				super(InlineTransformer.this, calledCFG, TopInliner.this.target());
				this.called = called;
				this.callExit = callExit;
				this.localReplacer = localReplacer;
			}
			
			@Override
			protected CFGStmt transform(CallStmt stmt) {
				final CFGStmt inlined;
				
				// if this call has already been unrolled
				// for the given number of unrollings, stop
				// otherwise, unroll it again
				if (called.count(stmt) >= numUnrolls) {
					inlined = null;
				} else {
					CFGStmt inlinedNext = apply(stmt.getNext());
					if (inlinedNext == null) {
						inlined = null;
					} else {
						Bag<CallStmt> called2 = new HashBag<CallStmt>(called);
						called2.add(stmt);
						called2 = SDGCollections.unmodifiableBag(called2);
						inlined = inlineCall(stmt.called(),
								replaceExprs(stmt.inArgs()),
								replaceVars(stmt.outArgs()),
								inlinedNext, called2);
					}
				}
				
				super.replace(stmt, inlined);
				return null;
			}
			
			@Override
			protected CFGStmt transform(BranchStmt stmt) {
				// replace the condition and
				// inline the left and right hand sides
				BranchStmt copy = super.copy(stmt);
				super.replace(stmt, copy);
				CFGStmt thenInlined = apply(stmt.getThen());
				CFGStmt elseInlined = apply(stmt.getElse());
				
//				System.out.println("stmt: " + stmt);
//				System.out.println("copy: " + copy);
//				System.out.println("thenInlined: " + thenInlined);
//				System.out.println("elseInlined: " + elseInlined);
				
				// if both branches deleted, delete stmt
				if (thenInlined == null && elseInlined == null) {
					super.replace(stmt, null);
					return null;
				}
				
				// if branches deleted, turn into assume
				else if (thenInlined == null) {
					SpecStmt assume = target().newAssume(copy.condition().not());
					assume.setNext(elseInlined);
					super.replace(stmt, assume);
					return assume;
				} 
				
				// if else branch deleted, turn into assume
				else if (elseInlined == null) {
					SpecStmt assume = target().newAssume(copy.condition());
					assume.setNext(thenInlined);
					super.replace(stmt, assume);
					return assume;
				}
				
				// if branches are the same, replace with then branch
				else if (thenInlined == elseInlined) {
					super.replace(stmt, thenInlined);
					return null;
				}
				
				// otherwise, just transform into copy
				else {
					copy.setThen(thenInlined);
					copy.setElse(elseInlined);
					super.replace(stmt, copy);
					return copy;
				}
			}

			@Override
			protected UpdateStmt transform(AssignStmt stmt) {
				return transformNext(stmt, super.copy(stmt));
			}
			
			@Override
			protected UpdateStmt transform(CreateStmt stmt) {
				return transformNext(stmt, super.copy(stmt));
			}
			
			@Override
			protected UpdateStmt transform(SpecStmt stmt) {
				return transformNext(stmt, super.copy(stmt));
			}
			
			private UpdateStmt transformNext(UpdateStmt stmt, UpdateStmt inlined) {
				super.replace(stmt, inlined);
				CFGStmt inlinedNext = apply(stmt.getNext());
				
				if (inlinedNext == null) {
					super.replace(stmt, null);
					return null;
				} else {
					inlined.setNext(inlinedNext);
					return inlined;
				}
			}

			@Override
			protected CFGStmt transform(ExitStmt stmt) {
				super.replace(stmt, callExit);
				return null;
			}

			@Override
			protected ForgeExpression replaceExpr(ForgeExpression expr) {
				return expr.accept(localReplacer);
			}
		}
	}
	
	/**
	 * Replaces local variables with a fresh copy of them.
	 */
	private static final class LocalReplacer extends ExpressionReplacer {
		
		private final String suffix;
		private final Bag<LocalVariable> quants;
		
		private LocalReplacer(int suffix) {
			this.suffix = "_" + suffix;
			quants = new HashBag<LocalVariable>();
		}

		@Override
		protected ForgeExpression visit(QuantifyExpression expr) {
			quants.addAll(expr.decls().locals());
			ForgeExpression replaced = super.visit(expr);
			quants.removeAll(expr.decls().locals());
			return replaced;
		}

		@Override
		protected ForgeVariable visit(ForgeVariable expr) {
			if (quants.contains(expr)) return expr;
			ForgeVariable replaced = expr.isGlobal() ? expr :
				freshLocal((LocalVariable)expr);
			super.putCache(expr, replaced);
			return replaced;
		}
		
		private LocalVariable freshLocal(LocalVariable local) {
			return local.program().newLocalVariable(
					local.name() + suffix, local.type());
		}
	}
	
}
