package forge.transform;

import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Set;

import forge.cfg.AssignStmt;
import forge.cfg.BranchStmt;
import forge.cfg.CFGStmt;
import forge.cfg.CallStmt;
import forge.cfg.CreateStmt;
import forge.cfg.ExitStmt;
import forge.cfg.ForgeCFG;
import forge.cfg.SpecStmt;
import forge.cfg.UpdateStmt;
import forge.dataflow.AnalysisResults;
import forge.dataflow.ReachabilityAnalysis;
import forge.program.ForgeExpression;

/**
 * Transformer that unrolls loops of the procedure and all called procedures.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class UnrollTransformer implements Transformer {
	
	private final int numUnrolls;
	
	private final List<TopUnroller> transformers;
	private boolean changed;

	/**
	 * Constructs a loop unroller that unrolls for the given number of times.
	 * 
	 * @throws IllegalArgumentException - numUnrolls < 1
	 */
	public UnrollTransformer(int numUnrolls) {
		if (numUnrolls < 1) {
			throw new IllegalArgumentException("numUnrolls = " + numUnrolls);
		}
		this.numUnrolls = numUnrolls;
		transformers = new LinkedList<TopUnroller>();
	}

	public ForgeCFG transform(ForgeCFG source) {
		transformers.clear();
		ForgeCFG target = source;
		changed = true;
		
		while (changed) {
			changed = false;
			TopUnroller transf = new TopUnroller();
			target = transf.transform(target);
			transformers.add(transf);
		}
		
		return target;
	}
	
	public CFGStmt sourceStmt(CFGStmt targetStmt) {
		CFGStmt sourceStmt = targetStmt;
		for (ListIterator<TopUnroller> i = transformers.listIterator(
				transformers.size()); i.hasPrevious();) {
			sourceStmt = i.previous().sourceStmt(sourceStmt);
		}
		return sourceStmt;
	}

	/**
	 * Unrolls the outermost loops.
	 */
	private final class TopUnroller extends AbstractTransformer {
		
		private TopUnroller() {}
		
		@Override
		protected TransformVisitor visitor(ForgeCFG source) {
			return new UnrollVisitor(source);
		}
		
		private final class UnrollVisitor extends TransformVisitor {
			
			private final AnalysisResults<Set<CFGStmt>> reachResults;
			
			private UnrollVisitor(ForgeCFG source) {
				super(TopUnroller.this, source);
				reachResults = ReachabilityAnalysis.ANALYSIS.analyze(source);
			}
			
			@Override
			protected CFGStmt transform(BranchStmt stmt) {
				// try to turn branch into a loop
				// if not a loop, visit it normally
				Loop loop = makeLoop(stmt);
				if (loop == null) return super.transform(stmt);
				
				// otherwise, unroll it
				UnrollTransformer.this.changed = true;
				BranchStmt unrolled = target().newBranch(loop.condition);
				super.replace(stmt, unrolled);
				unrolled.setElse(super.apply(loop.exit));
				LoopCopier looper = new LoopCopier(this, loop, 1);
				unrolled.setThen(looper.apply(loop.entry));
				return unrolled;
			}
			
			private final class LoopCopier extends TransformVisitor {
				
				private final TransformVisitor parent;
				private final Loop loop;
				private final int count;
				
				private LoopCopier(TransformVisitor parent, Loop loop, int count) {
					super(TopUnroller.this, parent.source(), parent.target());
					this.parent = parent;
					this.loop = loop;
					this.count = count;
				}
				
				/**
				 * Returns true iff the node is in the loop.
				 */
				private boolean inLoop(CFGStmt stmt) {
					return reachResults.dataBefore(stmt).contains(loop.branch);
				}

				@Override
				protected CFGStmt transform(BranchStmt stmt) {
					if (!inLoop(stmt)) return useParent(stmt);
					if (!stmt.equals(loop.branch)) return super.transform(stmt);
					
					// if reached the last unroll
					// make this impossible to reach
					if (count == numUnrolls) {
						SpecStmt assume = target().newAssume(loop.condition.not());
						super.replace(stmt, assume);
						assume.setNext(parent.apply(loop.exit));
						return assume;
					}
					
					// otherwise, unroll the loop again
					BranchStmt unrolled = target().newBranch(loop.condition);
					super.replace(stmt, unrolled);
					unrolled.setElse(parent.apply(loop.exit));
					LoopCopier looper = new LoopCopier(parent, loop, count + 1);
					unrolled.setThen(looper.apply(loop.entry));
					return unrolled;
				}

				@Override
				protected UpdateStmt transform(AssignStmt stmt) {
					return inLoop(stmt) ? super.transform(stmt) : useParent(stmt);
				}
				@Override
				protected CFGStmt transform(CallStmt stmt) {
					return inLoop(stmt) ? super.transform(stmt) : useParent(stmt);
				}
				@Override
				protected UpdateStmt transform(CreateStmt stmt) {
					return inLoop(stmt) ? super.transform(stmt) : useParent(stmt);
				}
				@Override
				protected UpdateStmt transform(SpecStmt stmt) {
					return inLoop(stmt) ? super.transform(stmt) : useParent(stmt);
				}
				
				@Override
				protected CFGStmt transform(ExitStmt stmt) {
					assert !inLoop(stmt) : "exit stmt should not be in any loop";
					return useParent(stmt);
				}

				private UpdateStmt useParent(CFGStmt stmt) {
					super.replace(stmt, parent.apply(stmt));
					return null;
				}
			}
			
			/**
			 * If this ifNode is a loop, returns a Loop that wraps it;
			 * otherwise, returns null. Throws an IllegalArgumentException
			 * if it is an infinite loop.
			 * 
			 * A loop has two subnodes - an entry node that is the first statement
			 * in the loop and an exit node at the end of the loop.
			 */
			private Loop makeLoop(BranchStmt stmt) {
				CFGStmt thenNode = stmt.getThen();
				CFGStmt elseNode = stmt.getElse();
				
				if (stmt.equals(thenNode) || stmt.equals(elseNode)) {
					throw new IllegalArgumentException(
							"cannot unroll infinite loops");
				}
				
				Set<CFGStmt> thenReach = reachResults.dataBefore(thenNode);
				Set<CFGStmt> elseReach = reachResults.dataBefore(elseNode);
				assert thenReach != null : "reach results null for " + thenNode;
				assert elseReach != null : "reach results null for " + elseNode;
				boolean thenLoop = thenReach.contains(stmt);
				boolean elseLoop = elseReach.contains(stmt);
				
				// if neither loop or both loop, not the header to a loop
				if (thenLoop == elseLoop) return null;
				return new Loop(stmt, thenLoop);
			}
		}
	}
	
	/**
	 * Wraps an BranchNode with a boolean to indicate the branch that loops.
	 */
	private static final class Loop {
			
		private final CFGStmt branch, entry, exit;
		private final ForgeExpression condition;

		private Loop(BranchStmt branch, boolean thenLoop) {
			this.branch = branch;
			if (thenLoop) {
				condition = branch.condition();
				entry = branch.getThen();
				exit = branch.getElse();
			} else {
				condition = branch.condition().not();
				entry = branch.getElse();
				exit = branch.getThen();
			}
		}
		
		public String toString() {
			return branch.toString();
		}
	}
}
