package forge.translate;

import java.util.HashSet;
import java.util.Set;

import kodkod.ast.Expression;
import kodkod.ast.Formula;
import edu.mit.csail.sdg.util.collections.Pair;
import forge.cfg.AssignStmt;
import forge.cfg.CreateStmt;
import forge.cfg.SpecStmt;
import forge.program.ForgeVariable;
import forge.program.InstanceDomain;
import forge.program.ForgeExpression.Modifiable;
import forge.translate.SymbolicExecutor.AssignHook;
import forge.translate.SymbolicExecutor.BranchHook;
import forge.translate.SymbolicExecutor.CreateHook;
import forge.translate.SymbolicExecutor.SpecHook;

/**
 * A translator strategy that introduces constraints wherever possible in lieu
 * of inlining expressions.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class ConstrainStrategy implements SEStrategy {
	
	public static ConstrainStrategy STRATEGY = new ConstrainStrategy();
	
	private ConstrainStrategy() {}

	public Formula execute(AssignHook hook) {
		AssignStmt stmt = hook.stmt();
		// important that expr is made before var is unconstrained
		// because unconstraining var can affect expr translation
		// make expression
		Expression kexpr = hook.toExpr(stmt.expr());
		
		// constrain new relations
		Expression kvar = hook.unconstrain(stmt.var());
		return kvar.eq(kexpr);
	}

	public Formula execute(CreateHook hook) {
		CreateStmt stmt = hook.stmt();
		
		// make expressions
		InstanceDomain dom = stmt.type();
		Expression currs = hook.toExpr(dom);
		
		// constrain new relations
		Expression kvar = hook.unconstrain(stmt.var());
		Expression kdom = hook.unconstrain(dom);
		
		Formula varOne = kvar.one();
		Formula varNew = kvar.in(currs).not();
		Formula domSet = kdom.eq(currs.union(kvar));
		return Formula.and(varOne, varNew, domSet);
	}

	public Formula execute(SpecHook hook) {
		SpecStmt stmt = hook.stmt();
		Set<Formula> constraints = new HashSet<Formula>();
		
		// each modified variable is represented by a new expression --
		// either a functional expr of previous values or by the spec relation
		for (ForgeVariable var : stmt.modified()) {
			Expression ktype = hook.toExpr(var.type());
			Expression kvar = hook.unconstrain(var);
			constraints.add(kvar.in(ktype));
		}
		
		// add condition to constraints and return conjunction
		Formula cond = hook.toForm(stmt.condition());
		constraints.add(cond);
		return Formula.and(constraints);
	}
	
	public Pair<Formula,Formula> merge(BranchHook hook, Modifiable mod,
			Expression thenExpr, Expression elseExpr) {
		Expression kmod = hook.unconstrain(mod);
		Formula thenCond = hook.cond().implies(kmod.eq(thenExpr));
		Formula elseCond = hook.notCond().implies(kmod.eq(elseExpr));
		return new Pair<Formula,Formula>(thenCond, elseCond);
	}
	
	public String toString() {
		return "Constrain Strategy";
	}
}
