package forge.translate;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import kodkod.ast.Expression;
import edu.mit.csail.sdg.annotations.Effects;
import edu.mit.csail.sdg.annotations.Modifies;
import edu.mit.csail.sdg.annotations.Returns;
import edu.mit.csail.sdg.annotations.SpecField;
import edu.mit.csail.sdg.annotations.Throws;
import forge.program.ForgeExpression;
import forge.program.ForgeProgram;
import forge.program.ForgeExpression.Modifiable;
import forge.util.ModifiableMap;

/**
 * Symbolic execution environment maintained by the translation.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
@SpecField("env : Modifiable -> Expression | env = exprs.elems")
public final class Environment {
	
	/** model containing relations in the environment expressions. */
	private final RelationalModel model;
	
	/** maps Forge modifiable to KK expressions. */
	private final ModifiableMap<Expression> exprs;
	
	//
	// Following two methods used only by translation.
	//
	
	/**
	 * Creates a new environment for translation down a branch.
	 */
	Environment branch() {
		return new Environment(model,
				new ModifiableMap<Expression>(exprs));
	}
	
	/**
	 * Maps modifiables to exprs that have changed since last branch.
	 */
	Map<Modifiable, Expression> diffExprs() {
		return Collections.unmodifiableMap(exprs.diffMap());
	}
	
	//
	// public methods
	//
	
	public Environment(RelationalModel model) {
		this(model, new ModifiableMap<Expression>());
	}
	
	public Environment copy() {
		return new Environment(model,
				new ModifiableMap<Expression>(
						new HashMap<Modifiable, Expression>(exprs)));
	}
	
	public Environment unmodifiable() {
		return new Environment(model,
				((ModifiableMap<Expression>)exprs).unmodifiable());
	}
	
	@Effects("this.model = model && this.exprs = modExprs")
	@Modifies("this.model, this.exprs")
	private Environment(RelationalModel model,
			ModifiableMap<Expression> modExprs) {
		this.model = model;
		this.exprs = modExprs;
	}
	
	/* inherits specification */
	public ForgeProgram program() {
		return model.program();
	}
	
	/* inherits specification */
	public RelationalModel model() {
		return model;
	}
	
	/**
	 * Returns the set of all modifiables in the environment.
	 */
	@Returns("this.env.Expression")
	public Set<Modifiable> mods() {
		return exprs.keySet();
	}
	
	/**
	 * Gets the expression for the given modifiable.
	 */
	@Returns("this.env[mod]")
	@Throws({"NullPointerException : elem = null",
		     "IllegalArgumentException : elem.program != this"})
	public Expression getExpr(Modifiable mod) {
		program().checkProgram(mod);
		return exprs.get(mod);
	}
	
	/**
	 * Sets the expression for the given modifiable.
	 */
	@Effects("this.env[mod] = expr")
	@Modifies("this.env[mod]")
	public void putExpr(Modifiable mod, Expression expr) {
		program().checkProgram(mod);
		checkKExpression(mod, expr);
		exprs.put(mod, expr);
	}
	
	private static final String NEW_LINE = System.getProperty("line.separator");
	public final String toString() {
		StringBuffer sb = new StringBuffer();
		for (Entry<Modifiable, Expression> e : exprs.entrySet()) {
			sb.append(e.getKey()).append(" = ").append(e.getValue()).append(NEW_LINE);
		}
		return sb.toString();
	}
	
	private static void checkKExpression(ForgeExpression expr, Expression kexpr) {
		if (kexpr == null) {
			throw new NullPointerException("null kexpr for expr " + expr);
		}
		if (kexpr != null && (expr.arity() != kexpr.arity())) {
			throw new IllegalArgumentException(kexpr + " arity does not match " + expr);
		}
	}
}
