package forge.translate;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import kodkod.ast.Decls;
import kodkod.ast.Expression;
import kodkod.ast.Formula;
import kodkod.ast.IfExpression;
import kodkod.ast.IntConstant;
import kodkod.ast.IntExpression;
import kodkod.ast.IntToExprCast;
import kodkod.ast.Variable;
import kodkod.ast.operator.ExprOperator;
import edu.mit.csail.sdg.annotations.Returns;
import edu.mit.csail.sdg.util.collections.OverrideMap;
import forge.program.BinaryExpression;
import forge.program.BooleanLiteral;
import forge.program.ConditionalExpression;
import forge.program.ExpressionVisitor;
import forge.program.ForgeDomain;
import forge.program.ForgeExpression;
import forge.program.ForgeLiteral;
import forge.program.ForgeProgram;
import forge.program.ForgeType;
import forge.program.ForgeVariable;
import forge.program.InstanceDomain;
import forge.program.InstanceLiteral;
import forge.program.IntegerLiteral;
import forge.program.LocalVariable;
import forge.program.OldExpression;
import forge.program.ProjectionExpression;
import forge.program.QuantifyExpression;
import forge.program.UnaryExpression;
import forge.program.ForgeExpression.Modifiable;
import forge.translate.RelationalModel.DomainRelations;


/**
 * Translates Forge expressions into Kodkod expressions.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public class ExprTranslator {
	
	/** Kodkod zero constant. */
	private static final IntExpression ZERO = IntConstant.constant(0);

	/** the environment that is the subject of the translation. */
	private final Environment env;
	
	/** derived field: model. */
	private final RelationalModel model;
	
	/** maps Forge variables to old KK expressions. */
	private final Map<ForgeVariable, Expression> oldExprs;
	
	/**
	 * @effects this.env = env
	 */
	public ExprTranslator(Environment env) {
		this.env = env;
		model = env.model();
		oldExprs = new HashMap<ForgeVariable, Expression>();
	}
	
	/* inherits specification */
	public final ForgeProgram program() {
		return model.program();
	}
	
	/* inherits specification */
	public final RelationalModel model() {
		return model;
	}
	
	@Returns("this.env")
	public final Environment env() {
		return env;
	}
	
	/**
	 * Pins the old expression for the duration of the next translation.
	 */
	public final void pinOldExpr(ForgeVariable var) {
		// if there is no old expression for var,
		// then this has no effect, which is fine
		oldExprs.put(var, env.getExpr(var));
	}
	
	/**
	 * Unpins all the old expressions.
	 */
	public final void unpinOldExprs() {
		oldExprs.clear();
	}
	
	/**
	 * Translates the given Forge expression to a Kodkod expression.
	 */
	public final Expression toExpr(ForgeExpression expr) {
		return expr.accept(new ExprTranslVisitor());
	}
	
	/**
	 * Faster translates for a modifiable.
	 */
	public final Expression toExpr(Modifiable mod) {
		return env.getExpr(mod);
	}
	
	/**
	 * Translates the given Forge expression to a Kodkod formula.
	 */
	public final Formula toForm(ForgeExpression expr) { 
		return expr2Form(toExpr(expr));
	}
	
	/**
	 * Translates the given Forge expression to a Kodkod integer expression.
	 */
	public final IntExpression toInt(ForgeExpression expr) { 
		return expr2Int(toExpr(expr));
	}
	
	private final class ExprTranslVisitor extends ExpressionVisitor<Expression> {
		
		/** maps Forge variables to quantified KK expressions. */
		private Map<LocalVariable, Expression> quantExprs;
		
		private ExprTranslVisitor() {
			quantExprs = new HashMap<LocalVariable, Expression>();
		}
		
		protected Expression visit(ForgeLiteral expr) {
			final Expression kexpr;
			
			switch (expr.type().kind()) {
			case BOOLEAN:
				BooleanLiteral boolLit = (BooleanLiteral)expr;
				kexpr = boolLit.value() ? model.trueRelation() : model.falseRelation();
				break;
				
			case INTEGER:
				IntegerLiteral intLit = (IntegerLiteral)expr;
				kexpr = int2Expr(model.intConstant(intLit));
				break;
				
			case INSTANCE:
				InstanceLiteral instLit = (InstanceLiteral)expr;
				kexpr = model.literalRelation(instLit);
				break;
				
			default:
				assert false;
				kexpr = null;
				break;
			}
			
			super.putCache(expr, kexpr);
			return kexpr;
		}
	
		protected Expression visit(ForgeType expr) {
			Expression kexpr = typeExpr(expr, false);
			super.putCache(expr, kexpr);
			return kexpr;
		}
		
		//
		// Do not cache the rest, because they could contain quantified local variables
		//
	
		protected Expression visit(ForgeVariable expr) {
			Expression varExpr = quantExprs.get(expr);
			if (varExpr != null) return varExpr;
			varExpr = env.getExpr(expr);
			if (varExpr == null) {
				throw new IllegalStateException(expr + " not bound");
			}
			return varExpr;
		}
	
		@Override
		protected Expression visit(ProjectionExpression expr) {
			Expression sub = expr.sub().accept(this);
			IntExpression[] cols = new IntConstant[expr.columns().size()];
			
			int i = 0;
			for (int c : expr.columns()) {
				cols[i++] = IntConstant.constant(c);
			}
			
			return sub.project(cols);
		}
		
		@Override
		protected Expression visit(UnaryExpression expr) {
			Expression sub = expr.sub().accept(this);
			return apply(expr.op(), sub);
		}
		
		@Override
		protected Expression visit(BinaryExpression expr) {
			Expression kleft = expr.left().accept(this);
			final Expression kright;
			
			// special optimization for checking subset of type
			ForgeExpression right = expr.right();
			if (expr.op().equals(BinaryExpression.Op.SUBSET) &&
					right instanceof ForgeType) {
				kright = typeExpr((ForgeType)right, true);
			} else {
				kright = right.accept(this);
			}
			
			return compose(expr.op(), kleft, kright);
		}
		
		@Override
		protected Expression visit(QuantifyExpression expr) {
			List<LocalVariable> locals = expr.decls().locals();
			Map<LocalVariable, Variable> quantVars = new HashMap<LocalVariable, Variable>();
			
			// create kodkod variables to quantify over
			for (LocalVariable local : locals) {
				Variable kvar = Variable.nary(local.name(), local.arity());
				quantVars.put(local, kvar);
			}
			
			// translate the quantified formula
			Map<LocalVariable, Expression> prevQuants = quantExprs;
			quantExprs = OverrideMap.override(quantExprs, quantVars);
			Expression ksub = expr.sub().accept(this);
			quantExprs = prevQuants;

			// create the declaration
			Iterator<LocalVariable> i = locals.iterator();
			LocalVariable local = i.next();
			Variable kvar = quantVars.get(local);
			Decls kdecls = kvar.oneOf(typeExpr(local.type(), false));
			while (i.hasNext()) {
				local = i.next();
				kvar = quantVars.get(local);
				kdecls = kdecls.and(kvar.oneOf(typeExpr(local.type(), false)));
			}
			
			// create the full quantified formula
			switch(expr.op()) {
			case ALL:   return form2Expr(expr2Form(ksub).forAll(kdecls));
			case SOME:  return form2Expr(expr2Form(ksub).forSome(kdecls));
			case SUM:   return int2Expr(expr2Int(ksub).sum(kdecls));
			case UNION: return expr2Form(ksub).comprehension(kdecls);
			default:    assert false; return null;
			}
		}
		
		@Override
		protected Expression visit(ConditionalExpression expr) {
			Expression kcond = expr.condition().accept(this);
			Expression kthen = expr.thenExpr().accept(this);
			Expression kelse = expr.elseExpr().accept(this);
			return expr2Form(kcond).thenElse(kthen, kelse);
		}
		
		@Override
		protected Expression visit(OldExpression expr) {
			ForgeVariable var = expr.variable();
			Expression oldExpr = oldExprs.get(var);
			if (oldExpr == null) {
				throw new IllegalStateException("reference to old(" + var +
						") but " + var + " not declared to be modified");
			}
			return oldExpr;
		}
	
	}
	
	//
	// helpers
	//
	
	/** converts a Kodkod expression to a Kodkod formula. */
	private Formula expr2Form(Expression expr) {
		if (expr == model.trueRelation())  return Formula.TRUE;
		if (expr == model.falseRelation()) return Formula.FALSE;
		if (expr instanceof IfExpression) {
			IfExpression ifExpr = (IfExpression)expr;
			if (ifExpr.thenExpr().equals(model.trueRelation()) &&
				ifExpr.elseExpr().equals(model.falseRelation())) {
				return ifExpr.condition();
			}
		}
		return expr.in(model.trueRelation());
	}
	
	/** converts a Kodkod expression to a Kodkod integer. */
	private IntExpression expr2Int(Expression expr) {
		if (expr instanceof IntToExprCast) {
			return ((IntToExprCast)expr).intExpr();
		}
		
		return expr.sum();
	}
	
	/** converts a Kodkod formula to a Kodkod expression. */
	private Expression form2Expr(Formula form) {
		if (form == Formula.TRUE)  return model.trueRelation();
		if (form == Formula.FALSE) return model.falseRelation();
		return form.thenElse(model.trueRelation(), model.falseRelation());
	}
	
	/** converts a Kodkod integer to a Kodkod expression. */
	private Expression int2Expr(IntExpression intExpr) {
		return intExpr.toExpression();
	}
	
	/**
	 * Converts a type to an expression. If extent is true, converts it to the
	 * extent of that type; otherwise, the expression represents the instances
	 * of the type created up to that point.
	 */
	private Expression typeExpr(ForgeType type, boolean extent) {
		Set<? extends ForgeType.Tuple> tuples = type.tupleTypes();
		
		// empty case
		if (tuples.isEmpty()) {
			List<Expression> empties = Collections.nCopies(type.arity(), Expression.NONE);
			return Expression.product(empties);
		}
		
		// non-empty case
		List<Expression> tupleExprs = new ArrayList<Expression>(tuples.size());
		for (ForgeType.Tuple tuple : tuples) {
			tupleExprs.add(typeExpr(tuple, extent));
		}
		return Expression.union(tupleExprs);
	}

	/**
	 * Converts a tuple type to an expression. If extent is true, converts it
	 * to the extent of that type; otherwise, the expression represents the
	 * instances of the type created up to that point.
	 */
	private Expression typeExpr(ForgeType.Tuple tupleType, boolean extent) {
		List<ForgeDomain> doms = tupleType.domains();
		List<Expression> domExprs = new ArrayList<Expression>(doms.size());
		for (ForgeDomain dom : tupleType.domains()) {
			domExprs.add(typeExpr(dom, extent));
		}
		return Expression.compose(ExprOperator.PRODUCT, domExprs);
	}
	
	/**
	 * Converts a domain to an expression. If extent is true, converts it
	 * to the extent of that type; otherwise, the expression represents the
	 * instances of the type created up to that point.
	 */
	private Expression typeExpr(ForgeDomain domain, boolean extent) {
		switch(domain.kind()) {
		case BOOLEAN:
			return model.boolRelation();
			
		case INTEGER:
			return model.intRelation();
			
		case INSTANCE:
			InstanceDomain idom = (InstanceDomain)domain;
			DomainRelations domRels = model.domainRelations(idom);
			
			if (extent) {
				return domRels.extentExpr();
			} else {
				Expression currs = env.getExpr(idom);
				return domRels.literalsExpr().union(currs);
			}
		}
		
		assert false; return null;
	}
	
	/** applies Forge unary operator to Kodkod expression. */
	private Expression apply(UnaryExpression.Op op, Expression sub) {
		switch(op) {
		case CLOSURE: return sub.closure();
		
		case CARDINALITY: return int2Expr(sub.count());
		
		case NO:   return form2Expr(sub.no());
		case LONE: return form2Expr(sub.lone());
		case ONE:  return form2Expr(sub.one());
		case SOME: return form2Expr(sub.some());
		
		case BOOL: return form2Expr(expr2Form(sub));
		case NOT:  return not(sub);
		
		case SUM:     return int2Expr(expr2Int(sub));
		case NEG:     return int2Expr(ZERO.minus(expr2Int(sub)));
		case BIT_NOT: return int2Expr(expr2Int(sub).not());
		}
		
		assert false : "unknown unary operator: " + op;
		return null;
	}
	
	private Expression not(Expression expr) {
		if (expr == model.trueRelation() ) return model.falseRelation();
		if (expr == model.falseRelation()) return model.trueRelation();
		return form2Expr(expr2Form(expr).not());
	}
	
	/** composes Kodkod expressions with Forge binary operator. */
	private Expression compose(BinaryExpression.Op op, Expression left,
			Expression right) {
		switch(op) {
		case UNION:        return left.union(right);
		case DIFFERENCE:   return left.difference(right);
		case INTERSECTION: return left.intersection(right);
		case PRODUCT:      return left.product(right);
		case JOIN:         return left.join(right);
		case OVERRIDE:
			IntExpression[] cols = new IntExpression[left.arity() - 1];
			for (int i = 0; i < cols.length; i++) {
				cols[i] = IntConstant.constant(i);
			}
			Expression project = right.project(cols);
			Expression remove = project.product(Expression.UNIV);
			return left.difference(remove).union(right);
		
		case SUBSET:
			if (left == right) return model.trueRelation();
			return form2Expr(left.in(right));
		case EQUALS:
			if (left == right) return model.trueRelation();
			return form2Expr(left.eq(right));
		
		case AND:
			if (left == right) return left;
			if (left == model.trueRelation()  || right == model.falseRelation()) return right;
			if (left == model.falseRelation() || right == model.trueRelation() ) return left;
			return form2Expr(expr2Form(left).and(expr2Form(right)));
		case OR:
			if (left == right) return left;
			if (left == model.trueRelation()  || right == model.falseRelation()) return left;
			if (left == model.falseRelation() || right == model.trueRelation() ) return right;
			return form2Expr(expr2Form(left).or(expr2Form(right)));
		case IMPLIES:
			if (left == right) return model.trueRelation();
			if (left == model.falseRelation() || right == model.falseRelation()) return not(left);
			if (left == model.trueRelation()  || right == model.trueRelation() ) return right;
			return form2Expr(expr2Form(left).implies(expr2Form(right)));
		case IFF:
			if (left == right) return model.trueRelation();
			if (left  == model.trueRelation() ) return right;
			if (left  == model.falseRelation()) return not(right);
			if (right == model.trueRelation() ) return left;
			if (right == model.falseRelation()) return not(left);
			return form2Expr(expr2Form(left).iff(expr2Form(right)));
		case XOR:
			if (left == right) return model.falseRelation();
			if (left  == model.trueRelation() ) return not(right);
			if (left  == model.falseRelation()) return right;
			if (right == model.trueRelation() ) return not(left);
			if (right == model.falseRelation()) return left;
			Formula leftForm = expr2Form(left);
			Formula rightForm = expr2Form(right);
			Formula orForm = leftForm.or(rightForm);
			Formula nandForm = leftForm.and(rightForm).not();
			return form2Expr(orForm.and(nandForm));
		
		case LT: return form2Expr(expr2Int(left).lt(expr2Int(right)));
		case GT: return form2Expr(expr2Int(left).gt(expr2Int(right)));
		case LTE: return form2Expr(expr2Int(left).lte(expr2Int(right)));
		case GTE: return form2Expr(expr2Int(left).gte(expr2Int(right)));
		
		case PLUS:   return int2Expr(expr2Int(left).plus(expr2Int(right)));
		case MINUS:  return int2Expr(expr2Int(left).minus(expr2Int(right)));
		case TIMES:  return int2Expr(expr2Int(left).multiply(expr2Int(right)));
		case DIVIDE: return int2Expr(expr2Int(left).divide(expr2Int(right)));
		case MODULO: return int2Expr(expr2Int(left).modulo(expr2Int(right)));
		
		case BIT_AND: return int2Expr(expr2Int(left).and(expr2Int(right)));
		case BIT_OR:  return int2Expr(expr2Int(left).or(expr2Int(right)));
		case BIT_XOR: return int2Expr(expr2Int(left).xor(expr2Int(right)));
		case SHL:     return int2Expr(expr2Int(left).shl(expr2Int(right)));
		case SHR:     return int2Expr(expr2Int(left).sha(expr2Int(right)));
		case USHR:    return int2Expr(expr2Int(left).shr(expr2Int(right)));
		}
		
		assert false : "unknown binary operator: " + op;
		return null;
	}
	
}
