package forge.translate;

import java.util.Collections;
import java.util.List;

import kodkod.ast.Formula;
import forge.cfg.AssignStmt;
import forge.cfg.BranchStmt;
import forge.cfg.CFGStmt;
import forge.cfg.CFGVisitor;
import forge.cfg.CallStmt;
import forge.cfg.CreateStmt;
import forge.cfg.ExitStmt;
import forge.cfg.ForgeCFG;
import forge.cfg.SpecStmt;
import forge.cfg.StmtSet;
import forge.program.ForgeProgram;
import forge.program.ForgeExpression.Modifiable;

/**
 * A translation of a Forge procedure to Kodkod.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class ForgeTranslation {
	
	private final RelationalModel model;
	private final Environment initEnv, finlEnv;
	private final List<Formula> assumes;
	private final SliceLog slices;

	/**
	 * Translates the given cfg to Kodkod, logging the slice information
	 * iff the logSlices flag is true.
	 */
	public static ForgeTranslation translate(SEStrategy strategy,
			ForgeCFG cfg, boolean logSlices) {
		SEVisitor tv = new SEVisitor(strategy, cfg, logSlices);
		Environment initEnv = tv.se.env().copy();
		cfg.entry().accept(tv);
		return new ForgeTranslation(initEnv, tv.se);
	}
	
	private ForgeTranslation(Environment initEnv, SymbolicExecutor se) {
		this.initEnv = initEnv.unmodifiable();
		model   = initEnv.model();
		finlEnv = se.env().unmodifiable();
		assumes = Collections.unmodifiableList(se.assumes());
		slices  = se.slices() == null ? null : se.slices().unmodifiable();
	}
	
	public ForgeProgram program() {
		return model.program();
	}
	
	public RelationalModel model() {
		return model;
	}
	
	/**
	 * Returns the initial environment.
	 */
	public Environment initialEnv() {
		return initEnv;
	}
	
	/**
	 * Returns the final environment.
	 */
	public Environment finalEnv() {
		return finlEnv;
	}
	
	/**
	 * Returns the set of assume constraints accumulated across the translation.
	 */
	public List<Formula> assumes() {
		return assumes;
	}
	
	/**
	 * Returns the set of statements responsible for the given assume formula.
	 */
	public StmtSet slice(Formula assume) {
		return checkSlices().getSlice(assume);
	}
	
	/**
	 * Returns the set of statements that impact the given modifiable.
	 */
	public StmtSet slice(Modifiable mod) {
		return checkSlices().getSlice(mod);
	}
	
	private SliceLog checkSlices() {
		if (slices == null) {
			throw new IllegalStateException("slices were not logged");
		}
		return slices;
	}
	
	private static final class SEVisitor implements CFGVisitor {
		
		/** symbolic executor. */
		private final SymbolicExecutor se;
		
		/** join-points information. */
		private final JoinPoints joins;
		
		/** current join point. */
		private final CFGStmt join;
		
		SEVisitor(SEStrategy strategy, ForgeCFG cfg,
				boolean logSlices) {
			se = SymbolicExecutor.initialize(strategy, cfg, logSlices);
			joins = JoinPoints.find(cfg);
			join = cfg.exit();
		}
		
		private SEVisitor(SEVisitor parent, CFGStmt join) {
			SymbolicExecutor pse = parent.se;
			se = new SymbolicExecutor(pse.strategy(), pse.env().branch(),
					pse.slices() == null ? null : pse.slices().branch());
			joins = parent.joins;
			this.join = join;
		}
		
		public void visit(AssignStmt stmt) {
			if (stmt.equals(join)) return;
			se.execute(stmt);
			stmt.getNext().accept(this);
		}
		
		public void visit(CreateStmt stmt) {
			if (stmt.equals(join)) return;
			se.execute(stmt);
			stmt.getNext().accept(this);
		}
		
		public void visit(SpecStmt stmt) {
			if (stmt.equals(join)) return;
			se.execute(stmt);
			stmt.getNext().accept(this);
		}
		
		public void visit(CallStmt stmt) {
			assert false : "all calls must have been inlined";
		}
		
		public void visit(BranchStmt stmt) {
			if (stmt.equals(join)) return;
			CFGStmt newJoin = joins.joinPointOf(stmt);
			
			// translate the then and else branches and merge
			SEVisitor thenTV = new SEVisitor(this, newJoin);
			SEVisitor elseTV = new SEVisitor(this, newJoin);
			stmt.getThen().accept(thenTV);
			stmt.getElse().accept(elseTV);
			se.merge(stmt, thenTV.se, elseTV.se);

			// continue from the join point of this branch
			newJoin.accept(this);
		}
		
		public void visit(ExitStmt stmt) {
			return;
		}
	}
}