package forge.translate;

import java.util.HashSet;
import java.util.Set;

import kodkod.ast.Expression;
import kodkod.ast.Formula;
import edu.mit.csail.sdg.util.collections.Pair;
import forge.cfg.AssignStmt;
import forge.cfg.CreateStmt;
import forge.cfg.SpecStmt;
import forge.program.ForgeExpression;
import forge.program.ForgeVariable;
import forge.program.InstanceDomain;
import forge.program.ForgeExpression.Modifiable;
import forge.translate.SymbolicExecutor.AssignHook;
import forge.translate.SymbolicExecutor.BranchHook;
import forge.translate.SymbolicExecutor.CreateHook;
import forge.translate.SymbolicExecutor.SpecHook;

/**
 * A translator that inlines for updates but adds constraints for branches
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class HybridStrategy implements SEStrategy {
	
	public static final HybridStrategy STRATEGY = new HybridStrategy();
	
	private HybridStrategy() {}
	
	public Formula execute(AssignHook hook) {
		AssignStmt stmt = hook.stmt();
		
		// make kexpr
		Expression kexpr = hook.toExpr(stmt.expr());
		
		// bind var
		hook.bind(stmt.var(), kexpr);
		return Formula.TRUE;
	}
	
	public Formula execute(CreateHook hook) {
		CreateStmt stmt = hook.stmt();
		
		// make exprs
		InstanceDomain dom = stmt.type();
		Expression currs = hook.toExpr(dom);
		Expression order = hook.orderExpr(dom);
		Expression first = hook.firstExpr(dom);
		Expression next  = currs.no().thenElse(first, currs.join(order).difference(currs));

		// bind mods
		hook.bind(stmt.var(), next);
		hook.bind(stmt.type(), currs.union(next));
		return next.some();
	}
	
	public Formula execute(SpecHook hook) {
		SpecStmt stmt = hook.stmt();
		Set<Formula> constraints = new HashSet<Formula>();
		
		// each modified variable is represented by a new expression --
		// either a functional expr of previous values or by the spec relation
		for (ForgeVariable var : stmt.modified()) {
			ForgeExpression explicit = stmt.explicitExpr(var);
			
			if (explicit == null) {
				Expression ktype = hook.toExpr(var.type());
				Expression kvar = hook.unconstrain(var);
				constraints.add(kvar.in(ktype));
			} else {
				Expression kexpr = hook.toExpr(explicit);
				hook.bind(var, kexpr);
			}
		}
		
		// add condition to constraints and return conjunction
		Formula cond = hook.toForm(stmt.condition());
		constraints.add(cond);
		return Formula.and(constraints);
	}
	
	public Pair<Formula,Formula> merge(BranchHook hook, Modifiable mod,
			Expression thenExpr, Expression elseExpr) {
		Expression kmod = hook.unconstrain(mod);
		Formula thenCond = hook.cond().implies(kmod.eq(thenExpr));
		Formula elseCond = hook.notCond().implies(kmod.eq(elseExpr));
		return new Pair<Formula,Formula>(thenCond, elseCond);
	}
	
	public String toString() {
		return "Hybrid Strategy";
	}
}
