package forge.translate;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import edu.mit.csail.sdg.annotations.Returns;
import forge.cfg.BranchStmt;
import forge.cfg.CFGStmt;
import forge.cfg.ForgeCFG;
import forge.dataflow.AnalysisResults;
import forge.dataflow.InverseDominatorAnalysis;

/**
 * Mapping from branchs to their join points.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
final class JoinPoints {
	
	/**
	 * Finds the join points for the given procedure.
	 */
	static JoinPoints find(ForgeCFG cfg) {
		AnalysisResults<Set<CFGStmt>> invDomResults =
			InverseDominatorAnalysis.ANALYSIS.analyze(cfg);
		return new JoinPoints(invDomResults, new HashMap<BranchStmt, CFGStmt>());
	}
	
	private final AnalysisResults<Set<CFGStmt>> invDomResults;
	private final Map<BranchStmt, CFGStmt> branch2Join;
	
	private JoinPoints(AnalysisResults<Set<CFGStmt>> invDomResults,
			Map<BranchStmt, CFGStmt> branch2Join) {
		this.invDomResults = invDomResults;
		this.branch2Join = branch2Join;
	}
	
	/**
	 * Returns the join point for the given branch.
	 */
	@Returns("this.branch2Join[branch]")
	CFGStmt joinPointOf(BranchStmt branch) {
		CFGStmt join = branch2Join.get(branch);
		if (join != null) return join;
		Set<CFGStmt> invDoms = invDomResults.dataBefore(branch);
		assert invDoms != null : "null inverse dominators for " + branch;
		
		// the join point is the stmt whose inverse dominators
		// contains all the inverse dominators of the if stmt
		// excluding the if stmt itself
		for (CFGStmt j : invDoms) {
			if (invDomResults.dataBefore(j).size() == invDoms.size() - 1) {
				branch2Join.put(branch, j);
				return j;
			}
		}
		
		throw new IllegalArgumentException("no join point for " + branch);
	}
	
}
