package forge.translate;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import kodkod.ast.Expression;
import kodkod.ast.IntConstant;
import kodkod.ast.Relation;
import edu.mit.csail.sdg.annotations.Effects;
import edu.mit.csail.sdg.annotations.Returns;
import edu.mit.csail.sdg.util.collections.HashMultiMap;
import edu.mit.csail.sdg.util.collections.MultiMap;
import forge.cfg.CFGStmt;
import forge.program.ForgeProcedure;
import forge.program.ForgeProgram;
import forge.program.GlobalVariable;
import forge.program.InstanceDomain;
import forge.program.InstanceLiteral;
import forge.program.IntegerLiteral;
import forge.program.LocalVariable;
import forge.program.ProgramElement;
import forge.program.ForgeExpression.Modifiable;

/**
 * Stores the relations for the translator.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class RelationalModel implements ProgramElement {

	private final ForgeProcedure proc;
	private final Relation boolRel, intRel, trueRel, falseRel;
	private final Map<GlobalVariable, Relation> global2Rel;
	private final Map<LocalVariable, Relation> param2Rel;
	private final Map<InstanceLiteral, Relation> lit2Rel;
	private final Map<InstanceDomain, DomainRelations> dom2Rels;
	private final Map<CFGStmt, StmtRelations> stmt2Rels;
	
	//
	// methods only called by the translation
	//
	
	RelationalModel(ForgeProcedure proc) {
		this.proc = proc;
		ForgeProgram prog = proc.program();
		
		// create relations for the boolean literals
		trueRel  = Relation.nary(prog.trueLiteral().name(), 1);
		falseRel = Relation.nary(prog.falseLiteral().name(), 1);
		
		// create relations for the instance literals
		lit2Rel = new HashMap<InstanceLiteral, Relation>();
		MultiMap<InstanceDomain, InstanceLiteral> idom2Literals =
			new HashMultiMap<InstanceDomain, InstanceLiteral>();
		for (InstanceLiteral lit : prog.instanceLiterals()) {
			idom2Literals.put(lit.type(), lit);
			addLiteralRelation(lit);
		}
		
		// create relations for the primitive domains
		boolRel = Relation.nary(prog.booleanDomain().name(), 1);
		intRel  = Relation.nary(prog.integerDomain().name(), 1);
		
		// create relations for the instance domains
		dom2Rels = new HashMap<InstanceDomain, DomainRelations>();
		for (InstanceDomain idom : prog.instanceDomains()) {
			idom2Literals.addKey(idom);
			Set<InstanceLiteral> literals = idom2Literals.get(idom);
			addDomainRelation(idom, literals);
		}
		
		// create relation for each global variable
		global2Rel = new HashMap<GlobalVariable, Relation>();
		for (GlobalVariable var : prog.globalVariables()) {
			addGlobalRelation(var);
		}
		
		// create relation for each in parameter
		param2Rel = new HashMap<LocalVariable, Relation>();
		for (LocalVariable var : proc.inParams().locals()) {
			addParamRelation(var);
		}
		
		// create relations for all the assume statements
		stmt2Rels = new HashMap<CFGStmt, StmtRelations>();
	}
	
	Relation ensureRelation(CFGStmt stmt, Modifiable mod) {
		assert stmt.cfg().procedure().equals(proc);
		
		StmtRelations stmtRels = stmt2Rels.get(stmt);
		if (stmtRels == null) {
			stmtRels = new StmtRelations();
			stmt2Rels.put(stmt, stmtRels);
		}
		Relation rel = stmtRels.mod2Rel.get(mod);
		if (rel == null) {
			rel = Relation.nary(mod.name() + "_" + stmt.id(), mod.arity());
			stmtRels.mod2Rel.put(mod, rel);
		}
		return rel;
	}
	
	//
	// public methods
	//
	
	/* inherits specification */
	public ForgeProgram program() {
		return proc.program();
	}
	
	@Returns("this.boolRel")
	public Relation boolRelation() {
		return boolRel;
	}
	
	@Returns("this.intRel")
	public Relation intRelation() {
		return intRel;
	}
	
	@Returns("this.dom2Rels.elems[dom]")
	public DomainRelations domainRelations(InstanceDomain dom) {
		return dom2Rels.get(dom);
	}
	
	@Returns("this.trueRel")
	public Relation trueRelation() {
		return trueRel;
	}
	
	@Returns("this.falseRel")
	public Relation falseRelation() {
		return falseRel;
	}
	
	@Effects("return.value = lit.value")
	public IntConstant intConstant(IntegerLiteral lit) {
		return IntConstant.constant(lit.value());
	}
	
	@Returns("this.lit2Rel.elems[lit]")
	public Relation literalRelation(InstanceLiteral lit) {
		return lit2Rel.get(lit);
	}
	
	@Returns("this.param2Rel.elems[param]")
	public Relation paramRelation(LocalVariable param) {
		return param2Rel.get(param);
	}
	
	@Returns("this.global2Rel.elems[global]")
	public Relation globalRelation(GlobalVariable global) {
		return global2Rel.get(global);
	}
	
	public Iterable<Entry<CFGStmt,StmtRelations>> stmtRelations() {
		return Collections.unmodifiableSet(stmt2Rels.entrySet());
	}
	
	//
	// Classes for multiple relations.
	//
	
	/**
	 * Relations created for an InstanceDomain.
	 */
	public static final class DomainRelations {
		private final Relation first, create, order, init;
		private final Expression literals, extent;
		
		private DomainRelations(Relation create, Relation first, Relation order,
				Relation init, Expression literals) {
			this.create = create;
			this.first = first;
			this.order = order;
			this.init = init;
			this.literals = literals;
			extent = literals.union(create);
		}
		
		public Relation createRelation()  { return create;  }
		public Relation firstRelation()   { return first;   }
		public Relation orderRelation()   { return order;   }
		public Relation initialRelation() { return init;    }
		
		public Expression literalsExpr() { return literals; }
		public Expression extentExpr()   { return extent;   }
	}
	
	/**
	 * Relations created for a statement.
	 */
	public static final class StmtRelations
			implements Iterable<Entry<Modifiable,Relation>> {
		
		private final Map<Modifiable, Relation> mod2Rel;
		
		private StmtRelations() {
			mod2Rel = new HashMap<Modifiable, Relation>();
		}
		
		public Iterator<Entry<Modifiable,Relation>> iterator() {
			return Collections.unmodifiableSet(mod2Rel.entrySet()).iterator();
		}
	}
	
	//
	// Private helper methods for adding relations
	//
	
	private void addDomainRelation(InstanceDomain domain,
			Set<InstanceLiteral> literals) {
		String name = domain.name();
		Relation create  = Relation.nary(name, 1);
		Relation first   = Relation.nary(name + "_first", 1);
		Relation initial = Relation.nary(name + "_initial", 1);
		Relation order   = Relation.nary(name + "_order", 2);
		
		Expression litsExpr = Expression.NONE;
		for (InstanceLiteral lit : literals) {
			Relation litRel = lit2Rel.get(lit);
			litsExpr = litsExpr.union(litRel);
		}
		
		DomainRelations instanceRels =
			new DomainRelations(create, first, order, initial, litsExpr);
		dom2Rels.put(domain, instanceRels);
	}
	
	private Relation addLiteralRelation(InstanceLiteral literal) {
		Relation rel = Relation.nary(literal.name(), 1);
		lit2Rel.put(literal, rel);
		return rel;
	}
	
	private void addGlobalRelation(GlobalVariable var) {
		Relation rel = Relation.nary(var.name(), var.arity());
		global2Rel.put(var, rel);
	}
	
	private void addParamRelation(LocalVariable var) {
		Relation rel = Relation.nary(var.name(), var.arity());
		param2Rel.put(var, rel);
	}
}
