package forge.translate;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import kodkod.ast.Expression;
import kodkod.ast.Formula;
import kodkod.ast.Relation;
import edu.mit.csail.sdg.util.collections.Pair;
import forge.cfg.AssignStmt;
import forge.cfg.BranchStmt;
import forge.cfg.CreateStmt;
import forge.cfg.ForgeCFG;
import forge.cfg.ReferenceStmt;
import forge.cfg.SpecStmt;
import forge.cfg.StmtSet;
import forge.cfg.UpdateStmt;
import forge.program.ForgeExpression;
import forge.program.ForgeProgram;
import forge.program.ForgeVariable;
import forge.program.GlobalVariable;
import forge.program.InstanceDomain;
import forge.program.LocalVariable;
import forge.program.ForgeExpression.Modifiable;
import forge.translate.RelationalModel.DomainRelations;
import forge.util.ExpressionUtil;

/**
 * Symbolically executes statements.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class SymbolicExecutor extends ExprTranslator {
	
	//
	// Note on assumes: the order in which constraints are added affects the
	// order in which they are removed during unsat score extraction
	//

	/** se strategy */
	private final SEStrategy strategy;
	
	/** the assumes being accumulated. */
	private final List<Formula> assumes;
	
	/** records slices when not null. */
	private final SliceLog slices;
	
	static public SymbolicExecutor initialize(SEStrategy strategy,
			ForgeCFG cfg, boolean logSlices) {
		RelationalModel model = new RelationalModel(cfg.procedure());
		Environment env = new Environment(model);
		SymbolicExecutor se = new SymbolicExecutor(strategy, env,
				logSlices ? new SliceLog(cfg) : null);
		
		// set the initial environment and constraints
		ForgeProgram program = cfg.program();
		for (InstanceDomain dom : program.instanceDomains()) {
			DomainRelations domRels = model.domainRelations(dom);
			Relation initRel = domRels.initialRelation();
			env.putExpr(dom, initRel);
			Formula prefix = domRels.orderRelation().join(initRel).in(initRel);
			se.assumes.add(prefix);
		}
		for (GlobalVariable global : program.globalVariables()) {
			Relation globalRel = model.globalRelation(global);
			env.putExpr(global, globalRel);
			Formula exists = globalRel.in(se.toExpr(global.type()));
			se.assumes.add(exists);
		}
		for (LocalVariable param : cfg.procedure().inParams().locals()) {
			Relation paramRel = model.paramRelation(param);
			env.putExpr(param, paramRel);
			Formula exists = paramRel.in(se.toExpr(param.type()));
			se.assumes.add(exists);
		}
		
		return se;
	}
	
	public SymbolicExecutor(SEStrategy strategy, Environment env,
			SliceLog slices) {
		super(env);
		this.strategy = strategy;
		this.assumes = new LinkedList<Formula>();
		this.slices = slices;
	}
	
	public SEStrategy strategy() {
		return strategy;
	}
	
	public List<Formula> assumes() {
		return assumes;
	}
	
	public SliceLog slices() {
		return slices;
	}
	
	public void execute(AssignStmt stmt) {
		AssignHook hook = new AssignHook(stmt);
		addAssumes(strategy.execute(hook), hook);
	}
	
	public void execute(CreateStmt stmt) {
		CreateHook hook = new CreateHook(stmt);
		addAssumes(strategy.execute(hook), hook);
	}
	
	public void execute(SpecStmt stmt) {
		for (ForgeVariable var : stmt.modified()) {
			super.pinOldExpr(var);
		}
		SpecHook hook = new SpecHook(stmt);
		addAssumes(strategy.execute(hook), hook);
		super.unpinOldExprs();
	}
	
	private void addAssumes(Formula formula, UpdateHook<?> hook) {
		Set<Formula> assumes = ExpressionUtil.roots(formula);
		for (Formula assume: assumes) {
			this.assumes.add(assume);
			if (slices != null) {
				StmtSet slice = slices.getSlice(assume);
				slice.addAll(hook.refSlice);
			}
		}
	}
	
	/**
	 * Merge symbolic execution from both sides of a branch.
	 */
	public void merge(BranchStmt stmt, SymbolicExecutor seThen, SymbolicExecutor seElse) {
		Map<Modifiable, Expression> thenDiff = seThen.env().diffExprs();
		Map<Modifiable, Expression> elseDiff = seElse.env().diffExprs();
		BranchHook hook = new BranchHook(stmt, seThen.slices, seElse.slices, elseDiff);
		
		// merge then exprs and slices
		hook.thenHook = true;
		for (Entry<Modifiable, Expression> e: thenDiff.entrySet()) {
			Modifiable mod = e.getKey();
			Expression thenExpr = e.getValue();
			Expression elseExpr = seElse.toExpr(mod);
			if (elseExpr == null) continue;
			
			// merge expressions
			mergeMod(hook, mod, thenExpr, elseExpr);
		}
		// merge then assumes and slices
		mergeAssumes(hook.cond(), seThen, hook.refSlice);
		
		// merge else exprs and slices
		hook.thenHook = false;
		for (Entry<Modifiable, Expression> e: elseDiff.entrySet()) {
			Modifiable mod = e.getKey();
			
			// where mod is initialized but assigned only on else branch
			if (thenDiff.containsKey(mod)) continue;
			Expression elseExpr = e.getValue();
			Expression currExpr = super.toExpr(mod);
			if (currExpr == null) continue;
			
			// merge mod
			mergeMod(hook, mod, currExpr, elseExpr);
		}
		// merge else assumes and slices
		mergeAssumes(hook.notCond(), seElse, hook.refSlice);
	}
	
	private void mergeMod(BranchHook hook, Modifiable mod, Expression thenExpr,
			Expression elseExpr) {
		if (thenExpr == null || elseExpr == null) throw new NullPointerException();
		
		Pair<Formula,Formula> assume = strategy.merge(hook, mod, thenExpr, elseExpr);
		Set<Formula> thenAssumes = ExpressionUtil.roots(assume.first());
		Set<Formula> elseAssumes = ExpressionUtil.roots(assume.second());
		for (Formula f: thenAssumes) assumes.add(f);
		for (Formula f: elseAssumes) assumes.add(f);
		
		if (slices != null) {
			for (Formula thenAssume: thenAssumes) {
				StmtSet thenSlice = slices.getSlice(thenAssume);
				thenSlice.addAll(hook.refSlice);
				thenSlice.addAll(hook.thenSlices.getSlice(mod));
			}
			for (Formula elseAssume: elseAssumes) {
				StmtSet elseSlice = slices.getSlice(elseAssume);
				elseSlice.addAll(hook.refSlice);
				elseSlice.addAll(hook.elseSlices.getSlice(mod));
			}
		}
	}
	
	private void mergeAssumes(Formula cond, SymbolicExecutor se, StmtSet branchSlice) {
		for (Formula assume : se.assumes) {
			Formula branchAssume = cond.implies(assume);
			assumes.add(branchAssume);
			
			if (slices != null) {
				StmtSet slice = slices.getSlice(branchAssume);
				slice.addAll(se.slices.getSlice(assume));
				slice.addAll(branchSlice);
			}
		}
	}
	
	/**
	 * The callback hook called by the strategy methods.
	 */
	private abstract class Hook<S extends ReferenceStmt> {
		
		private final S stmt;
		final StmtSet refSlice;
		
		private Hook(S stmt) {
			this.stmt = stmt;
			refSlice = (slices == null) ? null : referenceSlice(stmt);
		}
		
		public final S stmt() {
			return stmt;
		}
		
		/**
		 * Returns the total ordering over instances of the given domain.
		 */
		public final Expression orderExpr(InstanceDomain dom) {
			return SymbolicExecutor.this.model().domainRelations(dom).orderRelation(); 
		}
		
		/**
		 * Returns the first element in the total ordering of instances.
		 */
		public final Expression firstExpr(InstanceDomain dom) {
			return SymbolicExecutor.this.model().domainRelations(dom).firstRelation(); 
		}
		
		/**
		 * Translates the given Forge expression into a Kodkod formula.
		 */
		public final Formula toForm(ForgeExpression expr) {
			return SymbolicExecutor.this.toForm(expr);
		}
		
		/**
		 * Translates the given Forge expression into a Kodkod expression.
		 */
		public final Expression toExpr(ForgeExpression expr) {
			return SymbolicExecutor.this.toExpr(expr);
		}
		
		/**
		 * Unconstrains the given modifiable and returns its new expression.
		 */
		public final Expression unconstrain(Modifiable mod) {
			Relation rel = SymbolicExecutor.this.model().ensureRelation(stmt, mod);
			SymbolicExecutor.this.env().putExpr(mod, rel);
			if (slices != null) unconstrainSlices(mod);
			return rel;
		}
		
		abstract void unconstrainSlices(Modifiable mod);
		
		/**
		 * Binds the given modifiable to the given Kodkod expression.
		 */
		public final void bind(Modifiable mod, Expression kexpr) {
			SymbolicExecutor.this.env().putExpr(mod, kexpr);
			if (slices != null) bindSlices(mod);
		}
		
		abstract void bindSlices(Modifiable mod);
	}
	
	/**
	 * The callback hook for the execute methods.
	 */
	private abstract class UpdateHook<S extends UpdateStmt> extends Hook<S> {
		
		private UpdateHook(S stmt) {
			super(stmt);
		}
		
		void unconstrainSlices(Modifiable mod) {
			slices.getSlice(mod).clear();
		}
		
		void bindSlices(Modifiable mod) {
			StmtSet slice = slices.getSlice(mod);
			slice.clear();
			slice.addAll(refSlice);
		}
	}
	
	/**
	 * Callback hook for execute(AssignStmt) method.
	 */
	public final class AssignHook extends UpdateHook<AssignStmt> {
		private AssignHook(AssignStmt stmt) {
			super(stmt);
		}
	}
	
	/**
	 * Callback hook for execute(CreateStmt) method.
	 */
	public final class CreateHook extends UpdateHook<CreateStmt> {
		private CreateHook(CreateStmt stmt) {
			super(stmt);
		}
	}
	
	/**
	 * Callback hook for execute(AssignStmt) method.
	 */
	public final class SpecHook extends UpdateHook<SpecStmt> {
		private SpecHook(SpecStmt stmt) {
			super(stmt);
		}
	}
	
	/**
	 * Callback hook for the merge method.
	 */
	public final class BranchHook extends Hook<BranchStmt> {
		
		private final Formula cond, notCond;
		private final SliceLog thenSlices, elseSlices;
		private final Map<Modifiable, Expression> elseDiff;
		private boolean thenHook;
		
		private BranchHook(BranchStmt stmt, SliceLog thenSlices,
				SliceLog elseSlices, Map<Modifiable, Expression> elseDiff) {
			super(stmt);
			this.thenSlices = thenSlices;
			this.elseSlices = elseSlices;
			this.elseDiff = elseDiff;
			cond = SymbolicExecutor.this.toForm(stmt.condition());
			notCond = cond.not();
		}
		
		public Formula cond() {
			return cond;
		}
		
		public Formula notCond() {
			return notCond;
		}
		
		void unconstrainSlices(Modifiable mod) {
			// ensure the then and else slice logs have copies for this mod
			// before the parent slices are cleared.
			thenSlices.getSlice(mod);
			elseSlices.getSlice(mod);
			slices.getSlice(mod).clear();
		}
		
		void bindSlices(Modifiable mod) {
			StmtSet slice = slices.getSlice(mod);
			
			if (thenHook) {
				// where mod is assigned on both branches
				if (elseDiff.containsKey(mod)) {
					slice.clear();
					slice.addAll(elseSlices.getSlice(mod));
				}
				slice.addAll(thenSlices.getSlice(mod));
			} else {
				slice.addAll(elseSlices.getSlice(mod));
			}
			
			slice.addAll(refSlice);
		}
	}
	
	/**
	 * Returns an unmod copy of the current set of statements influencing
	 * the modifiables referenced by the given statement (does include stmt).
	 */
	private StmtSet referenceSlice(ReferenceStmt stmt) {
		StmtSet slice = StmtSet.hash(stmt.cfg());
		slice.add(stmt);
		for (Modifiable mod : stmt.referenced()) {
			slice.addAll(slices.getSlice(mod));
		}
		return slice.unmodifiable();
	}
}
