package forge.translate;

import java.util.HashSet;
import java.util.Set;

import kodkod.ast.BinaryExpression;
import kodkod.ast.Expression;
import kodkod.ast.Formula;
import kodkod.ast.IfExpression;
import kodkod.ast.UnaryExpression;
import edu.mit.csail.sdg.util.collections.Pair;
import forge.cfg.AssignStmt;
import forge.cfg.CreateStmt;
import forge.cfg.SpecStmt;
import forge.program.ForgeVariable;
import forge.program.InstanceDomain;
import forge.program.ForgeExpression.Modifiable;
import forge.translate.SymbolicExecutor.AssignHook;
import forge.translate.SymbolicExecutor.BranchHook;
import forge.translate.SymbolicExecutor.CreateHook;
import forge.translate.SymbolicExecutor.SpecHook;

/**
 * 
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public final class ThesisStrategy implements SEStrategy {
	
	public final boolean inlineAssign, inlineBranch, inlineCreate;
	
	public ThesisStrategy(boolean inlineAssign, boolean inlineBranch,
			boolean inlineCreate) {
		this.inlineAssign = inlineAssign;
		this.inlineBranch = inlineBranch;
		this.inlineCreate = inlineCreate;
	}
	
	public Formula execute(AssignHook hook) {
		AssignStmt stmt = hook.stmt();
		return inlineAssign ? inlineAssign(stmt, hook) :
			                  constrainAssign(stmt, hook);
	}

	public Formula execute(CreateHook hook) {
		CreateStmt stmt = hook.stmt();
		return inlineCreate ? inlineCreate(stmt, hook) :
			                  constrainCreate(stmt, hook);
	}
	
	public Formula execute(SpecHook hook) {
		SpecStmt stmt = hook.stmt();
		Set<Formula> constraints = new HashSet<Formula>();
		
		// each modified variable is represented by a new expression --
		// either a functional expr of previous values or by the spec relation
		for (ForgeVariable var : stmt.modified()) {
			Expression ktype = hook.toExpr(var.type());
			Expression kvar = hook.unconstrain(var);
			constraints.add(kvar.in(ktype));
		}
		
		// add condition to constraints and return conjunction
		Formula cond = hook.toForm(stmt.condition());
		constraints.add(cond);
		return Formula.and(constraints);
	}
	
	public Pair<Formula,Formula> merge(BranchHook hook, Modifiable mod,
			Expression thenExpr, Expression elseExpr) {
		return inlineBranch ? inlineMerge(hook, mod, thenExpr, elseExpr) :
				              constrainMerge(hook, mod, thenExpr, elseExpr);
	}
	
	public String toString() {
		return "a" + (inlineAssign ? "I" : "C") + " " +
		       "b" + (inlineBranch ? "I" : "C") + " " +
		       "c" + (inlineCreate ? "I" : "C");
	}
	
	private static Formula inlineAssign(AssignStmt stmt, AssignHook hook) {
		// make kexpr
		Expression kexpr = hook.toExpr(stmt.expr());
		
		// bind var
		hook.bind(stmt.var(), kexpr);
		return Formula.TRUE;
	}
	
	private static Formula constrainAssign(AssignStmt stmt, AssignHook hook) {
		// important that expr is made before var is unconstrained
		// because unconstraining var can affect expr translation
		// make expression
		Expression kexpr = hook.toExpr(stmt.expr());
		
		// constrain new relations
		Expression kvar = hook.unconstrain(stmt.var());
		return kvar.eq(kexpr);
	}
	
	private static Formula inlineCreate(CreateStmt stmt, CreateHook hook) {
		// make exprs
		InstanceDomain dom = stmt.type();
		Expression currs = hook.toExpr(dom);
		Expression order = hook.orderExpr(dom);
		Expression first = hook.firstExpr(dom);
		Expression next  = currs.no().thenElse(first, currs.join(order).difference(currs));

		// bind mods
		hook.bind(stmt.var(), next);
		hook.bind(stmt.type(), currs.union(next));
		return next.some();
	}
	
	private static Formula constrainCreate(CreateStmt stmt, CreateHook hook) {
		// make exprs
		InstanceDomain dom = stmt.type();
		Expression currs = hook.toExpr(dom);
		Expression order = hook.orderExpr(dom);
		Expression first = hook.firstExpr(dom);
		Expression next  = currs.no().thenElse(first, currs.join(order).difference(currs));
		
		// constrain new relations
		Expression kvar = hook.unconstrain(stmt.var());
		Expression kdom = hook.unconstrain(dom);
		Formula varEq = kvar.eq(next);
		Formula domEq = kdom.eq(currs.union(kvar));
		return varEq.and(domEq);
	}
	
	/** Constant that merge always returns. */
	private static final Pair<Formula,Formula> TRUE_PAIR =
		new Pair<Formula, Formula>(Formula.TRUE, Formula.TRUE);
	
	private static Pair<Formula, Formula> inlineMerge(BranchHook hook,
			Modifiable mod, Expression thenExpr, Expression elseExpr) {
		hook.bind(mod, merge(hook, thenExpr, elseExpr));
		return TRUE_PAIR;
	}
	
	private static Pair<Formula, Formula> constrainMerge(BranchHook hook,
			Modifiable mod, Expression thenExpr, Expression elseExpr) {
		Expression kmod = hook.unconstrain(mod);
		Formula thenCond = hook.cond().implies(kmod.eq(thenExpr));
		Formula elseCond = hook.notCond().implies(kmod.eq(elseExpr));
		return new Pair<Formula,Formula>(thenCond, elseCond);
	}
	
	private static Expression merge(BranchHook hook, Expression thenExpr, Expression elseExpr) {
		// if equal return it
		if (thenExpr == elseExpr) return thenExpr;

		// collapse two ifs into one if possible
		if (thenExpr instanceof IfExpression) {
			IfExpression thenIf = (IfExpression)thenExpr;
			Formula thenCond = thenIf.condition();
			if (thenIf.thenExpr() == elseExpr) {
				return hook.cond().and(thenCond.not()).
					thenElse(thenIf.elseExpr(), elseExpr);
			}
			if (thenIf.elseExpr() == elseExpr) {
				return hook.cond().and(thenCond).
					thenElse(thenIf.thenExpr(), elseExpr);
			}
		}

		// collapse two ifs into one if possible
		if (elseExpr instanceof IfExpression) {
			IfExpression elseIf = (IfExpression)elseExpr;
			Formula elseCond = elseIf.condition();
			if (elseIf.thenExpr() == thenExpr) {
				return hook.notCond().and(elseCond.not()).
					thenElse(elseIf.elseExpr(), thenExpr);
			}
			if (elseIf.elseExpr() == thenExpr) {
				return hook.notCond().and(elseCond).
					thenElse(elseIf.thenExpr(), thenExpr);
			}
		}

		// optimized merging of binary expressions
		if (thenExpr instanceof BinaryExpression
				&& elseExpr instanceof BinaryExpression) {
			BinaryExpression thenBin = (BinaryExpression) thenExpr;
			BinaryExpression elseBin = (BinaryExpression) elseExpr;
			if (thenBin.op().equals(elseBin.op())) {
				if (thenBin.left() == elseBin.left()) {
					return thenBin.left().compose(thenBin.op(),
							merge(hook, thenBin.right(), elseBin.right()));
				} else if (thenBin.right() == elseBin.right()) {
					return merge(hook, thenBin.left(), elseBin.left()).compose(
							thenBin.op(), thenBin.right());
				}
			}
		}

		// optimized merging of unary expressions
		if (thenExpr instanceof UnaryExpression
				&& elseExpr instanceof UnaryExpression) {
			UnaryExpression thenUn = (UnaryExpression) thenExpr;
			UnaryExpression elseUn = (UnaryExpression) elseExpr;
			if (thenUn.op().equals(elseUn.op())) {
				return merge(hook, thenUn.expression(), elseUn.expression())
						.apply(thenUn.op());
			}
		}

		// return a new if expression
		return hook.cond().thenElse(thenExpr, elseExpr);
	}
}
