package forge.util;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.Queue;
import java.util.Set;

import forge.cfg.AssignStmt;
import forge.cfg.BranchStmt;
import forge.cfg.CFGStmt;
import forge.cfg.CFGVisitor;
import forge.cfg.CallStmt;
import forge.cfg.CreateStmt;
import forge.cfg.ExitStmt;
import forge.cfg.ForgeCFG;
import forge.cfg.SpecStmt;

/**
 * Performs a breadth-first traversal of the call graph and then the
 * statements within the call graph.
 *
 * @author Greg Dennis (gdennis@mit.edu)
 */
public abstract class BreadthFirstVisitor implements CFGVisitor {
	
	protected BreadthFirstVisitor() {}
	
	/**
	 * Performs a breadth-first traversal of the cfgs in the call graph, and
	 * then of the statements within each cfg.
	 */
	public final void traverse(ForgeCFG cfg) {
		Queue<ForgeCFG> cfgWorklist = new LinkedList<ForgeCFG>();
		Set<ForgeCFG> cfgVisited = new HashSet<ForgeCFG>();
		cfgWorklist.add(cfg);
		
		while(!cfgWorklist.isEmpty()) {
			ForgeCFG c = cfgWorklist.remove();
			
			if (cfgVisited.add(c)) {
				visit(c);
				
				Queue<CFGStmt> stmtWorklist = new LinkedList<CFGStmt>();
				Set<CFGStmt> stmtVisited = new HashSet<CFGStmt>();
				stmtWorklist.add(c.entry());
				stmtVisited.add(c.exit());
				
				while(!stmtWorklist.isEmpty()) {
					CFGStmt stmt = stmtWorklist.remove();
					if (stmtVisited.add(stmt)) {
						stmt.accept(this);
						stmtWorklist.addAll(stmt.succs());
						
						if (stmt instanceof CallStmt) {
							cfgWorklist.add(((CallStmt)stmt).called());
						}
					}
				}
				
				c.exit().accept(this);
			}
		}
	}
	
	protected void visit(ForgeCFG cfg) {}
	
	protected void defaultVisit(CFGStmt stmt) {}
	
	public void visit(AssignStmt stmt) { defaultVisit(stmt); }
	public void visit(BranchStmt stmt) { defaultVisit(stmt); }
	public void visit(CallStmt   stmt) { defaultVisit(stmt); }
	public void visit(CreateStmt stmt) { defaultVisit(stmt); }
	public void visit(ExitStmt   stmt) { defaultVisit(stmt); }
	public void visit(SpecStmt   stmt) { defaultVisit(stmt); }
	
}
