package forge.util;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.Iterator;

import kodkod.engine.bool.BooleanFormula;
import kodkod.engine.bool.Operator;
import kodkod.util.ints.IntBitSet;
import kodkod.util.ints.IntSet;
import forge.solve.ForgeReporter;

/**
 * A reporter that prints the final circuit in eDIMACS form
 * to a given file.
 * 
 * @specfield file: File
 * 
 * @author Emina Torlak, Greg Dennis (gdennis@mit.edu)
 */
public final class CircuitReporter extends ForgeReporter {
	
	private static final String NEW_LINE = System.getProperty("line.separator");
	
	private final File file;
	
	/**
	 * Constructs a new circuit reporter whose {@link
	 * #translatingToCNF(kodkod.engine.bool.BooleanFormula)} method will dump
	 * its argument to the given file, in eDIMACS format.
	 * 
	 * @effects this.file' = file
	 */
	public CircuitReporter(File file) {
		this.file = file;
	}
	
	/**
	 * Dumps the given circuit to this.file, in eDIMACS format, if the circuit
	 * is not just a variable. If it is a variable, the file is unchanged.
	 * 
	 * @effects this.file.contents' = circuit in BooleanVariable =>
	 *     this.file.contents else eDIMACS(circuit)
	 */
	@Override
	public final void translatingToCNF(Object circuitObj) {
		BooleanFormula circuit = (BooleanFormula)circuitObj;
		if (circuit.op() == Operator.VAR) return;
		Writer writer = null;
		
		try {
			writer = new BufferedWriter(new FileWriter(file));
			final int maxLabel = Math.abs(circuit.label());
			
			if (circuit.label() > 0) { 
				writeHeader(writer, maxLabel);
			} else { // must add a NOT gate at the top
				writeHeader(writer, maxLabel+1);
				writeGate(writer, circuit.op());
				writeIO(writer, maxLabel+1);
				writeInputs(writer, circuit.iterator());
			}
			
			write(writer, circuit, new IntBitSet(maxLabel+1));
		} catch (IOException e) {
			throw new RuntimeException(e);
		} finally {
			if (writer != null) {
				try { writer.close(); }
				catch (IOException ignore) {}
			}
		}
		
	}
	
	/**
	 * Writes the eDIMACS header to the given writer.
	 * 
	 * @effects writer.contents' = writer.contents + "p noncnf " + vars + " \n"
	 */
	private static void writeHeader(Writer writer, int vars) throws IOException {
		writer.write("p noncnf ");
		writer.write(String.valueOf(vars));
		writer.write(NEW_LINE);
	}
	
	/**
	 * Writes the given io label to the writer, followed by a space.
	 * 
	 * @effects writer.contents' = writer.contents + String.valueOf(i) + " "
	 */
	private static void writeIO(Writer writer, int i) throws IOException {
		writer.write(String.valueOf(i));
		writer.write(" ");
	}
	
	/**
	 * Writes the eDIMACS op code for the given operator to the writer, followed
	 * by a space, -1 (indicating the gate has no parameters), and another space.
	 * 
	 * @requires op in {AND, OR, ITE, NOT}
	 * @effects writer.contents' = writer.contents + eDIMACS(op) + " -1 "
	 */
	private static void writeGate(Writer writer, Operator op) throws IOException {
		switch(op.ordinal()) {
		case 0 : /* AND */ writer.write("4"); 	break;
		case 1 : /* OR  */ writer.write("6"); 	break;
		case 2 : /* ITE */ writer.write("12"); 	break;
		case 3 : /* NOT */ writer.write("3"); 	break;
		default: 
			throw new IllegalArgumentException("unsupported gate");
		}
		writer.write(" -1 ");
	}
	
	/**
	 * Writes out the labels of the gates returned by the inputs iterator to
	 * the writer, separated by spaces and terminated by "0" and an EOL.
	 * 
	 * @effects writes the labels of the gates returned by the inputs iterator
	 * to the writer, separated by spaces and terminated by "0" and an EOL.
	 */
	private static void writeInputs(Writer writer, Iterator<BooleanFormula> inputs) throws IOException {
		while(inputs.hasNext()) {
			writeIO(writer, inputs.next().label());
		}
		writer.write("0");
		writer.write(NEW_LINE);
	}
	
	/**
	 * Recursively dumps the eDIMACS representation of the given circuit to the
	 * given writer.  The visited set is used to keep track of visited gates.  
	 * The method only keeps track of gates with more than one input.
	 * 
	 * @throws IOException 
	 * @effects recursively dumps the eDIMACS representation of the given
	 * circuit to the given writer.
	 */
	private static void write(Writer writer, BooleanFormula circuit, IntSet visited) throws IOException {
		switch (circuit.size()) {
		case 0: /* variable */ break;
		case 1: /* negation */ write(writer, circuit.input(0), visited); break;
		default : /* and, or, ite */
			if (visited.add(circuit.label())) { // unseen
				writeGate(writer, circuit.op());
				writeIO(writer, circuit.label());
				writeInputs(writer, circuit.iterator());
				for(BooleanFormula in : circuit) {
					write(writer, in, visited);
				}
			}
		}
	}
	
}
