package forge.util;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;

import kodkod.ast.BinaryFormula;
import kodkod.ast.Formula;
import kodkod.ast.NaryFormula;
import kodkod.ast.operator.FormulaOperator;
import edu.mit.csail.sdg.util.collections.Bag;
import edu.mit.csail.sdg.util.collections.HashBag;
import forge.program.BinaryExpression;
import forge.program.ConditionalExpression;
import forge.program.ForgeDomain;
import forge.program.ForgeExpression;
import forge.program.ForgeLiteral;
import forge.program.ForgeType;
import forge.program.ForgeVariable;
import forge.program.InstanceDomain;
import forge.program.LocalVariable;
import forge.program.OldExpression;
import forge.program.ProjectionExpression;
import forge.program.QuantifyExpression;
import forge.program.UnaryExpression;
import forge.program.ForgeDomain.Kind;
import forge.program.ForgeExpression.Modifiable;
import forge.transform.ExpressionDefaulter;
import forge.transform.ExpressionDescender;
import forge.transform.ExpressionReplacer;

/**
 * Utilities with expressions.
 * 
 * @author Kuat Yessenov (kuat@mit.edu), Greg Dennis (gdennis@mit.edu)
 */
public final class ExpressionUtil {

	/** Suppress default public constructor. */
	private ExpressionUtil() {}
	
	/**
	 * Replace old expressions with variables under it.
	 */
	public static ForgeExpression bringGlobalsToPostState(ForgeExpression expr) {
		return expr.accept(new PostStateReplacer());
	}
	
	/**
	 * Replace all global variables with old expressions on these global variables.
	 */
	public static ForgeExpression bringGlobalsToPreState(ForgeExpression expr) {
		return expr.accept(new PreStateReplacer());
	}
	
	/**
	 * Replace all occurrences of variable from to variable to.
	 */
	public static ForgeExpression replaceVariable(ForgeExpression expr,
			ForgeVariable from, ForgeExpression to) {
		assert from != null && to != null;
		final Map<ForgeVariable, ForgeExpression> substitutions =
			new HashMap<ForgeVariable, ForgeExpression>();
		substitutions.put(from, to);
		return expr.accept(new VariableReplacer(substitutions));
	}
	
	/**
	 * Replaces variables according to the map (map is from variables to be
	 * replaced to their desired replacements.)
	 */
	public static ForgeExpression replaceVariables(ForgeExpression expr,
			Map<ForgeVariable, ForgeExpression> map) {
		assert map != null;
		return expr.accept(new VariableReplacer(map));
	}
	
	/**
	 * Pretty-prints the given expression to the given PrintStream.
	 * (Note: System.out is a PrintStream).
	 */
	public static String prettyPrint(ForgeExpression expr) {
		if (expr == null) return "null";
		ExprPrettyPrinter prettyPrinter = new ExprPrettyPrinter();
		expr.accept(prettyPrinter);
		return prettyPrinter.sb.toString();
	}
	
	/**
	 * Flattens conjunctions into a set of top-level constraints.
	 */
	public static Set<ForgeExpression> flatten(ForgeExpression expr) {
		ExprFlattener flattener = new ExprFlattener();
		expr.accept(flattener);
		return Collections.unmodifiableSet(flattener.toplevel);
	}
	
	/**
	 * Returns the set of variables found in the given expression, provided the
	 * expression does not contain any old expressions.
	 */
	public static Set<Modifiable> findModifiables(ForgeExpression expr) {
		ModFinder varFinder = new ModFinder();
		expr.accept(varFinder);
		return Collections.unmodifiableSet(varFinder.found);
	}
	
	/**
	 * Returns the non-trivial roots the formula.
	 */
	public static Set<Formula> roots(Formula formula) {
		final List<Formula> formulas = new LinkedList<Formula>();
		formulas.add(formula);
		
		for (ListIterator<Formula> i = formulas.listIterator(); i.hasNext();) {
			formula = i.next();
			if (formula.equals(Formula.TRUE)) {
				i.remove();
			} else if (formula instanceof BinaryFormula) {
				BinaryFormula bin = (BinaryFormula)formula;
				if (bin.op()==FormulaOperator.AND) {
					i.remove();
					i.add(bin.left());
					i.add(bin.right());
					i.previous();
					i.previous();
				}
			} else if (formula instanceof NaryFormula) {
				NaryFormula nary = (NaryFormula)formula;
				if (nary.op().equals(FormulaOperator.AND)) {
					i.remove();
					for (Formula child : nary) {
						i.add(child);
						i.previous();
					}
				}
			}
		}
		
		return new LinkedHashSet<Formula>(formulas);
	}
	
	/**
	 * Visitor the prints out an expression with indents.
	 * Only extends ExpressionDescender for the void return value.
	 */
	private static class ExprPrettyPrinter extends ExpressionDefaulter<Object> {
		
		private static final int INDENT = 3;
		private static final int MAX_LINE = 80;
		private static final String NEW_LINE = System.getProperty("line.separator");
		
		private final StringBuilder sb;
		private int depth = 0;
		
		private ExprPrettyPrinter() {
			sb = new StringBuilder();
		}
		
		@Override
		protected Object visit(BinaryExpression expr) {
			BinaryExpression.Op op = expr.op();
			
			switch(expr.op()) {
			// associative operators
			case AND: case OR:
				ForgeExpression left = expr.left();
				if ((left instanceof BinaryExpression) && (((BinaryExpression)left).op() == op))
					left.accept(this);
				else {
					depth++;		
					left.accept(this);
					depth--;
				}
				println(op.toString());
				ForgeExpression right = expr.right();
				if ((right instanceof BinaryExpression) && (((BinaryExpression)right).op() == op))
					right.accept(this);
				else {
					depth++;
					right.accept(this);
					depth --;
				}
				break;
				
			case IMPLIES:
				depth++;		
				expr.left().accept(this);
				depth--;
				println(op);
				depth++;
				expr.right().accept(this);
				depth --;
				break;
				
			default:
				println(expr);
			}
			return null;
		}

		@Override
		protected Object visit(UnaryExpression expr) {
			String exprString = expr.toString();
			if (exprString.length() > MAX_LINE) {
				println(expr.op());
				depth++;
				expr.sub().accept(this);
				depth--;
			} else {
				println(exprString);
			}
			return null;
		}

		@Override
		protected Object visit(ConditionalExpression expr) {
			println("if");
			depth++;
			expr.condition().accept(this);
			depth--;
			println("then");
			depth++;
			expr.thenExpr().accept(this);
			depth--;
			println("else");
			depth++;
			expr.elseExpr().accept(this);
			depth--;
			return null;
		}

		@Override
		protected Object visit(QuantifyExpression expr) {
			println(expr.op() + " " + expr.decls());
			return expr.sub().accept(this);
		}
		
		@Override
		protected Object defaultVisit(ForgeExpression expr) {
			println(expr);
			return null;
		}
		
		private void println(Object msg) {
			int indent = INDENT * depth;
			for (int i = 0; i < indent; i++) sb.append(' ');
			sb.append(msg).append(NEW_LINE);
		}
	}
	
	/**
	 * Flattens conjunctions into a set of top-level constraints.
	 */
	private static class ExprFlattener extends ExpressionDefaulter<Object> {
		
		private Set<ForgeExpression> toplevel = new LinkedHashSet<ForgeExpression>();
		
		private ExprFlattener() {}

		@Override
		protected Object visit(BinaryExpression expr) {
			if (expr.op().equals(BinaryExpression.Op.AND)) {
				expr.left().accept(this);
				expr.right().accept(this);
			} else {
				defaultVisit(expr);
			}
			return null;
		}

		@Override
		protected Object defaultVisit(ForgeExpression expr) {
			toplevel.add(expr);
			return null;
		}
	}
		
	private static final class ModFinder extends ExpressionDescender {
		
		private final Set<Modifiable> found;
		private final Bag<LocalVariable> quants;
		
		private ModFinder() {
			found = new HashSet<Modifiable>();
			quants = new HashBag<LocalVariable>();
		}
		
		@Override
		protected void descend(ForgeVariable expr) {
			if (!quants.contains(expr)) {
				found.add(expr);
			}
			super.putCache(expr, null);
		}
		
		@Override
		protected void descend(ForgeType expr) {
			if (expr.isDomain()) {
				ForgeDomain dom = (ForgeDomain)expr;
				if (dom.kind().equals(Kind.INSTANCE)) {
					found.add((InstanceDomain)dom);
				}
			}
			super.putCache(expr, null);
		}
		
		@Override
		protected void descend(QuantifyExpression expr) {
			quants.addAll(expr.decls().locals());
			expr.sub().accept(this);
			quants.removeAll(expr.decls().locals());
			super.putCache(expr, null);
		}
		
		@Override
		protected void descend(OldExpression expr) {
			throw new IllegalArgumentException(
					"findReferenced does not handle old expressions: " + expr);
		}
		
		@Override
		protected void descend(BinaryExpression expr) {
			expr.left().accept(this);
			expr.right().accept(this);
			super.putCache(expr, null);
		}
		@Override
		protected void descend(ConditionalExpression expr) {
			expr.condition().accept(this);
			expr.thenExpr().accept(this);
			expr.elseExpr().accept(this);
			super.putCache(expr, null);
		}
		@Override
		protected void descend(ProjectionExpression expr) {
			expr.sub().accept(this);
			super.putCache(expr, null);
		}
		@Override
		protected void descend(UnaryExpression expr) {
			expr.sub().accept(this);
			super.putCache(expr, null);
		}
		@Override
		protected void descend(ForgeLiteral expr) {
			super.putCache(expr, null);
		}
	}

	/** @author Kuat Yessenov */
	private static final class PostStateReplacer extends ExpressionReplacer {
		@Override protected ForgeExpression visit(OldExpression expr) {		
			ForgeExpression result = expr.variable();
			super.putCache(expr, result);
			return result;
		}
	}
	
	/** @author Kuat Yessenov */
	private static final class PreStateReplacer extends ExpressionReplacer {
		@Override protected ForgeExpression visit(ForgeVariable expr) {		
			ForgeExpression result = expr.isGlobal() ? expr.old() : expr;
			super.putCache(expr, result);
			return result;
		}
	}
	
	/**
	 * Transforms Forge expression tree by substituting variables.
	 * 
	 * @author Kuat Yessenov (kuat@mit.edu)
	 */
	private static final class VariableReplacer extends ExpressionReplacer {
		Map<ForgeVariable, ForgeExpression> substitutions;

		/**
		 * Replaces variables in a Forge expression. 
		 *  
		 * @param substitutions mapping from variables to be replaced to the desired variables
		 */
		VariableReplacer(Map<ForgeVariable, ForgeExpression> substitutions) {
			assert substitutions != null;
			assert !substitutions.keySet().contains(null);
			assert !substitutions.values().contains(null);
			this.substitutions = new HashMap<ForgeVariable, ForgeExpression>(substitutions);
		}

		@Override protected ForgeExpression visit(ForgeVariable expr) {
			ForgeExpression result = substitutions.containsKey(expr) ? substitutions.get(expr) : expr;
			super.putCache(expr, result);
			return result;
		}

		@Override
		protected ForgeExpression visit(OldExpression expr) {
			ForgeExpression replaced = bringGlobalsToPreState(expr.variable().accept(this));						
			super.putCache(expr, replaced);
			return replaced;
		}
	}
}
