package forge.util;

import java.util.AbstractList;
import java.util.AbstractSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import kodkod.engine.bool.BooleanFactory;
import kodkod.engine.bool.BooleanMatrix;
import kodkod.engine.bool.Dimensions;
import kodkod.engine.bool.Int;
import kodkod.engine.config.Options;
import kodkod.instance.Bounds;
import kodkod.instance.Tuple;
import kodkod.instance.TupleFactory;
import kodkod.instance.TupleSet;
import kodkod.instance.Universe;
import kodkod.util.ints.IntCollection;
import kodkod.util.ints.IntIterator;
import kodkod.util.ints.IntSet;
import forge.program.ForgeProgram;
import forge.program.ProgramElement;

/**
 * Implementation support for factories that construct relations with Kodkod
 * TupleSets.
 */
public abstract class RelationFactory<R, U extends R, T extends R, A extends U /*& T*/, O>
		implements ProgramElement {

	private final ForgeProgram program;
	private Universe univ;
	private TupleFactory/*<T>*/ tupler;
	private BooleanFactory booler;
	
	/** Caching relations created from TupleSets. */
	private final Map<TupleSet, R> tuples2Rel = new HashMap<TupleSet, R>();
	
	protected RelationFactory(ForgeProgram program) {
		this.program = program;
	}
	
	protected final void setUniverse(Set<O> atoms) {
		assert this.univ == null;
		univ = new Universe(atoms);
		tupler = univ.factory();
		booler = BooleanFactory.constantFactory(new Options());
	}
	
	//
	// public stuff
	//
	
	public final ForgeProgram program() {
		return program;
	}
	
	/**
	 * Return the kodkod universe used by this factory.
	 */
	public final Universe universe() {
		return univ;
	}
	
	/**
	 * Returns the empty set relation.
	 */
	public final U empty() {
		return makeUnary(emptyTupleSet(1));
	}
	
	/**
	 * Returns the empty relation of the given arity.
	 */
	public final R empty(int arity) {
		return make(emptyTupleSet(arity));
	}
	
	//
	// protected abstract stuff
	//
	
	/**
	 * Constructs a new instance of a TupleSet with a size > 1 and arity > 1.
	 */
	protected abstract R newBig(TupleSet tuples);
	
	/**
	 * Constructs a new instance of a TupleSet with a size = 1 and arity > 1.
	 */
	protected abstract U newUnary(TupleSet tuples);
	
	/**
	 * Constructs a new instance of a TupleSet with a size > 1 and arity = 1.
	 */
	protected abstract T newTuple(TupleSet tuples);
	
	/**
	 * Converts an object stored in the universe to an atom.
	 */
	protected abstract A toAtom(O object);
	
	//
	// protected stuff
	//
	
	/**
	 * Makes a relation from a TupleSet.
	 */
	protected final R make(TupleSet tuples) {
		R rel = tuples2Rel.get(tuples);
		
		if (rel == null) {
			if (tuples.arity() == 1 && tuples.size() == 1) {
				rel = extractAtom(tuples);
			} else if (tuples.arity() == 1) {
				rel = newUnary(tuples);
			} else if (tuples.size() == 1) {
				rel = newTuple(tuples);
			} else {
				rel = newBig(tuples);
			}
			tuples2Rel.put(tuples, rel);
		}
		
		return rel;
	}
	
	/**
	 * Makes a unary relation from a TupleSet.
	 */
	@SuppressWarnings("unchecked")
	protected final U makeUnary(TupleSet tuples) {
		assert tuples.arity() == 1;
		U rel = (U)tuples2Rel.get(tuples);
		
		if (rel == null) {
			if (tuples.size() == 1) {
				rel = extractAtom(tuples);
			} else {
				rel = newUnary(tuples);
			}
			tuples2Rel.put(tuples, rel);
		}
		
		return rel;
	}
	
	/**
	 * Makes a tuple relation from a TupleSet.
	 */
	@SuppressWarnings("unchecked")
	protected final T makeTuple(TupleSet tuples) {
		assert tuples.size() == 1;
		T rel = (T)tuples2Rel.get(tuples);
		
		if (rel == null) {
			if (tuples.arity() == 1) {
				rel = (T)extractAtom(tuples);
			} else {
				rel = newTuple(tuples);
			}
			tuples2Rel.put(tuples, rel);
		}
		
		return rel;
	}
	
	/**
	 * Wraps the given TupleSet as a set of tuples.
	 */
	protected final Set<T> setOfTuples(TupleSet tuples) {
		return new SetOfTuples(tuples);
	}
	
	/**
	 * Wraps the given TupleSet of unary tuples as a set of atoms.
	 */
	protected final Set<A> setOfAtoms(TupleSet tuples) {
		return new SetOfAtoms(tuples);
	}
	
	/**
	 * Wraps the given singleton TupleSet as a list of atoms.
	 */
	protected final List<A> listOfAtoms(TupleSet tuples) {
		return new ListOfAtoms(tuples);
	}
	
	/**
	 * Returns a new bounds object with the universe.
	 */
	protected final Bounds makeBounds() {
		return new Bounds(univ);
	}
	
	/**
	 * Returns the ith atom in the universe.
	 */
	@SuppressWarnings("unchecked")
	protected final O atom(int i) {
		return (O)univ.atom(i);
	}
	
	/**
	 * Constructs a Tuple with the given atoms.
	 */
	protected final Tuple tuple(A... atoms) {
		return tupler.tuple((Object[])atoms);
	}
	
	/**
	 * Returns an empty TupleSet of the given arity.
	 */
	protected final TupleSet emptyTupleSet(int arity) {
		return tupler.noneOf(arity);
	}
	
	/**
	 * Returns a singleton TupleSet containing the given atom.
	 */
	protected final TupleSet singletonTupleSet(O atom) {
		return tupler.setOf(atom);
	}
	
	/**
	 * Returns a singleton TupleSet containing the given tuple.
	 */
	protected final TupleSet singletonTupleSet(Tuple tuple) {
		return tupler.setOf(tuple);
	}
	
	/**
	 * Returns TupleSet that is the closure of a matrix.
	 */
	protected final TupleSet closureTupleSet(BooleanMatrix m) {
		return toTupleSet(m.closure());
	}
	
	/**
	 * Returns TupleSet that is the projection of a matrix over some columns.
	 */
	protected final TupleSet projectTupleSet(BooleanMatrix m, int... columns) {
		Int[] kkInts = new Int[columns.length];
		for (int i = 0; i < columns.length; i++) {
			kkInts[i] = booler.integer(columns[i]);
		}
		return toTupleSet(m.project(kkInts));
	}
	
	/**
	 * Returns TupleSet that is the join of two matrices.
	 */
	protected final TupleSet joinTupleSet(BooleanMatrix m1, BooleanMatrix m2) {
		return toTupleSet(m1.dot(m2));
	}
	
	/**
	 * Returns TupleSet that is the override of two matrices.
	 */
	protected final TupleSet overrideTupleSet(BooleanMatrix m1, BooleanMatrix m2) {
		return toTupleSet(m1.override(m2));
	}
	
	/**
	 * Returns TupleSet that is the union of two matrices.
	 */
	protected final TupleSet unionTupleSet(BooleanMatrix m1, BooleanMatrix m2) {
		return toTupleSet(m1.or(m2));
	}
	
	/**
	 * Returns TupleSet that is the difference of two matrices.
	 */
	protected final TupleSet differenceTupleSet(BooleanMatrix m1, BooleanMatrix m2) {
		return toTupleSet(m1.difference(m2));
	}
	
	/**
	 * Returns TupleSet that is the intersection of two matrices.
	 */
	protected final TupleSet intersectTupleSet(BooleanMatrix m1, BooleanMatrix m2) {
		return toTupleSet(m1.and(m2));
	}
	
	/**
	 * Converts a TupleSet to a BooleanMatrix.
	 */
	protected final BooleanMatrix toBooleanMatrix(TupleSet tuples) {
		IntSet is = tuples.indexView();
		Dimensions dim = Dimensions.square(univ.size(), tuples.arity());
		return booler.matrix(dim, is, is);
	}
	
	//
	// private stuff
	//
	
	/**
	 * Extracts an atom from a TupleSet with a size = 1 and arity = 1.
	 */
	private final A extractAtom(TupleSet tuples) {
		return extractAtom(tuples.iterator().next(), 0);
	}
	
	/**
	 * Extracts an atom at index from a Tuple.
	 */
	@SuppressWarnings("unchecked")
	private final A extractAtom(Tuple tuple, int index) {
		return toAtom((O)tuple.atom(index));
	}
	
	/**
	 * Converts a BooleanMatrix to a TupleSet.
	 */
	private TupleSet toTupleSet(BooleanMatrix matrix) {
		return tupler.setOf(matrix.dimensions().numDimensions(),
				new UncloningIntSet(matrix.denseIndices()));
	}
	
	/**
	 * Wraps a TupleSet as a set of tuples
	 */
	private final class SetOfTuples extends AbstractSet<T> {
		
		private final TupleSet tuples;
		
		private SetOfTuples(TupleSet rel) {
			this.tuples = rel;
		}
		
		public int size() {
			return tuples.size();
		}

		@Override
		public Iterator<T> iterator() {
			return new TupleIter();
		}
		
		private final class TupleIter implements Iterator<T> {
			
			private final Iterator<kodkod.instance.Tuple> i;
			
			private TupleIter() {
				i = tuples.iterator();
			}
			
			public boolean hasNext() {
				return i.hasNext();
			}
			
			public T next() {
				return makeTuple(singletonTupleSet(i.next()));
			}
			
			public void remove() {
				throw new UnsupportedOperationException();
			}
		}
	}
	
	/**
	 * Wraps a TupleSet of unary tuples wrapped as a set of atoms.
	 */
	private class SetOfAtoms extends AbstractSet<A> {
		
		private final TupleSet tuples;
		
		private SetOfAtoms(TupleSet tuples) {
			this.tuples = tuples;
		}
		
		public int size() {
			return tuples.size();
		}

		public Iterator<A> iterator() {
			return new AtomIter();
		}
		
		private final class AtomIter implements Iterator<A> {
			
			private final Iterator<kodkod.instance.Tuple> i;
			
			private AtomIter() {
				i = tuples.iterator();
			}
			
			public boolean hasNext() {
				return i.hasNext();
			}
			
			public A next() {
				return extractAtom(i.next(), 0);
			}
			
			public void remove() {
				throw new UnsupportedOperationException();
			}
		}
	}
	
	/**
	 * Wraps a tuple as a list of atoms.
	 */
	private class ListOfAtoms extends AbstractList<A> {
		
		private final kodkod.instance.Tuple tuple;
		
		private ListOfAtoms(TupleSet rel) {
			this.tuple = rel.iterator().next();
		}

		public A get(int index) {
			return extractAtom(tuple, index);
		}

		@Override
		public int size() {
			return tuple.arity();
		}
		
	}
	
	/**
	 * An IntSet wrapper whose clone method returns the IntSet itself.
	 */
	private static final class UncloningIntSet implements IntSet {

		private final IntSet ints;
		
		public UncloningIntSet(IntSet ints) {
			this.ints = ints;
		}
		
		public IntSet clone() throws CloneNotSupportedException {
			return this;
		}
		
		//
		// the rest just delegate
		//
		public boolean add(int i) {
			return ints.add(i);
		}
		public boolean addAll(IntCollection c) {
			return ints.addAll(c);
		}
		public int ceil(int i) {
			return ints.ceil(i);
		}
		public void clear() {
			ints.clear();
		}
		public boolean contains(int i) {
			return ints.contains(i);
		}
		public boolean containsAll(IntCollection c) {
			return ints.containsAll(c);
		}
		public boolean equals(Object o) {
			return ints.equals(o);
		}
		public int floor(int i) {
			return ints.floor(i);
		}
		public int hashCode() {
			return ints.hashCode();
		}
		public boolean isEmpty() {
			return ints.isEmpty();
		}
		public IntIterator iterator() {
			return ints.iterator();
		}
		public IntIterator iterator(int from, int to) {
			return ints.iterator(from, to);
		}
		public int max() {
			return ints.max();
		}
		public int min() {
			return ints.min();
		}
		public boolean remove(int i) {
			return ints.remove(i);
		}
		public boolean removeAll(IntCollection c) {
			return ints.removeAll(c);
		}
		public boolean retainAll(IntCollection c) {
			return ints.retainAll(c);
		}
		public int size() {
			return ints.size();
		}
		public int[] toArray() {
			return ints.toArray();
		}
		public int[] toArray(int[] array) {
			return ints.toArray(array);
		}
	}
	
}

