/*!
 * \mainpage Squander
 */

/*! \addtogroup API
 * 
 * Public API.
 *  
 * @{
 */
package edu.mit.csail.sdg.squander;

import edu.mit.csail.sdg.squander.engine.ISquander;
import edu.mit.csail.sdg.squander.engine.ISquanderResult;
import edu.mit.csail.sdg.squander.log.Log;
import edu.mit.csail.sdg.squander.options.GlobalOptions;
import edu.mit.csail.sdg.squander.parser.JFSLParserException;
import edu.mit.csail.sdg.squander.spec.TypeCheckException;

/**
 * Provides a public API to Squander.
 * 
 * @author Aleksandar Milicevic (aleks@csail.mit.edu)
 */
public final class Squander {
    
    private Squander() {
        throw new UnsupportedOperationException("instantiation forbidden");
    }
    
    private static ISquander lastSqImpl; 

    // =========================================================================
    // --------------------------- public API ----------------------------------
    // =========================================================================
    
    /**
     * Executes the caller method's specification.  Method parameter types
     * are inferred from <code>methodArgs</code>.  If there's any ambiguity 
     * about that, use {@link Squander#exe(Object, Class[], Object[])}.
     * 
     * <p>
     * <b><i> Use this method only to invoke Squander from within the method
     * which you want to squander an implementation for.</i></b>
     * </p>
     * 
     * @param <R> - the return type
     * @param caller - the caller instance (<code>null</code> if static method)
     * @param methodArgs - invocation arguments
     * @return - the return value of the invocation
     */
    @SuppressWarnings("unchecked")
    public static <R> R exe(Object caller, Object ... methodArgs) {
        Class<?>[] methodParamTypes = new Class<?>[methodArgs.length];
        for (int i = 0; i < methodArgs.length; i++)
            if (methodArgs[i] == null)
                methodParamTypes[i] = Object.class;
            else
                methodParamTypes[i] = methodArgs[i].getClass();
        return (R) exe(caller, methodParamTypes, methodArgs);
    }

    /**
     * Executes the caller method's specification.
     * 
     * <p>
     * <b><i> Use this method only to invoke Squander from within the method
     * which you want to squander an implementation for.</i></b>
     * </p>
     * 
     * @param <R> - the return type
     * @param caller - the caller instance (<code>null</code> if static method)
     * @param methodParamTypes - method parameter types
     * @param methodArgs - invocation arguments
     * @return - the return value of the invocation
     */
    @SuppressWarnings("unchecked")
    public static <R> R exe(Object caller, Class<?>[] methodParamTypes, Object[] methodArgs) {
        Throwable t = new Throwable();
        String clsName = getCallerClassName(t);
        String methodName = getCallerMethod(t);
        return (R) exe(caller, clsName, methodName, methodParamTypes, methodArgs);
    }

    /**
     * Executes the given method's specification.
     * <b><i>Assumes that the method doesn't have any parameters.</b></i>
     * 
     * @param <R> - the return type
     * @param caller - the caller instance (<code>null</code> if static method)
     * @param clsName - declaring class full name
     * @param methodName - method name
     * @return - the return value of the invocation
     */
    @SuppressWarnings("unchecked")
    public static <R> R exe(Object caller, String clsName, String methodName) {
        return (R) exe(caller, clsName, methodName, new Class<?>[0], new Object[0]);
    }

    /**
     * The most general form
     * 
     * @param <R> - the return type
     * @param caller - the caller instance (<code>null</code> if static method)
     * @param clsName - declaring class full name
     * @param methodName - method name
     * @param methodParamTypes - method parameter types
     * @param methodArgs - invocation arguments
     * @return - the return value of the invocation
     */
    @SuppressWarnings("unchecked")
    public static <R> R exe(Object caller, String clsName, String methodName, 
            Class<?>[] methodParamTypes, Object[] methodArgs) {
        Log.log("Using %s engine", GlobalOptions.INSTANCE.engine);
        ISquander sq = GlobalOptions.INSTANCE.getSquanderImpl();
        lastSqImpl = sq;
        try {
            return (R) sq.magic(caller, clsName, methodName, methodParamTypes, methodArgs);
        } catch (TypeCheckException e) {
            StringBuilder msg = new StringBuilder("Typecheck Error:\n");
            msg.append(e.getMessage()).append("\n");
            if (e.source() != null ) {
                msg.append(e.source()).append("\n");
            }
            throw new TypeCheckException(msg.toString(), e);
        } catch (JFSLParserException e) {
            StringBuilder msg = new StringBuilder("Parser Error:\n");
            msg.append(e.getMessage()).append("\n"); 
            if (e.source() != null) {
                msg.append(e.source()).append("\n");
                int col = e.column();
                for (int i = 0; i < e.source().length(); i++) 
                    msg.append(i == col ? "^" : "-");
            }
            throw new JFSLParserException(msg.toString(), e);
        }
    }

    /**
     * Returns the result of the last Squander execution.
     */
    public static ISquanderResult getLastResult() {
        if (lastSqImpl != null)
            return lastSqImpl.getLastResult();
        return null;
    }

    private static String getCallerMethod(Throwable t) {
        return getRightStackTraceElem(t).getMethodName();
    }

    private static String getCallerClassName(Throwable t) {
        return getRightStackTraceElem(t).getClassName();
    }

    private static StackTraceElement getRightStackTraceElem(Throwable t) {
        StackTraceElement[] trace = t.getStackTrace();
        for (int i = 0; i < trace.length; i++) {
            if (!trace[i].getClassName().equals(Squander.class.getName()))
                return trace[i];
        }
        return null;
    }
}

/*! @} */