package edu.mit.csail.sdg.squander.forge_examples;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import forge.cfg.ForgeCFG;
import forge.program.ForgeExpression;
import forge.program.ForgeProcedure;
import forge.program.ForgeProgram;
import forge.program.GlobalVariable;
import forge.program.InstanceDomain;
import forge.program.InstanceLiteral;
import forge.program.LocalVariable;
import forge.solve.ForgeBounds;
import forge.solve.ForgeConstant;
import forge.solve.ForgeSolution;
import forge.solve.ForgeSolver;
import forge.solve.InstanceAtom;
import forge.solve.IntegerAtom;
import forge.solve.SolveOptions;
import forge.solve.ForgeConstant.Tuple;
import forge.util.SystemOutReporter;

/**
 * An example of a Forge program that implements and Binary Search Tree data
 * structure and a couple of operations on it.
 * 
 * @author Aleksandar Milicevic (aleks@csail.mit.edu)
 */
public class BinarySearchTree {

    private ForgeProgram program;
    private InstanceDomain node;
    private InstanceDomain bst;

    private InstanceLiteral nodeInsts[];
    private InstanceLiteral bstInst;

    private GlobalVariable root;    // : bst -> node
    private GlobalVariable key;     // : node -> integer
    private GlobalVariable left;    // : node -> node
    private GlobalVariable right;   // : node -> node
    private GlobalVariable parent;  // : node -> node

    private LocalVariable minRet;

    private ForgeProcedure minProc;

    /**
     * Creates Forge programs, instance domains, global variables,
     * literal instances, procedures  and such
     */
    public BinarySearchTree() {
        program = new ForgeProgram();

        node = program.newInstanceDomain("Node");
        bst = program.newInstanceDomain("BST");

        bstInst = program.newInstanceLiteral("bst", bst);
        int nNodes = 6;
        nodeInsts = new InstanceLiteral[nNodes];
        for (int i = 0; i < nNodes; i++) {
            nodeInsts[i] = program.newInstanceLiteral("node" + Integer.toString(i), node);
        }

        root = program.newGlobalVariable("root", bst.product(node));
        key = program.newGlobalVariable("key", node.product(program.integerDomain()));
        left = program.newGlobalVariable("left", node.product(node));
        right = program.newGlobalVariable("right", node.product(node));
        parent = program.newGlobalVariable("parent", node.product(node));

        minRet = program.newLocalVariable("min_return", program.integerDomain());

        minProc = program.newProcedure("min", program.emptyDecls(), minRet);
    }

    /**
     * Runs the <code>min</code> operation on a binary search tree 
     * purely based on specs for that operation (no method implementation)
     */
    public void runMin() {
        ForgeCFG cfg = cfgMin();
        System.out.println(cfg);
        ForgeSolver solver = new ForgeSolver(cfg, options());
        ForgeBounds bounds = bounds();
        ForgeSolution solution = solver.run(bounds);
        if (solution.trace() == null) {
            System.out.println("NO SOLUTION");
            return;
        }
        ForgeConstant x = solution.trace().evaluate(minRet);
        System.out.println(((IntegerAtom) x.tuples().iterator().next().atoms().get(0)).value());
    }

    /**
     * Returns Forge options for this purpose
     */
    private static SolveOptions options() {
        SolveOptions options = new SolveOptions.Builder(1).reporter(
                SystemOutReporter.REPORTER).build();
        return options;
    }

    /**
     * Returns the specs for the <code>min</code> operation
     */
    private ForgeCFG cfgMin() {
        LocalVariable k = program.newLocalVariable("k", node);
        LocalVariable l = program.newLocalVariable("l", node);
        ForgeExpression condition = minRet.one().and(k.join(key).eq(minRet).forSome(k)).and(
                l.join(key).gte(minRet).forAll(l));
        return ForgeCFG.specification(minProc, Collections.<GlobalVariable> emptySet(), condition);
    }

    /**
     * Returns Forge bounds that bound initial state space, i.e. encode a binary search tree
     * into Forge bounds using <code>boundInitial</code> method.
     */
    private ForgeBounds bounds() {
        Map<InstanceDomain, Integer> dom2bound = new HashMap<InstanceDomain, Integer>();
        dom2bound.put(bst, 0);
        dom2bound.put(node, 0);
        ForgeBounds bounds = new ForgeBounds(program, 4, dom2bound);
        int nNodes = 6;
        InstanceAtom[] nodeAtoms = new InstanceAtom[nNodes];
        for (int i = 0; i < nNodes; i++) {
            nodeAtoms[i] = bounds.instanceAtom(nodeInsts[i]);
        }
        InstanceAtom bstAtom = bounds.instanceAtom(bstInst);

        Tuple rootBound = bstAtom.product(nodeAtoms[0]);
        bounds.boundInitial(root, rootBound, rootBound);

        ForgeConstant leftBound = (nodeAtoms[0].product(nodeAtoms[1])).union(
                nodeAtoms[2].product(nodeAtoms[4])).union(nodeAtoms[1].product(nodeAtoms[5]));
        bounds.boundInitial(left, leftBound, leftBound);

        ForgeConstant rightBound = (nodeAtoms[0].product(nodeAtoms[2])).union(nodeAtoms[1]
                .product(nodeAtoms[3]));
        bounds.boundInitial(right, rightBound, rightBound);

        ForgeConstant parentBound = (nodeAtoms[1].product(nodeAtoms[0])).union(
                nodeAtoms[2].product(nodeAtoms[0])).union(nodeAtoms[3].product(nodeAtoms[1]))
                .union(nodeAtoms[4].product(nodeAtoms[2]))
                .union(nodeAtoms[5].product(nodeAtoms[1]));
        bounds.boundInitial(parent, parentBound, parentBound);

        ForgeConstant keyBound = (nodeAtoms[0].product(bounds.intAtom(5))).union(
                nodeAtoms[1].product(bounds.intAtom(3))).union(
                nodeAtoms[2].product(bounds.intAtom(7))).union(
                nodeAtoms[3].product(bounds.intAtom(4))).union(
                nodeAtoms[4].product(bounds.intAtom(6))).union(
                nodeAtoms[5].product(bounds.intAtom(2)));
        bounds.boundInitial(key, keyBound, keyBound);

        return bounds;
    }

    public static void main(String[] args) {
        BinarySearchTree b = new BinarySearchTree();
        b.runMin();
    }

}
