package edu.mit.csail.sdg.squander.forge_examples;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

import forge.cfg.ForgeCFG;
import forge.program.ForgeExpression;
import forge.program.ForgeProcedure;
import forge.program.ForgeProgram;
import forge.program.GlobalVariable;
import forge.program.InstanceDomain;
import forge.program.InstanceLiteral;
import forge.program.LocalVariable;
import forge.solve.ForgeAtom;
import forge.solve.ForgeBounds;
import forge.solve.ForgeConstant;
import forge.solve.ForgeSolution;
import forge.solve.ForgeSolver;
import forge.solve.InstanceAtom;
import forge.solve.IntegerAtom;
import forge.solve.SolveOptions;
import forge.solve.ForgeConstant.Tuple;
import forge.util.SystemOutReporter;

/**
 * 
 * @author Kuat Yessenov
 */
public class TopologicalSort {
    final int num = 16;
    final int width = 4; // should be at least ceiling(log(num))
    final ForgeProgram program;
    final InstanceDomain node;
    final InstanceDomain list;
    final InstanceLiteral[] nodes = new InstanceLiteral[num];
    final GlobalVariable value, next, elems;
    final ForgeProcedure proc;
    final LocalVariable out;
    
    public TopologicalSort() {
        program = new ForgeProgram();
        node = program.newInstanceDomain("Node");
        list = program.newInstanceDomain("List");        
        out = program.newLocalVariable("out", list);
                
        for (int i = 0; i < num; i++)
            nodes[i] = program.newInstanceLiteral("node"+i, node);

        value = program.newGlobalVariable("value", node.product(program.integerDomain()));
        next = program.newGlobalVariable("next", node.product(node));
        elems = program.newGlobalVariable("elems", list.product(program.integerDomain()).product(node));
        
        proc = program.newProcedure("sort", program.emptyDecls(), out);            
    }
    
    ForgeExpression condition() {
        ForgeExpression list = out.join(elems);
        LocalVariable i1 = program.newLocalVariable("i1", program.integerDomain());
        LocalVariable i2 = program.newLocalVariable("i5", program.integerDomain());
        LocalVariable i3 = program.newLocalVariable("i6", program.integerDomain());
        
        LocalVariable n1 = program.newLocalVariable("n1", node);
        
        return out.one()
            .and(i1.join(list).lone().forAll(i1)) // partial function
            .and(list.join(n1).one().forAll(n1)) // pre-image always a singleton
            .and(i2.lt(i3)
//                    .and(i2.join(list).one()) // aleks commented out
//                    .and(i3.join(list).one()) // aleks commented out
                    .implies(i2.join(list).in(i3.join(list).join(next)).not())
                    .forAll(i2)
                    .forAll(i3)) // consistent with next field
            ;
    }
            
    ForgeCFG cfg() {
        return ForgeCFG.specification(proc, Collections.singleton(elems), condition());
    }
    
    ForgeBounds bounds() {        
        Map<InstanceDomain, Integer> dom2bound = new HashMap<InstanceDomain, Integer>();        
        dom2bound.put(node, 0);
        dom2bound.put(list, 1);        
        ForgeBounds bounds = new ForgeBounds(program, width, dom2bound);
        
        // Extract atoms for instances
        InstanceAtom[] a = new InstanceAtom[num];
        for (int i = 0; i < num; i++)
            a[i] = bounds.instanceAtom(nodes[i]);
        
        // Set-up initial value of next field: from -> to
        ForgeConstant next = a[0].product(a[1]);
        for (int i = 1; i < num - 1; i++) {
            next = next.union(a[i].product(a[i+1]));
            if (i + 2 < num)
                next = next.union(a[i].product(a[i+2]));
        }
        
        System.err.println("Next: " + next);
        bounds.boundInitial(this.next, next, next);
        return bounds;
    }
    
    void run() {
        // Forge solver        
        SolveOptions options = new SolveOptions.Builder(1).reporter(SystemOutReporter.REPORTER).build();    
        ForgeSolver solver = new ForgeSolver(cfg(), options);        
        ForgeSolution solution = solver.run(bounds());
        if (solution.trace() == null) {
            System.err.println("no trace");
            return;
        }
        // Show field value
        ForgeConstant value = solution.trace().evaluate(out.join(elems));
                 
        SortedMap<Integer, ForgeAtom> map = new TreeMap<Integer, ForgeAtom>();
        for (Tuple tuple : value.tuples()) {
            map.put(((IntegerAtom) tuple.atoms().get(0)).value(), tuple.atoms().get(1));
        }
        System.err.println("Sorting: " + map.values());
    }
    
    public static void main(String[] args) {
        TopologicalSort sort = new TopologicalSort();
        sort.run();
    }
}