/*! \addtogroup Specification Specification 
 * This module contains classes for reading and maintaining specifications and converting them to Forge expressions. 
 * @{ 
 */
package edu.mit.csail.sdg.squander.spec;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import edu.mit.csail.sdg.squander.spec.JType.Factory;
import edu.mit.csail.sdg.squander.spec.JType.Unary;
import edu.mit.csail.sdg.squander.utils.ReflectionUtils;



@SuppressWarnings("unchecked")
public class JavaScene {
    
    private Map<JType.Unary, ClassSpec> classes = new HashMap<JType.Unary, ClassSpec>();
    private Map<Object, ClassSpec> obj2spec = new IdentityHashMap<Object, ClassSpec>();
    
    private JMethod method;  
    private Object caller;
    private Object[] methodArgs = new Object[0]; 
    
    private boolean finished = false;
    
    public boolean isFinished() { return finished; }
    
    public void addObj2spec(Object object, ClassSpec cs) { obj2spec.put(object, cs); }
    
    public ClassSpec ensureClass(Class<?> cls) {
        return ensureClass(Factory.instance.newJType(cls));
    }
    
    public ClassSpec ensureClass(JType.Unary type) {
        ClassSpec clzSpec = classes.get(type);
        if (clzSpec == null) {
            //assert !finished : "can't add more classes after calling finish()";
            clzSpec = createClassSpec(type);
            classes.put(type, clzSpec);
            // typecheck
            clzSpec.typecheck();
            // ensure super classes and interfaces
            Class<?> clz = type.clazz();
            for (Class<?> cls : ReflectionUtils.getImmParents(clz)) {
                JType.Unary superType = Factory.instance.newJType(cls, type.typeParams());
                ClassSpec superClzSpec = ensureClass(superType);
                clzSpec.addSuper(superClzSpec);
            }
        }
        return clzSpec;
    }

    public void finish() {
        assert !finished : "don't call finish() more than once";
        // for all classes ensure fields used in their super classes
        while (true) {
            Collection<ClassSpec> toVisit = new ArrayList<ClassSpec>(classSpecs());
            for (ClassSpec cs : toVisit) {
                for (JField jf : cs.usedFieldsAll()) {
                    if (jf.declaringType().clazz() == cs.clz())
                        continue;
                    cs.ensureField(jf.name(), false);
                }
            }
            if (toVisit.size() == classSpecs().size())
                break;
        }
        finished = true;
    }
    
//    private Unary merge(List<Unary> jtypes) {
//        assert jtypes.size() > 1;
//        Unary[] typeParams = null;
//        for (Unary t : jtypes) {
//           Unary[] tp = t.typeParams();
//           if (typeParams == null) {
//               typeParams = tp;
//               continue;
//           }
//           if (typeParams.length != tp.length)
//               throw new RuntimeException("incompatible types: " + Arrays.toString(typeParams) + ", " + Arrays.toString(tp));
//           for (int i = 0; i < typeParams.length; i++)
//               typeParams[i] = JType.Factory.instance.newJType(ReflectionUtils.lca(typeParams[i].clazz(), tp[i].clazz()));
//        }
//        return JType.Factory.instance.newJType(jtypes.iterator().next().clazz(), typeParams);
//    }

    public void setMethod(JMethod method) {
        this.method = method;
        method.spec().typecheck(this);
    }

    public void setArgs(Object[] methodArgs) { 
        if (methodArgs == null)
            this.methodArgs = new Object[0];
        else 
            this.methodArgs = methodArgs; 
    }

    public JMethod method()                    { return method; }
    public MethodSpec methodSpec()             { return method.spec(); }
    public Collection<ClassSpec> classSpecs()  { return classes.values(); }
//    public Collection<ClassSpec> classSpecs()  {
//        Collection<ClassSpec> mergedClassSpecs = new ArrayList<ClassSpec>(classes.size());
//        for (ClassSpec cs : classes.values()) {
//            if (cs.mergedSpec() != null && cs.mergedSpec() != cs)
//                continue;
//            mergedClassSpecs.add(cs);
//        }
//        return mergedClassSpecs; 
//    }
    public Collection<Class<?>> classes()      { 
        Set<Class<?>> classes = new HashSet<Class<?>>();
        for (JType.Unary jtype : this.classes.keySet()) {
            classes.add(jtype.clazz());
        }
        return classes;
    }
    public List<JType.Unary> jtypes(Class<?> cls) {
        List<JType.Unary> result = new LinkedList<Unary>();
        for (JType.Unary t : classes.keySet()) {
            if (t.clazz() == cls)
                result.add(t);
        }
        return result;
    }
    //public ClassSpec classSpec(Class cls)      { return classes.get(Factory.instance.newJType(cls)); }
    public ClassSpec classSpecForObj(Object o) { return obj2spec.get(o); }
    public JType.Unary jtypeForObj(Object o)   { return obj2spec.get(o).jtype(); }
    public ClassSpec classSpec(Unary jtype)    { return classes.get(jtype); }
    public ClassSpec findSpec(Class cls, boolean ensureSingle) {
        ClassSpec result = null; 
        for (Entry<Unary, ClassSpec> e : classes.entrySet()) {
            if (e.getKey().clazz() == cls) {
                if (ensureSingle)
                    if (result != null)
                        throw new RuntimeException("Multiple specs found for class " + cls.getName());
                result = e.getValue();
            }
        }
        return result;
    }
    public void setCaller(Object caller)       { this.caller = caller; }
    public Object caller()                     { return caller; }
    public Object[] methodArgs()               { return methodArgs; }
    
    public Collection<JField> specFields() {
        List<JField> specFields = new ArrayList<JField>();
        for (ClassSpec clsSpec : classSpecs()) {
            for (JField fld : clsSpec.usedFields()) 
                if (fld.isSpec())
                    specFields.add(fld);
        }
        return specFields;
    }

    public Collection<ClassSpec> subTypes(JType.Unary jtype) {
        return subTypes(classSpec(jtype));
    }
    
    public Collection<ClassSpec> subTypes(Class<?> cls) {
        return subTypes(findSpec(cls, false));
    }
    
    private Collection<ClassSpec> subTypes(ClassSpec clsSpec) {
        assert clsSpec != null : "clsSpec is null";
        if (((Class<?>) clsSpec.clz()).isArray())
            return new ArrayList<ClassSpec>(0);
        Collection<ClassSpec> subs = clsSpec.subs();
        if (subs == null) {
            subs = new LinkedList<ClassSpec>();
            for (ClassSpec cs : classes.values()) {
                if (cs == clsSpec)
                    continue;
                if (clsSpec.jtype().isAssignableFrom(cs.jtype()))
                    subs.add(cs);
            }
            clsSpec.setSubs(subs);
        }
        return subs;
    }
    
    public void translateSpecs(ForgeScene forgeScene) {
        assert finished : "call finish() first";
        for (ClassSpec clsSpec : classSpecs()) {
            if (clsSpec.isEmpty()) 
                continue;
            clsSpec.translateSpecs(forgeScene);
        }
        if (method != null) {
            ForgeEnv env = forgeScene.getEnv(forgeScene.thisVar());
            method.spec().translateSpecs(env, forgeScene);
        }
    }
    
    private ClassSpec createClassSpec(JType.Unary jtype) {
        //TODO (after Heap2Lit is used for initial heap traversal): 
        //assert !isFinished() : "can't create class spec after calling finish(); cls = " + jtype;
        ClassSpec clzSpec = new ClassSpec(jtype, this);
        ISpecProvider isp = new CompositeSpecProvider(
                new SpecFileSpecProvider(), new ReflectiveSpecProvider());
        List<Source> clsSpec = isp.extractClassSpec(jtype);
        for (Source src : clsSpec) {
            src.parse();
            if (src.isClause()) {
                clzSpec.addInvariant(src);
            } else if (src.isDecl()) {
                clzSpec.addSpecFieldSource(src);
            } else 
                throw new RuntimeException("invalid class spec: " + src);
        }
        return clzSpec;
    }

}
/*! @} */
