/*! \addtogroup Specification Specification 
 * This module contains classes for reading and maintaining specifications and converting them to Forge expressions. 
 * @{ 
 */
package edu.mit.csail.sdg.squander.spec;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import edu.mit.csail.sdg.annotations.Case;
import edu.mit.csail.sdg.annotations.Effects;
import edu.mit.csail.sdg.annotations.Ensures;
import edu.mit.csail.sdg.annotations.FuncField;
import edu.mit.csail.sdg.annotations.Helper;
import edu.mit.csail.sdg.annotations.Invariant;
import edu.mit.csail.sdg.annotations.Modifies;
import edu.mit.csail.sdg.annotations.NonNull;
import edu.mit.csail.sdg.annotations.Pure;
import edu.mit.csail.sdg.annotations.Requires;
import edu.mit.csail.sdg.annotations.Returns;
import edu.mit.csail.sdg.annotations.SpecField;
import edu.mit.csail.sdg.annotations.Specification;
import edu.mit.csail.sdg.annotations.Throws;
import edu.mit.csail.sdg.squander.annotations.Fresh;
import edu.mit.csail.sdg.squander.annotations.FreshObjects;
import edu.mit.csail.sdg.squander.annotations.Options;
import edu.mit.csail.sdg.squander.spec.Source.Rule;
import edu.mit.csail.sdg.squander.utils.ReflectionUtils;


public class ReflectiveSpecProvider implements ISpecProvider {
    
    private final String TYPE_EXCEPTIONAL = "EXCEPTIONAL";
    /*** Format strings for de-sugaring specifications */
    
    private final String CLAUSE_DEFAULT = "true";
    private final String RETURNS_FORMAT = "return = (%s)";
    private final String NONNULL_FORMAT = "(%s) != null";
        
    public MethodSpec extractMethodSpec(Method method, NameSpace ns) {
        MethodSpec result = new MethodSpec();
        extractMethodSpec(method, result, ns);
        return result;
    }
    
    private void extractMethodSpec(Method method, MethodSpec result, NameSpace ns) {
        extractMethodSpecNonRecursive(method, result, ns);
//        if (!result.cases().isEmpty())
//            return;
        if (!Modifier.isStatic(method.getModifiers())) {
            Class<?> methodDeclCls = method.getDeclaringClass();
            for (Class<?> cls : ReflectionUtils.getImmParents(methodDeclCls)) {
                try {
                    Method m = ReflectionUtils.getMethod(cls, method.getName(), method.getParameterTypes());
                    extractMethodSpec(m, result, ns);
//                    if (!result.cases().isEmpty())
//                        return;
                } catch (NoSuchMethodException e) {
                }
            }
        }
        return;
    }
    
    private void extractMethodSpecNonRecursive(Method method, MethodSpec result, NameSpace ns) {
        // initialize collections for the lightweight spec cases
        Set<String> requires = new LinkedHashSet<String>();
        Set<String> ensures = new LinkedHashSet<String>();
        Set<String> modifies = new LinkedHashSet<String>();
        
        // scan method annotations
        for (Annotation ann : method.getAnnotations()) {
            if (ann instanceof Requires) {
                requires.addAll(convertArray(((Requires) ann).value()));
            } else if (ann instanceof Ensures) {
                ensures.addAll(convertArray(((Ensures) ann).value()));
            } else if (ann instanceof Effects) {
                ensures.addAll(convertArray(((Effects) ann).value()));
            } else if (ann instanceof Modifies) {
                modifies.addAll(convertArray(((Modifies) ann).value()));
            } else if (ann instanceof Returns) {
                ensures.add(String.format(RETURNS_FORMAT, ((Returns) ann).value()));
            } else if (ann instanceof Pure) {
                result.makePure();
            } else if (ann instanceof Helper) {
                result.makePure();
            } else if (ann instanceof NonNull) {
                ensures.add(String.format(NONNULL_FORMAT, "return"));
            } else if (ann instanceof Throws) {
                throw new RuntimeException("Throws annotation not supported");
            } else if (ann instanceof Specification) {
                // full specification mode
                for (Case cs : ((Specification) ann).value()) {
                    // Process each case in the specification
                    final List<String> caseRequires = new ArrayList<String>();
                    final List<String> caseEnsures = new ArrayList<String>();
                    final List<String> caseModifies = new ArrayList<String>();
                    boolean caseExceptional = false;

                    caseRequires.addAll(convertArray(cs.requires()));
                    caseEnsures.addAll(convertArray(cs.ensures()));
                    caseModifies.addAll(convertArray(cs.modifies()));
                    caseExceptional = TYPE_EXCEPTIONAL.equals(cs.type().name());

                    Source pres = new Source(clause(caseRequires), ns, Rule.CLAUSE);
                    Source posts = new Source(clause(caseEnsures), ns, Rule.CLAUSE); 
                    Source frames = new Source(frame(caseModifies), ns, Rule.FRAME); 
                        
                    result.addCase(pres, posts, frames, caseExceptional);
                }
            } else if (ann instanceof Fresh) {
                Fresh fresh = (Fresh) ann;
                for (FreshObjects freshObj : fresh.value()) {
                    result.addFreshObj(freshObj.cls(), freshObj.typeParams(), freshObj.num());
                }
            } else if (ann instanceof FreshObjects) {
                FreshObjects freshObj = (FreshObjects) ann;
                result.addFreshObj(freshObj.cls(), freshObj.typeParams(), freshObj.num());
            } else if (ann instanceof Options) {
                Options opt = (Options) ann;
                result.addOptions(opt);
            }
        }

        // augment lightweight specification case
        if (requires.size() > 0 || ensures.size() > 0 || modifies.size() > 0)
            result.addCase(new Source(clause(requires), ns, Rule.CLAUSE), 
                    new Source(clause(ensures), ns, Rule.CLAUSE), 
                    new Source(frame(modifies), ns, Rule.FRAME), 
                    false);

    }

    /* (non-Javadoc)
     * @see squander.spec.ISpecProvider#extractClassSpec(java.lang.Class)
     */
    public List<Source> extractClassSpec(JType.Unary jtype) {
        List<Source> result = new ArrayList<Source>();
        for (Annotation ann : jtype.clazz().getAnnotations()) {
            if (ann instanceof Invariant) {
                for (String inv : ((Invariant) ann).value())
                    result.add(new Source(inv, NameSpace.forClass(jtype), Rule.CLAUSE));
            } else if (ann instanceof SpecField) {
                for (String sf : ((SpecField) ann).value())
                    result.add(new Source(sf, NameSpace.forClass(jtype), Rule.DECLARATION, false));
            } else if (ann instanceof FuncField) {
                for (String sf : ((FuncField) ann).value())
                    result.add(new Source(sf, NameSpace.forClass(jtype), Rule.DECLARATION, true));
            }
        }
        return result;
    }

    /* (non-Javadoc)
     * @see squander.spec.ISpecProvider#extractFieldSpec(java.lang.reflect.Field)
     */
    public List<Source> extractFieldSpec(Field field, JType.Unary declType) {
        List<Source> result = new ArrayList<Source>();
        for (Annotation ann : field.getAnnotations()) {
            if (ann instanceof NonNull) {
                result.add(new Source(String.format(NONNULL_FORMAT, "this." + field.getName()), 
                        NameSpace.forClass(declType), Rule.CLAUSE));
            } else if (ann instanceof Invariant) {
                for (String inv : ((Invariant) ann).value())
                    result.add(new Source(inv, NameSpace.forClass(declType), Rule.CLAUSE));
            }
        }
        return result;
    }


    /** Extracts elements of a string array annotation */
    private List<String> convertArray(String[] strs) {
        List<String> result = new ArrayList<String>();
        for (String str : strs) {
            if ("@nothing".equals(str))
                continue;
            result.add(str);
        }
        return result;
    }

    private String clause(Collection<String> clauses) {
        if (clauses.size() == 0)
            return CLAUSE_DEFAULT;

        final StringBuilder sb = new StringBuilder();
        for (String s : clauses) {
            if (sb.length() > 0)
                sb.append(" && ");
            sb.append('(').append(s).append(')');
        }
        return sb.toString();
    }

    private String frame(Collection<String> frames) {
        if (frames.size() == 0)
            return "";
                
        final StringBuilder sb = new StringBuilder();
        for (String s : frames) {
            if (sb.length() > 0)
                sb.append(" , ");
            sb.append(s);
        }

        return sb.toString();
    }
}
/*! @} */
